/*! \file Rational.h
    \brief Generic class for Rational numbers
    We define a general class for Rational numbers. We can use this class
    if we have a number type that does not have a field number type, 
    for example @int@.
*/

#ifndef SYNAPS_ARITHM_RATIONAL_H
#define SYNAPS_ARITHM_RATIONAL_H

#include <cassert>
#include <synaps/init.h>
#include <synaps/base/comparison.h>
#include <synaps/arithm/sign.h>

__BEGIN_NAMESPACE_SYNAPS


/** 
 * @brief Generic rational arithmetic class.
 * 
 */
template < typename  NT_ >
class Rational
{
 public:
  typedef NT_        NT;
  Rational() : num_(0), den_(1) { }
  template < typename  RT >
  Rational(const RT& n) : num_(n), den_(1) {}
  template < typename  RT1, typename  RT2 >
  Rational(const RT1& n, const RT2& d) : num_(n), den_(d)
  {
    assert( d != 0);
    adjust_sign();
  }
  
  Rational<NT>& operator+= (const Rational<NT>& r);
  Rational<NT>& operator-= (const Rational<NT>& r);
  Rational<NT>& operator*= (const Rational<NT>& r);
  Rational<NT>& operator/= (const Rational<NT>& r);
  Rational<NT>& operator+= (const NT& r);
  Rational<NT>& operator-= (const NT& r);
  Rational<NT>& operator*= (const NT& r);
  Rational<NT>& operator/= (const NT& r);
  
  Rational<NT>&    normalize();
  Rational<NT>&    adjust_sign();
  
  const NT&   numerator()   const { return num_; }
  const NT&   denominator() const { return den_; }

 public:
  NT   num_;     ///< Numerator
  NT   den_;     ///< Denominator
};
//--------------------------------------------------------------------
namespace arithm 
{
  template<class T>  struct isfield_;
  template<class X>  struct isfield_<Rational<X> > { typedef true_t T; };
}
//--------------------------------------------------------------------

template<class C> inline
bool adjust_denum( const C & x ) { return x < 0;  };

template < typename NT >  inline
Rational<NT>&
Rational<NT>::adjust_sign()
{
  if ( adjust_denum( den_ ) ) { num_ = -num_; den_ = -den_; }
  return *this;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>&
Rational<NT>::normalize()
{
  if ( num_ == den_ ) {
    num_ = den_ = 1;
    return *this;
  }
  
  if (-num_ == den_) {
      num_ = -1;
      den_ = 1;
      return *this;
  }
  NT ggt = gcd(num_, den_);
  if (ggt != 1 ) {
    num_ /= ggt;
    den_ /= ggt;
    adjust_sign();
  }
  return *this;
}
//------------------------------------------------------------------------------------------

// Arithmetic operators
template < typename NT >  inline
Rational<NT>&
Rational<NT>::operator+= (const Rational<NT>& r)
{
  num_ = num_ * r.den_ + r.num_ * den_;
  den_ *= r.den_;
  return *this;
}
//------------------------------------------------------------------------------------------


template < typename NT >  inline
Rational<NT>&
Rational<NT>::operator-= (const Rational<NT>& r)
{
  num_ = num_ * r.den_ - r.num_ * den_;
  den_ *= r.den_;
  return *this;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>&
Rational<NT>::operator*= (const Rational<NT>& r)
{
  num_ *= r.num_;
  den_ *= r.den_;
  return *this;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>& Rational<NT>::operator/= (const Rational<NT>& r)
{
  assert( r.num_ != 0 );
  num_ *= r.den_;
  den_ *= r.num_;
  adjust_sign();
  return *this;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>& Rational<NT>::operator+= (const NT& r)
{
  num_ += r * den_;
  return *this;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>& Rational<NT>::operator-= (const NT& r)
{
  num_ -= r * den_;
  return *this;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>&
Rational<NT>::operator*= (const NT& r)
{
  num_ *= r;
  return *this;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>&
Rational<NT>::operator/= (const NT& r)
{
  assert( r != 0 );
  den_ *= r;
  adjust_sign();
  return *this;
}
//------------------------------------------------------------------------------------------

// Help function. Do not use it :-)
template < typename NT >  inline
int quotient_cmp( const Rational<NT>& x, const Rational<NT>& y )
{
  int xsign = sign(x.num_);
  int ysign = sign(y.num_);
  
  if ((xsign == 0) && (ysign == 0)) return EQUAL;
  if (xsign < ysign) return SMALLER;
  if (xsign > ysign) return LARGER;
  // now x and y have the same sign
  NT ta = x.num_ * y.den_;
  NT tb = y.num_ * x.den_;
  NT t = ta - tb;
  return sign(t);
}
//------------------------------------------------------------------------------------------

template < typename NT > inline
int compare(const Rational<NT>& x, const Rational<NT>& y)
{ return quotient_cmp(x, y); }
//------------------------------------------------------------------------------------------

template < typename NT > inline
std::ostream&
operator<<(std::ostream& s, const Rational<NT>& r)
{
   return s << r.numerator() << "/" << r.denominator();
}
//------------------------------------------------------------------------------------------


//------------------------------------------------------------------------------------------
// Arithmetic with Rationals
//------------------------------------------------------------------------------------------
template < typename NT >  inline
Rational<NT>
operator+(const Rational<NT>& x, const Rational<NT>& y)
{
  Rational<NT> z = x;
  return z += y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator-(const Rational<NT>& x, const Rational<NT>& y)
{ return (Rational<NT>(x) -= y); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator*(const Rational<NT>& x, const Rational<NT>& y)
{
  Rational<NT> z = x;
  return z *= y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator/(const Rational<NT>& x, const Rational<NT>& y)
{
  Rational<NT> z = x;
  return z /= y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator-(const Rational<NT>& x)
{ return Rational<NT>(-x.num_,x.den_); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator+(const NT& x, const Rational<NT>& y)
{
  Rational<NT> z(x);
  return z += y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator+(const Rational<NT>& x, const NT& y)
{
  Rational<NT> z = x;
  return z += y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator-(const NT& x, const Rational<NT>& y)
{
  Rational<NT> z(x);
  return z -= y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator-(const Rational<NT>& x, const NT& y)
{
  Rational<NT> z = x;
  return z -= y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator*(const NT& x, const Rational<NT>& y)
{
  Rational<NT> z(x);
  return z *= y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator*(const Rational<NT>& x, const NT& y)
{
  Rational<NT> z = x;
  return z *= y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator/(const NT& x, const Rational<NT>& y)
{
  Rational<NT> z(x) ;
  return z /= y;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
Rational<NT>
operator/(const Rational<NT>& x, const NT& y)
{
  Rational<NT> z = x;
  return z /= y;
}
//------------------------------------------------------------------------------------------



//------------------------------------------------------------------------------------------
// Booleans operations
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator==(const Rational<NT>& x, const Rational<NT>& y)
{ return x.num_ * y.den_ == x.den_ * y.num_; }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator==(const Rational<NT>& x, const NT& y)
{ return x.den_ * y == x.num_; }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator==(const NT& x, const Rational<NT>& y)
{ return y == x; }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator!=(const Rational<NT>& x, const Rational<NT>& y)
{ return ! (x == y); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator!=(const Rational<NT>& x, const NT& y)
{ return ! (x == y); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator!=(const NT& x, const Rational<NT>& y)
{ return ! (x == y); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator<(const Rational<NT>& x, const Rational<NT>& y)
{
  return quotient_cmp(x,y) == SMALLER;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator<(const Rational<NT>& x, const NT& y)
{
  return quotient_cmp(x,Rational<NT>(y)) == SMALLER;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator<(const NT& x, const Rational<NT>& y)
{
  return quotient_cmp(Rational<NT>(x),y) == SMALLER;
}
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator>(const Rational<NT>& x, const Rational<NT>& y)
{ return y < x; }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator>(const Rational<NT>& x, const NT& y)
{ return y < x; }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator>(const NT& x, const Rational<NT>& y)
{ return y < x; }
//------------------------------------------------------------------------------------------


template < typename NT >  inline
bool
operator<=(const Rational<NT>& x, const Rational<NT>& y)
{ return ! (y < x); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator<=(const Rational<NT>& x, const NT& y)
{ return ! (y < x); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator<=(const NT& x, const Rational<NT>& y)
{ return ! (y < x); }
//------------------------------------------------------------------------------------------


template < typename NT >  inline
bool
operator>=(const Rational<NT>& x, const Rational<NT>& y)
{ return ! (x < y); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator>=(const Rational<NT>& x, const NT& y)
{ return ! (x < y); }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
bool
operator>=(const NT& x, const Rational<NT>& y)
{ return ! (x < y); }
//------------------------------------------------------------------------------------------



// Change this asa!
template < typename NT >
double
to_double(const Rational<NT>& q)
{
  return (to_double(q.numerator()) / to_double(q.denominator()));
}
//------------------------------------------------------------------------------------------


// Global functions for the numerator and the denominator
template < typename NT >  inline
const NT&
denominator(const Rational<NT>& q)
{ return q.den_ ; }
//------------------------------------------------------------------------------------------

template < typename NT >  inline
const NT&
numerator(const Rational<NT>& q)
{ return q.num_ ; }
//------------------------------------------------------------------------------------------

namespace let 
{
  template < typename T, typename S > inline
  void assign(T & a, const Rational<S>& x)
  {
    a = x.numerator()/x.denominator();
  }
  template < typename T > inline
  void assign(Rational<T>& a, const T& x)
	{
		a = Rational<T>(x);
	}

  inline void assign(double& a, const Rational<double>& x)
  {
    a = x.numerator()/x.denominator();
  }

}

__END_NAMESPACE_SYNAPS

#endif // SYNAPS_ARITHM_RATIONAL_H

