/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*      Author(s): B. Mourrain, GALAAD, INRIA                         *
**********************************************************************
History: Initialized from the Posso file.
         Modified in order to handel different internal representation.
$Id: Scl.h,v 1.1 2005/07/11 10:39:47 mourrain Exp $
**********************************************************************/


#ifndef SYNAPS_ARITHM_SCL_H
#define SYNAPS_ARITHM_SCL_H

#ifndef NDEBUG
	#include <cassert>
#endif

//----------------------------------------------------------------------
#include <cstdio>
#include <gmp.h>
#include <synaps/init.h>
#include <synaps/arithm/REF.h>
#include <synaps/base/shared_object.h>
//----------------------------------------------------------------------


__BEGIN_NAMESPACE_SYNAPS

#define WITH_TEMP_EXP
/**
  The general interface for scalars.
**/
template<class T> struct Scl
{
public:
#ifdef WITH_TEMP_EXP
  T data;
  inline T &       rep()       {return (data);}
  inline const T & rep() const {return (data);}
#else
  shared_object<T>  data; 
  T &       rep()       {return (*data);}
  const T & rep() const {return (*data);}
#endif
  Scl ();
  Scl (const Scl<T>& b);    //{ data=b.data;}
  Scl (double d);
  Scl (signed long sl);
  Scl (unsigned long ul);
  Scl (int si);
  Scl (const char* string, unsigned int b=10);
  Scl (const T & r);
  
  template<class S>
  Scl(const VAL<S> & P) {this->init();let::assign(*this,P.rep());}

  ~Scl () {}

  bool                  operator == (const Scl<T>& rhs) const;
  bool                  operator != (const Scl<T>& rhs) const;
  bool                  operator >  (const Scl<T>& rhs) const;
  bool                  operator >= (const Scl<T>& rhs) const;
  bool                  operator <  (const Scl<T>& rhs) const;
  bool                  operator <= (const Scl<T>& rhs) const;

  bool                  operator == (long sl) const;
  bool                  operator != (long sl) const;
  bool                  operator >  (long sl) const;
  bool                  operator >= (long sl) const;
  bool                  operator <  (long sl) const;
  bool                  operator <= (long sl) const;

  bool                  operator == (int si) const;
  bool                  operator != (int si) const;
  bool                  operator >  (int si) const;
  bool                  operator >= (int si) const;
  bool                  operator <  (int si) const;
  bool                  operator <= (int si) const;

  bool                  operator == (unsigned long ul) const;
  bool                  operator != (unsigned long ul) const;
  bool                  operator >  (unsigned long ul) const;
  bool                  operator >= (unsigned long ul) const;
  bool                  operator <  (unsigned long ul) const;
  bool                  operator <= (unsigned long ul) const;

  // functions that modify at least one argument or `*this' 

  Scl<T>& operator  = (const Scl<T>& rhs);// {data=rhs.data; return *this;}
  Scl<T>& operator += (const Scl<T>& rhs);
  Scl<T>& operator -= (const Scl<T>& rhs);
  Scl<T>& operator *= (const Scl<T>& rhs);
  Scl<T>& operator /= (const Scl<T>& rhs);
  Scl<T>& operator %= (const Scl<T>& rhs);
  
  Scl<T>&               operator =  (int rhs);
  Scl<T>&               operator += (int rhs);
  Scl<T>&               operator -= (int rhs);
  Scl<T>&               operator *= (int rhs);
  Scl<T>&               operator /= (int rhs);
  Scl<T>&               operator %= (int rhs);

  Scl<T>&               operator =  (unsigned long rhs);
  Scl<T>&               operator += (unsigned long rhs);
  Scl<T>&               operator -= (unsigned long rhs);
  Scl<T>&               operator *= (unsigned long rhs);
  Scl<T>&               operator /= (unsigned long rhs);
  Scl<T>&               operator %= (unsigned long rhs);

  Scl<T>&               operator *= (double rhs);
  Scl<T>&               operator += (double rhs);

  Scl<T>&               operator =  (long rhs);
  Scl<T>&               operator += (long rhs);
  Scl<T>&               operator -= (long rhs);
  Scl<T>&               operator *= (long rhs);
  Scl<T>&               operator /= (long rhs);
//   Scl<T>&               operator %= (long rhs);

  template<class S> 
  Scl<T>& operator   = (const VAL<S> & x) 
  {
    let::assign(*this,x.rep());return *this;
  }
  template<class S> 
  Scl<T>& operator  += (const VAL<S> & x) 
    {Scl<T> tmp(x);*this+=tmp;return *this;}
  template<class S> 
  Scl<T>& operator  -= (const VAL<S> & x) 
    {Scl<T> tmp(x); *this-=tmp;return *this;}
  template<class S> Scl<T>& operator  *= (const VAL<S> & x) 
    {Scl<T> tmp(x); *this*=tmp;return *this;}
  template<class S> Scl<T>& operator  /= (const VAL<S> & x) 
    {Scl<T> tmp(x); *this/=tmp;return *this;}
  template<class S> Scl<T>& operator  %= (const VAL<S> & x) 
    {Scl<T> tmp(x); *this%=tmp;return *this;}

  //-----------------------------------------------------------------
  void                  operator ++ ( );
  void                  operator -- ( );
  //-----------------------------------------------------------------
  void                  Div2Exp   (unsigned long exponent_of_2);
  void                  GCD       (const Scl<T>& b2);
  void                  Mod2Exp   (unsigned long exponent_of_2);
  void                  Mul2Exp   (unsigned long exponent_of_2);
  void                  PowMod    (const Scl<T>& exp, const Scl<T>& m);
  void                  PowMod    (unsigned long exp, const Scl<T>& m);
  void                  abs       ( );
  void                  factorial (unsigned long n);
  Scl<T>&               negate    ( );
  Scl<T>&               pow       (unsigned long exp);
  Scl<T>&               pow       (unsigned long base, unsigned long exp);
  void                  quo       (const Scl<T>& divisor);
  void                  quo       (unsigned long divisor);
  void                  rem       (const Scl<T>& divisor);
  unsigned long         rem       (unsigned long divisor);
  void                  sqrt      ( );

private:
  void init();
};

//======================================================================
// THE ARITHMETIC OPERATORS
//======================================================================
#ifdef WITH_TEMP_EXP
template <class R> inline
VAL<OP<'+',Scl<R>,Scl<R> > > operator+(const Scl<R> & a, const Scl<R> & b)
{
  return VAL<OP<'+',Scl<R>,Scl<R> > >(OP<'+',Scl<R>,Scl<R> >(a,b));
}
//----------------------------------------------------------------------
template <class R> inline 
VAL<OP<'-',Scl<R>,Scl<R> > > operator-(const Scl<R> & a, const Scl<R> & b)
{
  return VAL<OP<'-',Scl<R>,Scl<R> > >(OP<'-',Scl<R>,Scl<R> >(a,b));
}
//----------------------------------------------------------------------
// Product of polynomials 
template <class R>  inline 
VAL<OP<'*',Scl<R>,Scl<R> > > operator*(const Scl<R> & a, const Scl<R> & b)
{
  return VAL<OP<'*',Scl<R>,Scl<R> > >(OP<'*',Scl<R>,Scl<R> >(a,b));
}
//----------------------------------------------------------------------
// Division of polynomials 
template <class R>  inline 
VAL<OP<'/',Scl<R>,Scl<R> > > operator/(const Scl<R> & a, const Scl<R> & b)
{
  return VAL<OP<'/',Scl<R>,Scl<R> > >(OP<'/',Scl<R>,Scl<R> >(a,b));
}
//----------------------------------------------------------------------
// Division of polynomials 
template <class R>  inline 
VAL<OP<'%',Scl<R>,Scl<R> > > operator%(const Scl<R> & a, const Scl<R> & b)
{
  return VAL<OP<'%',Scl<R>,Scl<R> > >(OP<'%',Scl<R>,Scl<R> >(a,b));
}
//----------------------------------------------------------------------
#endif
//======================================================================
template<class T> inline 
Scl<T> operator - (const Scl<T>& b) 
{
  Scl<T> r(b); return (r.negate());
}
//----------------------------------------------------------------------}
template<class T> Scl<T>& operator + (const Scl<T>& b, unsigned long ul);
template<class T> Scl<T>& operator - (const Scl<T>& b, unsigned long ul);
template<class T> Scl<T>& operator * (const Scl<T>& b, unsigned long ul);
template<class T> Scl<T>& operator / (const Scl<T>& b, unsigned long ul);
template<class T> Scl<T>& operator % (const Scl<T>& b, unsigned long ul);
template<class T> Scl<T>& operator + (unsigned long ul, const Scl<T>& b);
template<class T> Scl<T>& operator * (unsigned long ul, const Scl<T>& b);
template<class T> Scl<T>& operator + (const Scl<T>& b, long sl);
template<class T> Scl<T>& operator - (const Scl<T>& b, long sl);
template<class T> Scl<T>& operator * (const Scl<T>& b, long sl);
template<class T> Scl<T>& operator / (const Scl<T>& b, long sl);
template<class T> Scl<T>& operator % (const Scl<T>& b, long sl);
template<class T> Scl<T>& operator + (long sl, const Scl<T>& b);
template<class T> Scl<T>& operator * (long sl, const Scl<T>& b);
template<class T> Scl<T>& operator * (double d, const Scl<T>& b);
//----------------------------------------------------------------------
template<class T> Scl<T>& Div2Exp (const Scl<T>& b, 
				   unsigned long exponent_of_2);
template<class T> Scl<T>& GCD     (const Scl<T>& b1, const Scl<T>& b2);
template<class T> Scl<T>& gcd     (const Scl<T>& b1, const Scl<T>& b2);
template<class T> Scl<T>& Mod2Exp (const Scl<T>& b, 
				   unsigned long exponent_of_2);
template<class T> Scl<T>& Mul2Exp (const Scl<T>& b, unsigned long exponent_of_2);
template<class T> Scl<T>& PowMod  (const Scl<T>& b, const Scl<T>& exp, const Scl<T>& m);
template<class T> Scl<T>& PowMod  (const Scl<T>& b, unsigned long exp, 
				   const Scl<T>& m);
template<class T> Scl<T> abs     (const Scl<T>& b);
// template<class T> Scl<T>& negate  (const Scl<T>& b);
template<class T> Scl<T> pow     (const Scl<T>& base, unsigned long exp);
template<class T> Scl<T>& quo     (const Scl<T>& dividend, 
				   const Scl<T>& divisor);
template<class T> Scl<T>& quo     (const Scl<T>& dividend, 
				   unsigned long divisor);
template<class T> Scl<T>& rem     (const Scl<T>& dividend, 
				   const Scl<T>& divisor);
template<class T> Scl<T> sqrt    (const Scl<T>& b);
template<class T> unsigned long rem (const Scl<T>& b, unsigned long divisor);

template<class T> Scl<T>& BigIntFactorial (unsigned long n);

//
// -------- the corresponding `const' functions -----------------------------
//

#define NO_NRV 1
#if defined(__GNUG__) && !defined(NO_NRV)
#define NRVAL(name) return name
#define NRCODE(code) code
#define NONRCODE(code) 
#else
#define NRVAL(name) 
#define NRCODE(code)
#define NONRCODE(code) code
#endif

#define OPCODE(a, b, op) Scl<T>& r =*new Scl<T>(a); return (r op b);


// template<class T> inline 
// Scl<T>& operator + (const Scl<T>& b1, const Scl<T>& b2) 
// {
//   OPCODE(b1,b2,+=)
// }

// template<class T> inline 
// Scl<T>& operator - (const Scl<T>& b1, const Scl<T>& b2) 
// {
//   OPCODE(b1,b2,-=)
// }

// template<class T> inline 
// Scl<T>& operator * (const Scl<T>& b1, const Scl<T>& b2) 
// {
//   OPCODE(b1,b2,*=)
// }

// template<class T> inline 
// Scl<T>& operator / (const Scl<T>& b1, const Scl<T>& b2) 
// {
//   OPCODE(b1,b2,/=)
// }

// template<class T> inline 
// Scl<T>& operator % (const Scl<T>& b1, const Scl<T>& b2) 
// {
//   OPCODE(b1,b2,%=)
// }

//--------------------------
template<class T> inline 
Scl<T>& operator + (const Scl<T>& b, int ul) 
{
  OPCODE(b,ul,+=)
}
template<class T> inline 
Scl<T>& operator + (const Scl<T>& b, unsigned long ul) 
{
  OPCODE(b,ul,+=)
}

template<class T> inline 
Scl<T>& operator - (const Scl<T>& b, unsigned long ul) 
{
  OPCODE(b,ul,-=)
}

template<class T> inline Scl<T>&
operator * (const Scl<T>& b, unsigned long ul) 
{
  OPCODE(b,ul,*=)
}

template<class T> inline Scl<T>&
operator / (const Scl<T>& b, unsigned long ul) 
{
  OPCODE(b,ul,/=)
}

template<class T> inline 
Scl<T>& operator % (const Scl<T>& b, unsigned long ul) 
{
  OPCODE(b,ul,%=)
}

//----------------------------------------------------------------------
template<class T> inline 
Scl<T>& operator + (unsigned long ul, const Scl<T>& b)
{
  return b + ul;
}
//----------------------------------------------------------------------
template<class T> inline 
Scl<T>& operator * (unsigned long ul, const Scl<T>& b)
{
  return b * ul;
}
//----------------------------------------------------------------------
template<class T> inline 
Scl<T>& operator + (const Scl<T>& b, long sl)  
{
  OPCODE(b,sl,+=)
}


template<class T> inline 
Scl<T>& operator - (const Scl<T>& b, long sl)  NRVAL(r(b))
{
  OPCODE(b,sl,-=)
}

template<class T> inline 
Scl<T>& operator * (const Scl<T>& b, int sl)  NRVAL(r(b))
{
  OPCODE(b,sl,*=)
}

template<class T> inline 
Scl<T>& operator * (const Scl<T>& b, long sl)  NRVAL(r(b))
{
  OPCODE(b,sl,*=)
}

template<class T> inline
Scl<T>& operator / (const Scl<T>& b, long sl)  NRVAL(r(b))
{
  OPCODE(b,sl,/=)
}

template<class T> inline Scl<T>&
operator % (const Scl<T>& b, long sl)  NRVAL(r(b))
{
  OPCODE(b,sl,%=)
}

//--------------------------

template<class T> inline 
Scl<T>& operator + (long sl, const Scl<T>& b)
{
  return b + sl;
}

template<class T> inline 
Scl<T>& operator * (long sl, const Scl<T>& b)
{
  return b * sl;
}
//--------------------------

template<class T> inline 
Scl<T>& Div2Exp (const Scl<T>& b, unsigned long exponent_of_2) NRVAL(r(b))
{
  NRCODE(r.Div2Exp(exponent_of_2);)
  NONRCODE(Scl<T> r(b); r.Div2Exp(exponent_of_2); return r;)
}

template<class T> inline 
Scl<T>& GCD (const Scl<T>& b1, const Scl<T>& b2) NRVAL(r(b1))
{
  NRCODE(r.GCD(b2);)
  NONRCODE(Scl<T> r(b1); r.GCD(b2); return r;)
}

template<class T> inline 
Scl<T>& gcd (const Scl<T>& b1, const Scl<T>& b2) NRVAL(r(b1))
{
  NRCODE(r.GCD(b2);)
  NONRCODE(Scl<T> r(b1); r.GCD(b2); return r;)
}

template<class T> inline 
Scl<T>& Mod2Exp (const Scl<T>& b, unsigned long exponent_of_2) NRVAL(r(b))
{
  NRCODE(r.Mod2Exp(exponent_of_2);)
  NONRCODE(Scl<T> r(b); r.Mod2Exp(exponent_of_2); return r;)
}

template<class T> inline 
Scl<T>& Mul2Exp (const Scl<T>& b, unsigned long exponent_of_2) NRVAL(r(b))
{
  NRCODE(r.Mul2Exp(exponent_of_2);)
  NONRCODE(Scl<T> r(b); r.Mul2Exp(exponent_of_2); return r;)
}

template<class T> inline 
Scl<T>& PowMod  (const Scl<T>& b, const Scl<T>& exp, const Scl<T>& m) NRVAL(r(b))
{
  NRCODE(r.PowMod(exp, m);)
  NONRCODE(Scl<T> r(b); r.PowMod(exp,m); return r;)
}

template<class T> inline 
Scl<T>& PowMod  (const Scl<T>& b, unsigned long exp, const Scl<T>& m) NRVAL(r(b))
{
  NRCODE(r.PowMod(exp, m);)
  NONRCODE(Scl<T> r(b); r.PowMod(exp,m); return r;)
}

//template<class T> inline Scl<T>
template <class T> inline 
Scl<T> abs (const Scl<T>& b) NRVAL(r(b))
{
  NRCODE(r.abs();)
  NONRCODE(Scl<T> r(b); r.abs(); return r;)
}

// template<class T> inline Scl<T>&
// negate (const Scl<T>& b) NRVAL(r(b))
// {
//   NRCODE(r.negate();)
//   NONRCODE(Scl<T> r(b); r.negate(); return r;)  
// }

template<class T> inline 
Scl<T> pow (const Scl<T>& base, unsigned long exp) NRVAL(r(base))
{
  NRCODE(r.pow(exp);)
  NONRCODE(Scl<T> r(base); r.pow(exp); return r;)
}

template<class T> inline 
Scl<T>& quo (const Scl<T>& dividend, const Scl<T>& divisor) NRVAL(r(dividend))
{
  NRCODE(r.quo(divisor);)
  NONRCODE(Scl<T> r(dividend); r.quo(divisor); return r;)
}

template<class T> inline 
Scl<T>& quo (const Scl<T>& dividend, unsigned long divisor) NRVAL(r(dividend))
{
  NRCODE(r.quo(divisor);)
  NONRCODE(Scl<T> r(dividend); r.quo(divisor); return r;)
}

template<class T> inline 
Scl<T>& rem (const Scl<T>& dividend, const Scl<T>& divisor) NRVAL(r(dividend))
{
  NRCODE(r.rem(divisor);)
  NONRCODE(Scl<T> r(dividend); r.rem(divisor); return r;)
}

template<class T> inline 
Scl<T> sqrt (const Scl<T>& b) NRVAL(r(b))
{
  NRCODE(r.sqrt();)
  NONRCODE(Scl<T> r(b); r.sqrt(); return r;)
}

template<class T> inline 
unsigned long rem (const Scl<T>& b, unsigned long divisor)
{
  Scl<T> r(b);
  return r.rem(divisor);
}

template<class T> inline 
Scl<T>& BigIntFactorial (unsigned long n) NRVAL(r)
{
  NRCODE(r.factorial(n);)
  NONRCODE(Scl<T> r; r.factorial(n); return r;)    
}

template<class T> inline 
Scl<T>& BigIntPow (unsigned long base, unsigned long exp) NRVAL(r)
{
  NRCODE(r.pow(base, exp);)
  NONRCODE(Scl<T> r; r.pow(base, exp); return r;)
}

template<class T>
 Scl<T> Size(const Scl<T> & r)
{
  return abs(r);
}

#undef NRVAL
#undef NRCODE
#undef NONRCODE


__END_NAMESPACE_SYNAPS

#endif // SYNAPS_ARITHM_SCL_H


