/********************************************************************
 *   This file is part of the source code of the SYNAPS library.
 *   Author(s): B. Mourrain, GALAAD, INRIA
 ********************************************************************/
#ifndef SYNAPS_ARITHM_REAL_NUMBER_H
#define SYNAPS_ARITHM_REAL_NUMBER_H
/********************************************************************/
#include <synaps/init.h>
#include <synaps/arithm/symbolic/Exp.h>
#include <synaps/arithm/Interval.h>
/********************************************************************/
__BEGIN_NAMESPACE_SYNAPS

namespace symbolic {

  template<class C>
  struct real_rep: 
    public val_rep<Interval<C> >
  {
    typedef val_rep<Interval<C> > base_t;

    real_rep(): base_t(Interval<C>(0,0)) {}
    real_rep(const C& a, const C& b): base_t(Interval<C>(a,b)) {}
    real_rep(Interval<C> u): base_t(u) {}
    virtual ~real_rep<C>() {};

    virtual C  lower() const {return this->rep_.lower();}
    virtual C& lower()       {return this->rep_.lower();}
    virtual C  upper() const {return this->rep_.upper();}
    virtual C& upper()       {return this->rep_.upper();}

    virtual C  prec()        {return (upper()-lower()); }

    Interval<C>  num_val() const {return this->rep_;}
    Interval<C>& num_val()       {return this->rep_;}

    virtual unsigned improve () { return 0; };
    virtual int      compare (real_rep<C>* a);

    virtual void  print(std::ostream & os) const
    {
      os<<num_val();
    }
  };
  //--------------------------------------------------------------------
  /** Generic class for real numbers. It points
      to the internal class @real_rep<C>@. The @lower@ and
      @upper@ functions return respectively a lower bound and upper
      bound of the underlying algebraic number.
  **/
  template<class C> struct real
  {
    //protected:
    real_rep<C>* rep_;
    const static C  eps;
  public:
    real<C>() : rep_(NULL) {}
    real(real_rep<C>* ptr) : rep_(ptr) {}
    real(int x) : rep_(new real_rep<C>(C(x),C(x))) {}
    real(const C& x) : rep_(new real_rep<C>(x,x)) {}
    real(const C& x, const C& y): 
      rep_(new real_rep<C>(x,y)) {}

    real_rep<C>* operator -> () {return rep_;}
    real_rep<C>* rep         () {return rep_;}

    Interval<C>    num_val () const    {return rep_->num_val();}
    Interval<C>&   num_val          () {return rep_->num_val();}
    
    unsigned improve ()               { return rep_->improve(); }
    int compare (real<C> a)    { return rep_->compare(a.rep()); }
    bool operator==(real<C> b) { return  rep_->compare(b.rep())==0;}
    
    virtual C  lower() const {return rep_->lower();}
    virtual C& lower()       {return rep_->lower();}
    virtual C  upper() const {return rep_->upper();}
    virtual C& upper()       {return rep_->upper();}

    C prec() { return rep_->prec(); }
    void print(std::ostream& out) const { rep_->print(out); }
  };
  //--------------------------------------------------------------------
  template<class C> const C real<C>::eps =1e-6;
  //--------------------------------------------------------------------
  template<class C>
  std::ostream& operator << (std::ostream& os, real<C> a)
  {
    a->print(os); return os;
  }
  //--------------------------------------------------------------------
  /** Comparison of two algebraic numbers. If the bounding intervals are
      separated, the answer is output. If not, the bounding intervals are
      improved and the test is applied again, until a fixed precision
      @real<C>::eps@ is reached. The output is -1 if b<a, 1 is
      a<b (certified) and 0 (equality within precision @eps@).
  **/
  template<class C>
  int real_rep<C>::compare(real_rep<C>* b)
  {
    if(this->upper() < b->lower()) return -1;
    else if (b->upper() < this->lower()) return 1;
    else if ( (this->prec())< real<C>::eps &&
	      (b->prec())< real<C>::eps)
      return 1000;
    else
      {
	int i1=this->improve(), i2=b->improve();
	if(i1 || i2)
	  return (this->compare(b));
	else
	  return 1000;
      }
  }
  //--------------------------------------------------------------------
  template<class Coef>
  int compare(real<Coef>  a, real<Coef> b)
  {
    return a.compare(b);
  }
  //--------------------------------------------------------------------
  /** Comparison of an real number with a classical number.
   **/
  template<class R>
  int compare(real<R> a, const R & b)
  {
    if(a.upper() < b) return -1;
    else if (b < a.lower()) return 1;
    else if ( (a.prec())< real<R>::eps)    return 0;
    else
      {
	if(a.improve())
	  return compare(a,b);
	else return 0;
      }
  } 
  //--------------------------------------------------------------------
  template<class C>
  struct op_rep<'+',Interval<C> > : public real_rep<C>
  {
  protected:
    real<C> op1, op2;
  public:
    friend class op<'+',C>;
    
    op_rep(real<C> a, real<C> b)
      : real_rep<C>((Interval<C>)(a.num_val()+b.num_val())), op1(a), op2(b) {}
    
    unsigned improve()
    {
      unsigned r = std::max(op1.improve(),op2.improve());
      this->num_val() = op1.num_val()+op2.num_val();
      return r;
    }
    void print(std::ostream& os) const {
      os<<"(+ "; op1.print(os); os<<" "; op2.print(os); os<<")";
    }
  };
  //--------------------------------------------------------------------
  template<class C>
  struct op<'+',Interval<C> >
  {
  protected:
    op_rep<'+',Interval<C> >* rep_;
  public:
    op(const real<C> & a, const real<C>& b): 
      rep_(new op_rep<'+',Interval<C> >(a,b)) {}
    inline operator real<C> () { return real<C>(rep_); }
  };
  //--------------------------------------------------------------------
  /** Addition of two numbers. It returns an unevaluated tree
      pointing to the two algebraic numbers <t>a</t> and <t>b</t> and coerced
      to an algebraic number. 
  **/
  template<class C>
  real<C> operator+ (real<C> a, real<C> b)
  {
    return op<'+',Interval<C> >(a,b);
  }
  //======================================================================
  template<class C>
  struct op_rep<'-',Interval<C> > : public real_rep<C>
  {
  protected:
    real<C> op1, op2;
  public:
    friend class op<'-', Interval<C> >;
    
    op_rep(real<C> a, real<C> b)
      : real_rep<C>((Interval<C>)(a.num_val()-b.num_val())), op1(a), op2(b) {}
    
    unsigned improve()
    {
      unsigned r = std::max(op1.improve(),op2.improve());
      this->num_val() = op1.num_val()-op2.num_val();
      return r;
    }
    void print(std::ostream& os) const {
      os<<"(- "; op1.print(os); os<<" "; op2.print(os); os<<")";
    }
  };
  //--------------------------------------------------------------------
  template<class C>
  struct op<'-',Interval<C> >
  {
  protected:
    op_rep<'-',Interval<C> >* rep_;
  public:
    op(const real<C> & a, const real<C>& b) : rep_(new op_rep<'-',Interval<C> >(a,b)) {}
    inline operator real<C> () { return real<C>(rep_); }
  };
  //--------------------------------------------------------------------
  /** Substraction of two algebraic numbers. It returns an unevaluated tree
      pointing to the two algebraic numbers <t>a</t> and <t>b</t> and coerced
      to an algebraic number. 
  **/
  template<class C>
  real<C> operator- (real<C> a, real<C> b)
  {
    return op<'-',Interval<C> >(a,b);
  }
  //======================================================================
  template<class C>
  struct op_rep<'*',Interval<C> > : public real_rep<C>
  {
  protected:
    real<C> op1, op2;
  public:
    friend class op<'*',Interval<C> >;
    
    op_rep(real<C> a, real<C> b)
      : real_rep<C>((Interval<C>)(a.num_val()*b.num_val())), op1(a), op2(b) {}
    
    unsigned improve()
    {
      unsigned r = std::max(op1.improve(),op2.improve());
      this->num_val() = op1.num_val()*op2.num_val();
      return r;
    }
    void print(std::ostream& os) const {
      os<<"(* "; op1.print(os); os<<" "; op2.print(os); os<<")";
    }
  };
  //--------------------------------------------------------------------
  template<class C>
  struct op<'*',Interval<C> >
  {
  protected:
    op_rep<'*',Interval<C> >* rep_;
  public:
    op<'*',Interval<C> >(const real<C> & a, const real<C>& b) : rep_(new op_rep<'*',Interval<C> >(a,b)) {}
    inline operator real<C> () { return real<C>(rep_); }
  };
  //--------------------------------------------------------------------
  /** Multiplication of two numbers. It returns an unevaluated tree
      pointing to the two algebraic numbers <t>a</t> and <t>b</t> and coerced
      to an algebraic number. 
  **/
  template<class C>
  real<C> operator* (real<C> a, real<C> b)
  {
    return op<'*',Interval<C> >(a,b);
  }
    
  //======================================================================
  template<class C>
  struct op_rep<'/',Interval<C> > : public real_rep<C>
  {
  protected:
    real<C> op1, op2;
  public:
    friend class op<'/',Interval<C> >;
    
    op_rep(real<C> a, real<C> b)
      : real_rep<C>((Interval<C>)(a.num_val()/b.num_val())), op1(a), op2(b) {}
    
    unsigned improve()
    {
      unsigned r = std::max(op1.improve(),op2.improve());
      this->num_val() = op1.num_val()/op2.num_val();
      return r;
    }
    void print(std::ostream& os) const {
      os<<"(/ "; op1.print(os); os<<" "; op2.print(os); os<<")";
    }
  };
  //--------------------------------------------------------------------
  template<class C>
  struct op<'/',Interval<C> >
  {
  protected:
    op_rep<'/',Interval<C> >* rep_;
  public:
    op(const real<C> & a, const real<C>& b) : rep_(new op_rep<'/',Interval<C> >(a,b)) {}
    inline operator real<C> () { return real<C>(rep_); }
  };
  //--------------------------------------------------------------------
  /** Division of two numbers. It returns an unevaluated tree
      pointing to the two numbers <t>a</t> and <t>b</t> and coerced
      to a number. 
  **/
  template<class C>
  real<C> operator/ (real<C> a, real<C> b)
  {
    return op<'/',Interval<C> >(a,b);
  }
  //--------------------------------------------------------------------
} //generic
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
/********************************************************************/
#endif // SYNAPS_ARITHM_REAL_NUMBER_H

