// Copyright (c) 2004  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)


#ifndef SYNAPS_UTIL_RAND_H
#define SYNAPS_UTIL_RAND_H

#include <cstdlib>
#include <limits>
#include <ctime>
#include <synaps/init.h>
#include <synaps/arithm/Abs.h>
#include <synaps/arithm/gmp.h>

__BEGIN_NAMESPACE_SYNAPS


/** Generator of pseudo-random numbers between @m@ and @M@ of type @C@.
 * Use: @Rand<C> h(min,max); h();@
 *
 *      or @h.get_random_bool()@ in torder to get a bool
 **/
template<class C = double>
class Rand
{
private:
  C                  m_;
  C                  M_;
  unsigned long int  seed_;

  void init()
  { std::srand(seed_); }

public:
  typedef C value_type;

  Rand()
    : m_(C(-100000)),
      M_(C(100000)),
      seed_( (unsigned long )std::time(NULL) )
  { init(); }

  Rand(const C& M)
    : m_(0), M_(M), seed_( (unsigned long)std::time(NULL) )
  { init(); }

  Rand(const C& m, const C& M)
    : m_(m), M_(M), seed_( (unsigned long)std::time(NULL) )
  { init(); }

  Rand(const C& m, const C& M, const unsigned long int& seed)
    : m_(m), M_(M), seed_(seed)
  { init(); }

  bool get_rand_bool()
  {
    int r = int( ( double(std::rand())/RAND_MAX ) * 10 );
    return (r < 5 ? false : true);
  }

  C operator()(void)
  {
    C U;
    if (m_ < 0)
      U = M_ + (m_ < 0 ? -m_ : m_) + 1;
    else
      U = M_ - m_ + 1;
    C r = C( ( double(std::rand())/RAND_MAX ) * U );
    return (r + m_);
  }

  C operator()(const C& m, const C& M)
  {
    C U;
    if (m < 0)
      U = M + (m_ < 0 ? -m_ : m_) + 1;
    else
      U = M - m + 1;
    C r = C( ( double(std::rand())/RAND_MAX ) * U );
    return (r + m);
  }

};

#ifdef SYNAPS_WITH_GMP

/** Generator of pseudo-random numbers between @m@ and @M@ of type @C@.
 * Use: @Rand<C> h(min,max); h();@
 *
 *      or @h.get_random_bool()@ in torder to get a bool
 *
 * Specialization for the ZZ (gmp) class
 **/
template<>
class Rand<ZZ>
{
private:
  ZZ                 m_;
  ZZ                 M_;
  gmp_randstate_t    rstate_;
  unsigned long int  seed_;

  void init()
  {
    gmp_randinit (rstate_, GMP_RAND_ALG_LC, 32);
    gmp_randseed_ui(rstate_, seed_);
  }

public:
  typedef ZZ value_type;

  Rand()
    : m_(std::numeric_limits<long>::min()),
      M_(std::numeric_limits<long>::max()),
      seed_( (unsigned)std::time(NULL) )
  { init(); }

  Rand(const ZZ& M)
    : m_(0), M_(M), seed_( (unsigned)std::time(NULL) )
  { init(); }

  Rand(const ZZ& m, const ZZ& M)
    : m_(m), M_(M), seed_( (unsigned)std::time(NULL) )
  { init(); }

  Rand(const ZZ& m, const ZZ& M, const unsigned long int& seed)
    : m_(m), M_(M), seed_(seed)
  { init(); }

  ZZ operator()(void)
  {
    ZZ U;
    if (m_ < 0)
      U = M_ + abs(m_) + 1;
    else
      U = M_ - m_ + 1;
    ZZ r;
    mpz_urandomm(r.get_mpz_t(), rstate_, U.get_mpz_t());
    return (r + m_);
  }

  ZZ operator()(const ZZ& m, const ZZ& M)
  {
    ZZ U;
    if (m < 0)
      U = M + abs(m) + 1;
    else
      U = M - m + 1;
    ZZ r;
    mpz_urandomm(r.get_mpz_t(), rstate_, U.get_mpz_t());
    return (r + m);
  }
  

  bool get_rand_bool()
  {
    ZZ r;
    ZZ t(2);
    mpz_urandomm(r.get_mpz_t(), rstate_, t.get_mpz_t());
    return (r == 0 ? false : true);
  }

  ~Rand()
  { gmp_randclear(rstate_); }
};

#endif //SYNAPS_WITH_GMP


__END_NAMESPACE_SYNAPS

#endif // SYNAPS_UTIL_RAND_H

