/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************
History: 
$Id: Decomp.h,v 1.1 2005/07/11 08:33:33 mourrain Exp $
**********************************************************************/
#ifndef synaps_linalg_Decomp_H
#define synaps_linalg_Decomp_H
//--------------------------------------------------------------------
#include <synaps/init.h>
#include <synaps/linalg/MethodName.h>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS

/** \file Decomp.h
    \brief Different decompositions of matrices.
**/
/** The function is performing a decomposition of the matrix into 
components. The parameter argument type @A@ is specifying the type of 
decomposition to be computed. It can be @LU@, @QR@, @Bareiss@, \dots 
Here is an exemple of how to calf the function, which compute a
LU-decomposition of a matrix @m@: 
\begin{xxx}
   Decomp<LU>(m);
\end{xxx}
The function @MATRIX::decomp(LU(),m.rep())@ will be called.
The specification of this function may depend on the container
type of @m.rep()@. In most of the cases, the output matrix will
be triangular.
 **/
template<class A, class M> inline
M Decomp(const M & mat)
{
   M x(mat);
   using MATRIX::decomp; decomp(A(),x.rep());
   return (x);
}
//--------------------------------------------------------------------------
/// Triang is doing the same thing as @Decomp@.
template<class A, class M> inline
M Triang(const M & mat)
{
  M x(mat);
  using MATRIX::decomp; decomp(A(),x.rep());
  return (x);
}
//----------------------------------------------------------------------
/** Apply the decomposition and return the second factor in @L@.
For instance, in a LU-decomposition, it will put in L the lower matrix
factor; in a QR-decomposition it will be Q, \dots It uses the function 
@decomp(A(),mat.rep(),L.rep())@.  
**/
template<class A, class M1, class M2> inline 
M1 Decomp(M1 & mat, M2 & L)
{
  M1 x(mat);
  using MATRIX::decomp; decomp(A(),x.rep(),L.rep());
  return (x);
}
//----------------------------------------------------------------------
/** Apply the decomposition and return the second factor in @L1@ and the
third factor if it exist in @L2@. For instance, in a LU-decomposition, 
the fisrt factor will the lower matrix L and the third factor will the 
pivots indices of the columns. It uses the function 
@decomp(A(),mat.rep(),L.rep(),S.rep())@. 
**/
template<class A, class M1, class M2, class M3> inline 
M1 Decomp(M1 & mat, M2 & L, M3 & S)
{
   M1 x(mat);
   using MATRIX::decomp; decomp(A(),x.rep(),L.rep(),S.rep());
   return (x);
}
//----------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
#endif // synaps_linalg_Decomp_H

