/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************
$Id: MatrSps.h,v 1.1 2005/07/11 08:33:34 mourrain Exp $
**********************************************************************/
#ifndef SYNAPS_LINALG_MATRSPS_H
#define SYNAPS_LINALG_MATRSPS_H
//---------------------------------------------------------------------
#include <synaps/init.h>
#include <synaps/arithm/GENERIC.h>
#include <synaps/linalg/MethodName.h>
#include <synaps/linalg/sparse.h>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS

//---------------------------------------------------------------------
/**
Sparse matrices. These are  matrices where only the non-zero coefficients
are stored. Scalar multiplication (and division) as well as multiplication by 
a vector are available. Addition and substraction of sparse matrices
are not provided. Row and column permutations are allowed.
**/
template<class C, class R=sparse::rep2d<C> > struct MatrSps
{
  shared_object<R>  data;
  R &       rep()       {return (*data);}
  const R & rep() const {return (*data);}

  typedef typename R::size_type  size_type;
  typedef C                      value_type;
  typedef MatrSps<C,R>           self_t;

  MatrSps() {}
  MatrSps(const R  & r): data(r)              {}
  MatrSps(int m, int n): data(R(m,n)) {}
  MatrSps(int m, int n, int nz) : data(R(n,m,nz)) {}
  MatrSps(int m, int n, int nz, value_type* t, size_type* idx): 
    data(R(m,n,nz,t,idx)) {}

  MatrSps(const char * str): data(R(str))         {}
  MatrSps(const self_t & M): data(M.data)       {}
  template<class T> 
  MatrSps(const VAL<T> & M) 
  {
    using let::assign; assign(*this,M.rep());
  } 

  self_t &  operator=(const self_t & M) {data=M.data; return *this;} 
  template<class T>
  self_t & operator=(const VAL<T> & M) 
  {
    using let::assign;assign(*this,M.rep()); return *this;
  } 

  self_t & transpose () {rep().transpose(); return *this;}

  size_type nbrow() const {return rep().nbrow();}
  size_type nbcol() const {return rep().nbcol();}
  size_type nzero() const {return rep().nzero();}

  value_type operator() (size_type i, size_type j) const {return rep()(i,j);} 

  self_t & operator*=(const value_type & c);
  self_t & operator/=(const value_type & c);

  self_t & swaprow(size_type i, size_type j);
  self_t & swapcol(size_type, size_type);
  self_t & permrow(size_type *pivot);
  self_t & permcol(size_type *pivot);

};
//---------------------------------------------------------------------------
/**
   Write a Sparse Matrix @M@ in Harwell-Boeing Format in the file 
   @file_to_write@.
 **/
template<class C,class R>
inline void WriteToHSL(const MatrSps<C,R> & M, const char *file_to_write)
{
  writetoHSL(M.rep(), file_to_write);
}
//---------------------------------------------------------------------------
/**
   read a Sparse Matrix @M@ in Harwell-Boeing Format from the file 
   @file_to_write@.
 **/
template<class C,class R>
inline void readFromHSL(MatrSps<C,R> & M, const char *file_to_write)
{
  readfromHSL(M.rep(), file_to_write);
}
//======================================================================
// INPUT/OUTPUT 
//======================================================================
/// Output operator for standard matrices.
template<class C,class R>
std::ostream & operator<<(std::ostream & os, const  MatrSps<C,R> & m)
{
  return print(os,m.rep());
}


__END_NAMESPACE_SYNAPS


#ifndef SEPARATE_COMPILATION
#include "synaps/linalg/MatrSps.C"
#endif
//----------------------------------------------------------------------

#endif // SYNAPS_LINALG_MATRSPS_H

