#ifndef SYNAPS_MPOLDSE_EENVOPS_H
#define SYNAPS_MPOLDSE_EENVOPS_H

#include <synaps/init.h>
#include <synaps/mpol/mpoldse/vctops.h>
#include <synaps/mpol/mpoldse/binomials.h>

__BEGIN_NAMESPACE_SYNAPS

namespace mpoldse
{
  /** Evaluation by horner scheme of multivariate polynomial data.
   *  @result@ is the result of the evaluation of the coefficients in dense monomial form starting at @data@ and corresponding 
   *  to the sizes and ordering of variables represented by @env@. 
   *  @p@ is the parameter array (operator[])supposed to be as large as the number of variables in @env@.
   *  @result@ is assumed to be assignable to the Type::Sup value of @P@ and @R@.
   *  Note that @levalm()@ index of variables are local to @env@, it's the difference with gevalm.
   **/
  template < class R, class C, class eenv, class P >
  void levalm (R & result, const C * data, const eenv & env, const P & p)
  {
    typedef typename Type::Sup < typename Type::ValueType < P >::result_t,
      C >::result_t result_type;

    // query for the environment information

    const int esz = env.sz ();	// number of coefficients
    const int nvr = env.nvr ();	// number of variables 
    const int *szs = env.szs ();	// sizes in each variables
    const int *str = env.str ();	// corresponding strides.

    result_type tmp[env.sz ()];	// temporary array used by the inplace evaluation loop
    std::copy (data, data + env.sz (), tmp);

    for (int v = nvr - 1; v >= 0; v--)
      for (int i = 0; i < esz; i += str[v - 1])
	(tmp + i)[0] = vct::horner(tmp + i, szs[v], p[v], str[v]);

    result = tmp[0];
  };

  template<typename real_t>
  void decasteljau( real_t * r, unsigned sz, const real_t& t, int str = 1 )
  {
    real_t *er, *p;
      for ( er = r + (sz-1)*str; er != r; er -= str )
	for ( p = r; p != er; p += str )
	  *p = (1.0-t)**p+t**(p+str);
  };  

  /** Evaluation in bernstein form as @levalm()@ **/
  template < class R, class C, class eenv, class P >
  void levalb (R & result, const C * data, const eenv & env, const P & p)
  {
    typename Type::Sup < typename Type::ValueType < P >::result_t,
      C >::result_t tmp[env.sz ()];
    
    const int esz  (env.sz ());
    const int nvr  (env.nvr ());
    const int *szs (env.szs ());
    const int *str (env.str ());

    std::copy (data, data + env.sz (), tmp);

    for (int v = nvr - 1; v >= 0; v--)
      for (int i = 0; i < esz; i += str[v - 1])
	decasteljau (tmp + i, szs[v], p[v], str[v]);

    result = tmp[0];
  };

  

  template < class OutputIterator, class C, class eenv  >
  void maxs (OutputIterator _maxs_, const C * data, const eenv & env, int v)
  {
    const int *eszs = env.szs ();
    const int *estr = env.str ();
    const int envr = env.nvr ();
    int p, k, i, j;
    for (p = 0, k = 0; k < eszs[v]; k++, data += estr[v], _maxs_++)
      {
	*_maxs_ = *data;
	for (i = 0; i < estr[-1]; i += estr[v - 1])
	  for (C * edata = data + i; edata < data + i + estr[v]; edata++)
	    if (*edata > *_maxs_)
	      *_maxs_ = *edata;
      };
  };
  
  template < class OutputIterator, class C, class eenv >
  void mins (OutputIterator _mins_, const C * data, const eenv & env, int v)
  {
    const int *eszs = env.szs ();
    const int *estr = env.str ();
    const int envr = env.nvr ();
    int p, k, i, j;
    for (p = 0, k = 0; k < eszs[v]; k++, p += estr[v], _mins_++)
      {
	*_mins_ = data[p];
	for (i = 0; i < estr[-1]; i += estr[v - 1])
	  for (j = i; j < i + estr[v]; j++)
	    if (data[j + p] < *_mins_)
	      *_mins_ = data[j + p];
      };
  };

  

  template < class C, class eenv >
  void vswap( C * data, const eenv & env, int *perm, int *supp = 0, int nsupp = 0  )
  {
    const int *szs = env.szs ();
    const int *str = env.str ();
    int ostr[env.nvr ()];
    int i;
    for (i = env.nvr () - 2, str[env.nvr () - 1] = 1; i >= -1;
	 ostr[i] = szs[perm[i + 1]] * str[i + 1], i--);
    if (supp == 0)
      {
	int a, o, v;
	for (i = 0; i < env.esz (); i++)
	  {
	    for (a = i, v = env.nvr () - 1; a; a /= szs[v], v--)
	      o += (a % szs[v]) * ostr[perm[v]];
	    std::swap (data[i], data[o]);
	  };
      }
    else
      {
	int a, o, v;
	for (i = 0; i < nsupp; supp[i++] = o)
	  {
	    for (a = supp[i], v = env.nvr () - 1; a; a /= szs[v], v--)
	      o += (a % szs[v]) * ostr[perm[v]];
	    std::swap (data[i], data[o]);
	  };
      };
    env.vswap (perm);
  };
  

  template < class C, class eenv  >
  void binoms (C * data, const eenv & env, binomials < C > & binms)
  {
    const int *estr = env.str ();
    const int *eszs = env.szs ();
    const int envr = env.nvr ();
    const int esz = env.sz ();
    const C *bins = binms[eszs[envr - 1] - 1];
    for (C * pdata = data; pdata < data + esz; pdata += estr[envr - 2])
      vct::copy (pdata, bins, eszs[envr - 1]);
    for (int v = 0; v < envr - 1; v++)
      {
	bins = binms[eszs[v] - 1];
	int k = 0;
	for (int i = 0; i < esz; i += estr[v - 1])
	  for (int j = 0; j < estr[v]; j++)
	    vct::pmul (data + i + j, bins, eszs[v], estr[v], 1);
      };
  };
  
  template < class C >
  void ibinoms (C * data, const eenv & env, binomials < C > & binms)
  {
    binoms(data,env,binms);
    vct::inverses(data,data+env.sz());
  };
  
  /** Convert a multivariate polynomial data in the bernstein basis to the scaled bernstein basis. **/ 
  template < class C, class eenv >
  void scale (C * data, const eenv & env, binomials < C > &binoms)
  {
    const int *estr = env.str ();
    const int *eszs = env.szs ();
    const int  envr = env.nvr ();
    const int  esz  = env.sz ();
    const C *  bins = binoms[eszs[envr - 1] - 1];
    
    for (C *  pdata = data; pdata < data + esz; vct::pmul (pdata, bins, eszs[envr - 1]), pdata += estr[envr - 2]);
    
    for (int v = 0; v < envr - 1; v++)
      {
	bins = binoms[eszs[v] - 1];
	int k = 0;
	for (int i = 0; i < esz; i += estr[v - 1])
	  for (int j = 0; j < estr[v]; j++)
	    vct::pmul (data + i + j, bins, eszs[v], estr[v], 1);
      };
  };
  
  /** Convert a multivariate polynomial data in the scaled bernstein basis to the bernstein basis **/
  template < class C, class eenv >
  void uscale (C * data, const eenv & env, binomials < C > &binoms)
  {
    const int *estr = env.str ();
    const int *eszs = env.szs ();
    const int envr = env.nvr ();
    const int esz = env.sz ();
    const C *bins = binoms[eszs[envr - 1] - 1];
    for (C * pdata = data; pdata < data + esz; pdata += estr[envr - 2])
      vct::pdiv (pdata, bins, eszs[envr - 1]);
    for (int v = 0; v < envr - 1; v++)
      {
	bins = binoms[eszs[v] - 1];
	int k = 0;
	for (int i = 0; i < esz; i += estr[v - 1])
	  for (int j = 0; j < estr[v]; j++)
	    vct::pdiv (data + i + j, bins, eszs[v], estr[v], 1);
      };
  };

  /** Convert a multivariate polynomial data in the monomials basis into the bernstein basis 
   ** along the @v@ variable.
   **/
  template < class C, class eenv >
  void convertm2b (C * data, const eenv & e, binomials < C > &bins, int v)
  {
    const int *esz = e.szs ();
    const int *est = e.str ();
    int i, j;
    for (i = 0; i < e.sz (); i += est[v - 1])
      for (j = i; j < i + est[v]; convertm2b (data + j, esz[v], est[v], bins), j++);
  };
  
  /** Convert a multivariate polynomial data in the monomials basis into the bernstein basis **/
  template < class C, class eenv >
  void convertm2b (C * data, const eenv & env, binomials < C > &bins)
  {
    for (int v = 0; v < env.nvr (); convertm2b (data, env, bins, v++));
  };
  
  /** Convert a multivariate polynomial data in the bernstein basis to the monomial basis 
      along the @v@ variable
  **/
  template < class C, class eenv >
  void convertb2m (C * data, const eenv & e, binomials < C > &bins, int v)
  {
    const int *esz = e.szs ();
    const int *est = e.str ();
    int i, j;
    for (i = 0; i < e.sz (); i += est[v - 1])
      for (j = i; j < i + est[v];
	   convertb2m (data + j, esz[v], est[v], bins), j++);
  };
  
  /** Convert a multivariate polynomial data in the bernstein basis to the monomial basis **/
  template < class C, class eenv  >
  void convertb2m (C * data, const eenv & env, binomials < C > &bins)
  {
    for (int v = 0; v < env.nvr (); convertb2m (data, env, bins, v++));
  };

  template<typename C> inline 
  void diff( C * b, const C * a, unsigned sz, int sb, int sa )
  
  {
    assert(sa==sb);
    for ( unsigned i = 1; i < sz; i ++ ) { b[(i-1)*sb] = i*a[i]; };
    //    for ( unsigned i = 1, a += sa; i < sz; *b = *a*i++, b += sb, a += sa ); 
  };
  
  template<typename C> inline /* differentiation */
  void diff( C * b, C * e, C const * a, int s )
  { 
    unsigned i;
    for ( i = 1, a += s; b != e; *b = *a*i++, b += s, a += s );
  };

    template<typename C> inline /* differentiation */
  void diff( C * b, C * e, C const * a )
  { 
    for ( int i = 1; b != e; *b++ = *(++a)*i++ );
  };

  /*
  template<typename C>
  static void mdiff( const eenv& o, const eenv& a, C * op, C const * ap, int v )
  {
    typedef int sz_t;
    sz_t is = 0;
    for ( sz_t i = 0; i < o.sz(); i += o.st(v-1), is += a.st(v-1) )
      for ( sz_t j = 0; j < a.st(v); j ++ )
	diff(op+i+j, ap+is+j,a.sz(v), o.st(v), a.st(v));
  };
  */

  template<typename C>
  static void mdiff( C * dst, const eenv& a, C const * const src, int v )
  {
    const int   ssz  = a.sz ();
    const int * sszs = a.szs();
    const int * sstr = a.str();
    int rst = sstr[v-1]-sstr[v];
    unsigned i;
    C * ed, * fd;
    const C * es, * fs;
    for ( ed = dst, es = src; es < src+ssz; ed += rst, es += sstr[v-1] )
      for ( i = 1, fd = ed, fs = es+sstr[v]; fs != es+sstr[v-1]; fd += sstr[v], fs += sstr[v], i ++  )
	vct::scmul(fd,fd+sstr[v],fs,C(i));
  }

  
  
  template<typename C>
  static void bdiff( C * dst, const eenv& a, C const * const src, int v )
  {
    const int   ssz  = a.sz ();
    const int * sszs = a.szs();
    const int * sstr = a.str();
    int rst = sstr[v-1]-sstr[v];
    unsigned i;
    C * ed; const C * es;
    for ( ed = dst, es = src; es < src+ssz; ed += rst, es += sstr[v-1] )
      {
	C * fd; const C * fs;
	for ( fs = es, fd = ed; fs < es + sstr[v]; fs++, fd ++ )
	  {
	    C * dp; const C * p;
	    for ( p = fs, dp = fd; p < fs+rst; p += sstr[v], dp += sstr[v] )
	      *dp = (sszs[v]-1)*(*(p+sstr[v])-*p);
	  };
      };
  }
  
  template<class C> inline 
  void bsplit_uloop( C * r, C * rlast, C * l, const C& t, int st )
  {
    C * er, * p;
    for ( er = rlast - st ; er != r; er -= st, l += st )
      {
	*l = *r;
	for ( p = r; p != er; p += st )
	  *p = (1.0-t)**p+t**(p+st);
      };
    *l = *r;
  };
  
  template < class C, class eenv > inline
  void bsplit (C * l, C * r, const eenv & env, const C & t, int v )
  {
    const int *sszs = env.szs ();
    const int *sstr = env.str ();
    int i, j, k;
    C *er,*el;
    for ( el = l, er = r; el != l + sstr[-1]; el += sstr[v-1], er += sstr[v-1] )
      {
	C *fr,*fl;
	for ( fl = el, fr = er; fl != el + sstr[v]; fl++, fr++ )
	  bsplit_uloop(fr,fr+sstr[v-1],fl,t,sstr[v]);
      };
  };
  
  
  template<class C> inline 
  void bsplit2_uloop( C * r, C * rlast, C * l,  int st )
  {
    C * er, * p;
    for ( er = rlast - st ; er != r; er -= st, l += st )
      {
	*l = *r;
	for ( p = r; p != er; p += st )
	  *p = (*p+*(p+st))/C(2);
      };
    *l = *r;
  };

  template < class C, class eenv > 
  void bsplit2 (C * l, C * r, const eenv & env, int v )
  {
    const int *sszs = env.szs ();
    const int *sstr = env.str ();
    int i, j, k;
    C *er,*el;
    for ( el = l, er = r; el != l + sstr[-1]; el += sstr[v-1], er += sstr[v-1] )
      {
	C *fr,*fl;
	for ( fl = el, fr = er; fl != el + sstr[v]; fl++, fr++ )
	  bsplit2_uloop(fr,fr+sstr[v-1],fl,sstr[v]);
      };
  };


  template<class C> inline 
  void brestrict_uloop( C * r, C * rlast, int st, const C& a, const C& b  )
  {
    C * er, * p;
    for ( er = rlast - st ; er != r; er -= st )
      for ( p = r; p != er; p += st )
	*p = (a**p+b**(p+st));
  };
  
  template<class C> inline
  void brestrict_mloop  (C * s, const eenv & env, int v, const C& a, const C& b, const C& c, const C& d  )
  {
    const int *sszs = env.szs ();
    const int *sstr = env.str ();
    int i, j, k;
    for ( C * es = s; es != s + sstr[-1]; es += sstr[v-1] )
      for ( C * fs = es; fs != es + sstr[v]; fs ++ )
	{
	  brestrict_uloop(fs,fs+sstr[v-1],sstr[v],a,b);
	  brestrict_uloop(fs+sstr[v-1]-sstr[v],fs-sstr[v],-sstr[v],c,d);
	};
  };

  template<class C, class eenv > inline
  void  brestrict( C * src, const eenv& env, int v, const C& a, const C& b )
  {
    brestrict_mloop   ( src, env, v, C(1-a), a, C((b-a)/(C(1)-a)), C((C(1)-b)/(C(1)-a)) );
  };

  template<class C>
  void lface( C * dst, const eenv& a, const C * src, int v )
  {
    const int   ssz  = a.sz ();
    const int * sszs = a.szs();
    const int * sstr = a.str();

    unsigned i;
    C * ed; const C * es;
    const C * fs;
    for ( ed = dst, es = src; es < src+ssz; es += sstr[v-1] )
      for ( fs = es; fs < es + sstr[v]; *ed++ = *fs++ );
  };

  template<class C>
  void rface( C * dst, const eenv& a, const C * src, int v )
  {
    const int   ssz  = a.sz ();
    const int * sszs = a.szs();
    const int * sstr = a.str();

    unsigned i;
    C * ed; const C * es;
    const C * fs;
    es = src +  sstr[v-1]-sstr[v];
    for ( ed = dst; es < src+ssz; es += sstr[v-1] )
      {
	//	fs = es + sstr[v-1] - sstr[v];
	for ( fs = es; fs < es + sstr[v]; *ed++ = *fs++ );
      };
  };

}

__END_NAMESPACE_SYNAPS
#endif
