#ifndef TREES_MREG_H
#define TREES_MREG_H
#include <vector>
#include <queue>
namespace trees
{
  template< typename prms >
  struct mrg
  {
    struct ngh;
    mrg * m_f, * m_l, * m_r;
    struct ngh
    {
      mrg * m_own;
      ngh * m_nxt;
      ngh * m_prv;
      ngh * m_sbb;
    };
    
    ngh  m_list;
    ngh * nbegin() { return m_list.m_nxt; };
    ngh * nend  () { return &m_list; };
    static ngh * next( ngh * n ) { return n->m_nxt; };
    static ngh * prev( ngh * n ) { return n->m_prv; };
    
    static inline void unlink( mrg::ngh * n )
    {
      n->m_prv->m_nxt = n->m_nxt;
      n->m_nxt->m_prv = n->m_prv;
    };
    
    void add( mrg::ngh * n )
    {
      n->m_own = this;
      n->m_nxt = m_list.m_nxt;
      n->m_nxt->m_prv = n;
      n->m_prv = &m_list;
      m_list.m_nxt = n;
    };
    
    
    static void link( mrg * ra, mrg * rb )
    {
      ngh * na = new ngh;
      ngh * nb = new ngh;
      na->m_sbb = nb; 
      ra->add(na);
      nb->m_sbb = na; 
      rb->add(nb);
    };
    
    static bool root( ngh * n )
    {
      return n->m_sbb->m_own->m_f == 0;
    };
    
    mrg()
    {
      m_f = m_l = m_r = 0;
      m_list.m_prv = &m_list;
      m_list.m_nxt = &m_list;
    };
  
    ~mrg()
    {
      ngh * n = nbegin();
      while ( n != nend() )
	{
	  ngh * tmp = next(n);
	  delete n;
	  n = tmp;
	};
    };
    
    prms m_data;

    mrg( const prms& prm ) : m_data(prm) 
    {
      m_f = m_l = m_r = 0;
      m_list.m_prv = &m_list;
      m_list.m_nxt = &m_list;
    };
    
    static void merge( mrg * ra, mrg * rb, mrg * data )
    {
      // merging sub-regions is an error
      assert(ra->m_f==0 && rb->m_f==0); 
      
      data->m_l = ra;
      data->m_r = rb;
      ra->m_f   = data;
      rb->m_f   = data;
      
      ngh * n;
      ngh * tmp;
      for ( n = ra->nbegin(); n != ra->nend(); n = tmp )
	{
	  tmp = next(n);
	  if ( root( n ) ) 
	    {
	      unlink(n);
	      data->add(n);
	    }
	};
      
      for ( n = rb->nbegin(); n != rb->nend(); n = tmp )
	{
	  tmp = next(n);
	  if ( root(n) )
	    {
	      unlink(n);
	      data->add(n);
	    };
	};
      
      prms::merge(data->m_data,ra->m_data,rb->m_data);
    };
    
    static mrg * merge( mrg * ra, mrg * rb )
    {
      mrg * tmp = new mrg;
      merge(ra,rb,tmp);
      return tmp;
    };
    
    
    bool root() { return m_f == 0; };
    
    
    typedef typename prms::score_t score_t;
    
    score_t score( mrg * rg )
    {
      if ( ! rg->root() ) return 0;
      return prms::score(m_data,rg->m_data);
    };
    
    static mrg * value( ngh *  n )
    {
      return n->m_sbb->m_own; 
    };
    mrg * best()
    {
      //      std::cout << "best";
      if ( nbegin() == nend() ) return 0;
      ngh * best = nbegin();
      score_t scr = score( value( best ) );
      ngh * n = next(nbegin());
      while ( n != nend() )
	{
	  if ( value(n)->root() )
	    {
	      score_t tmp = score(value(n));
	      if ( tmp >= scr )
		{
		  scr = tmp;
		  best = n;
		};
	    };
	  n = next( n);
	};
      if ( ! value(best)->root() ) return 0;
      //      std::cout << " \n";
      return value(best);
      
    };
    
    struct cmp
    {
      bool operator()( mrg * const a, mrg * const b ) const 
      {
	return ! (a->m_data < b->m_data);
      };
    };
    
    static void build( std::vector< mrg * >& result,
		       mrg ** begin, mrg ** end )
    {
      std::priority_queue< mrg*, std::vector<mrg*>, cmp > heap;
      for ( mrg ** a = begin; a != end; a++ ) { heap.push(*a); };
      while( ! heap.empty() )
	{
	  mrg * crr = heap.top();
	  //	  std::cout << crr->m_data.m_sz  << std::endl;
	  //std::cout << crr << std::endl;
	  heap.pop(); 
	  //std::cout << "pop \n";
	  
	  if ( crr->root() )
	    {
	      // std::cout << "root\n";
	      mrg * bst = crr->best();
	      //	      std::cout << "best\n";
	      if ( bst )
		{
		  heap.push( merge( crr, bst ) );
		  //		  std::cout << "push\n";
		}
	      else
		{
		  result.push_back(crr); 
		  //  std::cout << "result\n";
		};
	    };
	  
	};
    };

    void dump( std::ostream& out, int n = 0 )
    {
      for ( int i = 0; i < n; i ++ )
	out << "\t";
      out << m_data;
      out << std::endl;
      if ( m_l ) m_l->dump(out,n+1);
      if ( m_r ) m_r->dump(out,n+1);
    };
  };
  
  template< class prms >
  std::ostream& operator<<( std::ostream& out, mrg< prms > * m )
  {
    m->dump( out );
    return out;
  };
  
  template< typename prms > inline 
  mrg<prms> const * const left( mrg<prms> const * const m )
  { return m->m_l; };
  template< typename prms > inline 
  mrg<prms> *& left( mrg<prms> * m ) { return m->m_l; };
  
  template< typename prms > inline 
  mrg<prms> const * const right( mrg<prms> const * const m )
  { return m->m_r; };
  template< typename prms > inline 
  mrg<prms> *& right( mrg<prms> * m ) { return m->m_r; };
  template< typename prms > inline 
  mrg<prms> const * const father( mrg<prms> const * const m )
  { return m->m_l; };
  template< typename prms > inline 
  mrg<prms> *& father( mrg<prms> * m ) { return m->m_f; };
  template< typename prms > inline 
  prms&  data( mrg<prms> * m ) { return m->m_data; };
  template< typename prms > inline 
  const prms& data( mrg<prms> const *const  m ) { return m->m_data; };
}; 
#endif
