/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef synaps_implicit_quadric_h
#define synaps_implicit_quadric_h
/*********************************************************************/
#include <string>
#include <math.h>
#include <stdlib.h>
#include <synaps/arithm/let.h>
#include <synaps/linalg/MatrDse.h>
#include <synaps/mpol/mpoldse/triangular.h>
#include <synaps/mpol/MPolDse.h>
/*********************************************************************/
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
namespace implicit
{
  /// The implicit quadric class
  /**
    The implicit quadric is represented by the polynomial 
    type <tt>MPolDse<C,triangular::monomial<C> ></tt>, which is a dense
    representation in the monomial order.

    \param C is the type of the coefficients of the polynomial.

    The following constraints must be hold:
    \li <tt>the number of variable must be at least one</tt>.
    \li <tt>for each variable, the degree must be less or equal 2</tt>.
  **/
  template<class C>
  struct quadric: public MPolDse<C,triangular::monomial<C> >
  {
    typedef MPolDse<C,triangular::monomial<C> > polynomial_t;
    typedef C coeff_t;

    /*! The default constructor. */
    quadric():polynomial_t() {}

    /*!
      \param s a string which represents the polynomial.

      Construct an implicit quadric by giving a \c char \c * 
      which represents the polynomial (typically a polynomial stored 
      in a file).

      \attention If the degree of each variable is not less or equal 2, 
      we can't construct the object.
    */
    quadric(const char* s) : polynomial_t(s)
      {
	//	polynomial_t p=*this;
	//	std::cout<<p<<std::endl;
	//	if ( degree((polynomial_t&)*this) && (degree((polynomial_t)*this)<1) )
	// 	  {
	// 	    std::cerr<<"It is not a quadric!"<<std::endl;
	// 	    exit(1);
	// 	  } 
      }

    /*!
      \param eq the polynomial.

      Construct an implicit quadric by giving its polynomial.

      \attention If the degree of each variable is not less or equal 2,
      we can't construct the object.
    */
    quadric(const polynomial_t & eq): polynomial_t(eq){}

    int nbvar() const { return this->rep().nbvar_;}

  };

  /*!
    The \c operator \c << .\n
    This operator allows us to display a surface.
  */
  template<class C>
  std::ostream&  operator<<(std::ostream & os, const quadric<C>& q)
  {
    typedef typename quadric<C>::polynomial_t polynomial_t;
    os <<"ImplicitQuadric("<<(const polynomial_t&)q<<") ";
    return os;
  }

  template<class C>
  std::istream&  operator>>(std::istream & is, quadric<C>& q)
  {
    typedef typename quadric<C>::polynomial_t polynomial_t;
    is >>(polynomial_t)q;
    return is;
  }

  template<class C>
  void assign(MatrDse<C>& A, const implicit::quadric<C> & q)
  {
    int nbv=q.nbvar()+1,s=0;
    A.resize(nbv,nbv);
    for(unsigned j=0;j<nbv;j++)
      {
	for(unsigned i=j;i<nbv;i++)
	  {
	    A(i,j)=q[s+i-j];
	    if(i>j)
	      A(j,i)=A(i,j);
	    else
	      A(i,i)*=2;
	  }
	s+= nbv-j;
      }
  }

} //implicit
  

namespace xml 
{
  template<class ISTREAM, class C>
  ISTREAM & read(ISTREAM & is, implicit::quadric<C>& q)
  {
    xml::read(is,q.equation);
    return is;
  }

}
namespace let
{

  template<class C, class R>
  MatrDse<C> convert(const implicit::quadric<C> & q, As<MatrDse<C> > mth)
  {
    MatrDse<C> A(q.nbvar()+1,q.nbvar()+1);
    assign(A,q);
    return A;
  }
}
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
/*********************************************************************/
#endif // synaps_implicit_quadric_h
