/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): L. Deschamps                                          *
*              G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
/** \file                                                           **/
#ifndef axel_mesh_cube3d_H
#define axel_mesh_cube3d_H
/*********************************************************************/

#include <synaps/init.h>

#include <vector>

#include "synaps/topology/point.h"
#include "synaps/topology/vertex.h"
#include "synaps/topology/segment.h"
#include "synaps/topology/triangle.h"
#include <synaps/topology/mesh/scene.h>

/*********************************************************************/

__BEGIN_NAMESPACE_SYNAPS

template<class MP , 
         class REP3D  > struct scene;
namespace topology
{

  //! The structure that defines a cube.
  /*!
    Class BezierCube that defines a cube.
    This cube is represented by 12 edges each of these edges
    have at most one vertex.
  */
  template < class MP = MPol<double>,
	     class REP3D = linalg::rep3d< typename MP::coeff_t >,
	     class V = topology::vertex< topology::point< typename MP::coeff_t > >,
	     class S = topology::segment< topology::point< typename MP::coeff_t > >,
	     class T = topology::triangle< topology::point< typename MP::coeff_t > > > 

  struct cube3d 
  {
    typedef MP                       polynomial_t;
    typedef typename MP::coeff_t     coeff_t;
    typedef REP3D                    rep_t;
    typedef V                        vertex_t;
    typedef S                        segment_t;
    typedef T                        triangle_t;

    /*! There is, at most, one vertex per edges. */
    vertex_t * rootOnEdge_1;
    vertex_t * rootOnEdge_2;
    vertex_t * rootOnEdge_3;
    vertex_t * rootOnEdge_4;
    vertex_t * rootOnEdge_5;
    vertex_t * rootOnEdge_6;
    vertex_t * rootOnEdge_7;
    vertex_t * rootOnEdge_8;
    vertex_t * rootOnEdge_9;
    vertex_t * rootOnEdge_10;
    vertex_t * rootOnEdge_11;
    vertex_t * rootOnEdge_12;
  
    /*! Boolean use to know if the correspondant edge is ever computed. */
    
    bool edge_1_isProcessed;
    bool edge_2_isProcessed;
    bool edge_3_isProcessed;
    bool edge_4_isProcessed;
    bool edge_5_isProcessed;
    bool edge_6_isProcessed;
    bool edge_7_isProcessed;
    bool edge_8_isProcessed;
    bool edge_9_isProcessed;
    bool edge_10_isProcessed;
    bool edge_11_isProcessed;
    bool edge_12_isProcessed;

    /*! The 8 corner's sign. */
    int _000, _001, _010, _011, _100, _101, _110, _111;

    /*! Default constructor. */
    cube3d() :
    
      rootOnEdge_1(  NULL ),
      rootOnEdge_2(  NULL ),
      rootOnEdge_3(  NULL ),
      rootOnEdge_4(  NULL ),
      rootOnEdge_5(  NULL ),
      rootOnEdge_6(  NULL ),
      rootOnEdge_7(  NULL ),
      rootOnEdge_8(  NULL ),
      rootOnEdge_9(  NULL ),
      rootOnEdge_10( NULL ),
      rootOnEdge_11( NULL ),
      rootOnEdge_12( NULL ),
    
      edge_1_isProcessed( false ),
      edge_2_isProcessed( false ),
      edge_3_isProcessed( false ),
      edge_4_isProcessed( false ),
      edge_5_isProcessed( false ),
      edge_6_isProcessed( false ),
      edge_7_isProcessed( false ),
      edge_8_isProcessed( false ),
      edge_9_isProcessed( false ),
      edge_10_isProcessed( false ),
      edge_11_isProcessed( false ),
      edge_12_isProcessed( false )

    {}

  private :
  
    /*! 
      Copy constructor.
      This is private to prohibit the copy constructor. 
    */
    cube3d( const cube3d<MP,REP3D,V,S,T>& ) 
    {
      std::cerr << " cube3d(const cube3d<MP,REP3D,V,S,T>& \n"; 
    };

  public :

    /*! 
      Get all the roots of a cube.
    */
    void getAllRoots( std::vector< vertex_t * > &roots ) 
    {
      if ( rootOnEdge_1  != NULL ) roots.push_back( rootOnEdge_1  );
      if ( rootOnEdge_2  != NULL ) roots.push_back( rootOnEdge_2  );
      if ( rootOnEdge_3  != NULL ) roots.push_back( rootOnEdge_3  );
      if ( rootOnEdge_4  != NULL ) roots.push_back( rootOnEdge_4  );
      if ( rootOnEdge_5  != NULL ) roots.push_back( rootOnEdge_5  );
      if ( rootOnEdge_6  != NULL ) roots.push_back( rootOnEdge_6  );
      if ( rootOnEdge_7  != NULL ) roots.push_back( rootOnEdge_7  );
      if ( rootOnEdge_8  != NULL ) roots.push_back( rootOnEdge_8  );
      if ( rootOnEdge_9  != NULL ) roots.push_back( rootOnEdge_9  );
      if ( rootOnEdge_10 != NULL ) roots.push_back( rootOnEdge_10 );
      if ( rootOnEdge_11 != NULL ) roots.push_back( rootOnEdge_11 );
      if ( rootOnEdge_12 != NULL ) roots.push_back( rootOnEdge_12 );
    }

  
    /*! 
      Get roots on the face 1. 
      Return false if there is no root, true otherwise. 
    */
    bool getRootsOnFace_1( std::vector< vertex_t * > &roots ) 
    {
      if ( rootOnEdge_1  != NULL ) roots.push_back( rootOnEdge_1  );
      if ( rootOnEdge_3  != NULL ) roots.push_back( rootOnEdge_3  );
      if ( rootOnEdge_9  != NULL ) roots.push_back( rootOnEdge_9  );
      if ( rootOnEdge_10 != NULL ) roots.push_back( rootOnEdge_10 );
      
      return ( !roots.empty() );
    }
  
    /*!
      Get roots on the face 2. 
      Return false if there is no root, true otherwise.
    */
    bool getRootsOnFace_2( std::vector< vertex_t * > &roots ) 
    {
      if ( rootOnEdge_1  != NULL ) roots.push_back( rootOnEdge_1  );
      if ( rootOnEdge_2  != NULL ) roots.push_back( rootOnEdge_2  );
      if ( rootOnEdge_5  != NULL ) roots.push_back( rootOnEdge_5  );
      if ( rootOnEdge_6  != NULL ) roots.push_back( rootOnEdge_6  );
      
      return ( !roots.empty() );
    }
  
    /*!
      Get roots on the face 3. 
      Return false if there is no root, true otherwise. 
    */
    bool getRootsOnFace_3( std::vector< vertex_t * > &roots ) 
    {
      if ( rootOnEdge_4  != NULL ) roots.push_back( rootOnEdge_4  );
      if ( rootOnEdge_2  != NULL ) roots.push_back( rootOnEdge_2  );
      if ( rootOnEdge_11 != NULL ) roots.push_back( rootOnEdge_11 );
      if ( rootOnEdge_12 != NULL ) roots.push_back( rootOnEdge_12 );
      
      return ( !roots.empty() );
    }
  
    /*!
      Get roots on the face 4.
      Return false if there is no root, true otherwise. 
    */
    bool getRootsOnFace_4( std::vector< vertex_t * > &roots ) 
    {
      if ( rootOnEdge_6  != NULL ) roots.push_back( rootOnEdge_6  );
      if ( rootOnEdge_8  != NULL ) roots.push_back( rootOnEdge_8  );
      if ( rootOnEdge_10 != NULL ) roots.push_back( rootOnEdge_10 );
      if ( rootOnEdge_12 != NULL ) roots.push_back( rootOnEdge_12 );
      
      return ( !roots.empty() );
    }
 
    /*!
      Get roots on the face 5.
      Return false if there is no root, true otherwise.
    */
    bool getRootsOnFace_5( std::vector< vertex_t * > &roots )
    {
      if ( rootOnEdge_3  != NULL ) roots.push_back( rootOnEdge_3  );
      if ( rootOnEdge_4  != NULL ) roots.push_back( rootOnEdge_4  );
      if ( rootOnEdge_7  != NULL ) roots.push_back( rootOnEdge_7  );
      if ( rootOnEdge_8  != NULL ) roots.push_back( rootOnEdge_8  );
      
      return ( !roots.empty() );
    } 

    /*!
      Get roots on the face 6.
      Return false if there is no root, true otherwise. 
    */
    bool getRootsOnFace_6( std::vector< vertex_t * > &roots ) 
    {
      if ( rootOnEdge_5  != NULL ) roots.push_back( rootOnEdge_5  );
      if ( rootOnEdge_9  != NULL ) roots.push_back( rootOnEdge_9  );
      if ( rootOnEdge_7  != NULL ) roots.push_back( rootOnEdge_7  );
      if ( rootOnEdge_11 != NULL ) roots.push_back( rootOnEdge_11 );
      
      return ( !roots.empty() );
    }
  

    /*!
      Get roots on the face number 1 with good orientation.
      These orientation is based on the sign of corners.
      Return false if there is no root, true otherwise.
      "roots" is the list of roots in correct order for a good
      orientation of triangles.
      This order is differents if roots are taken on the current cube
      or if there are taken on a neighbor.
      "self" is true if roots are taken from the current cube 
      false otherwise. 
    */
    bool getLinkRootsOnFace_1( std::vector< vertex_t * > &roots, bool self ) 
    {
      bool isRootOnEdge_1 = false,
      isRootOnEdge_3 = false,
      isRootOnEdge_9 = false,
      isRootOnEdge_10 = false;
      self = !self;
    
      if ( rootOnEdge_1  != NULL ) isRootOnEdge_1  = true;
      if ( rootOnEdge_3  != NULL ) isRootOnEdge_3  = true;
      if ( rootOnEdge_9  != NULL ) isRootOnEdge_9  = true;
      if ( rootOnEdge_10 != NULL ) isRootOnEdge_10 = true;
      
      if ( isRootOnEdge_1 && isRootOnEdge_3) {
	if ( ( ( _000 > 0 ) && self ) || ( (_000 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_1 );
	  roots.push_back( rootOnEdge_3 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_3 );
	  roots.push_back( rootOnEdge_1 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_1 && isRootOnEdge_9) {
	if ( ( ( _000 > 0 ) && self ) || ( (_000 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_1 );
	  roots.push_back( rootOnEdge_9 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_9 );
	  roots.push_back( rootOnEdge_1 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_1 && isRootOnEdge_10) {
	if ( ( ( _000 > 0 )&& self ) || ( (_000 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_1 );
	  roots.push_back( rootOnEdge_10 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_10 );
	  roots.push_back( rootOnEdge_1 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_3 && isRootOnEdge_9) {
	if ( ( ( _100  > 0) && self ) || ( (_100 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_9 );
	  roots.push_back( rootOnEdge_3 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_3 );
	  roots.push_back( rootOnEdge_9 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_3 && isRootOnEdge_10) {
	if ( ( ( _100  > 0) && self ) || ( (_100 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_10 );
	  roots.push_back( rootOnEdge_3 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_3 );
	  roots.push_back( rootOnEdge_10 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_9 && isRootOnEdge_10) {
	if ( ( ( _000 > 0) && self ) || ( (_000 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_10 );
	  roots.push_back( rootOnEdge_9 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_9 );
	  roots.push_back( rootOnEdge_10 );
	  return true;
	}
      }
      
      return false;
    }
  
    /*! 
      Get roots on the face number 2 with good orientation.
      These orientation is based on the sign of corners.
      Return false if there is no root, true otherwise.
      "roots" is the list of roots in correct order for a good
      orientation of triangles.
      This order is differents if roots are taken on the current cube
      or if there are taken on a neighbor.
      "self" is true if roots are taken from the current cube 
      false otherwise. 
    */
    bool getLinkRootsOnFace_2( std::vector< vertex_t * > &roots, bool self ) 
    {
      bool isRootOnEdge_1 = false,
      isRootOnEdge_2 = false,
      isRootOnEdge_5 = false,
      isRootOnEdge_6 = false;
      self = !self;
      
      if ( rootOnEdge_1 != NULL ) isRootOnEdge_1 = true;
      if ( rootOnEdge_2 != NULL ) isRootOnEdge_2 = true;
      if ( rootOnEdge_5 != NULL ) isRootOnEdge_5 = true;
      if ( rootOnEdge_6 != NULL ) isRootOnEdge_6 = true;
      
      if ( isRootOnEdge_1 && isRootOnEdge_2 ) {
	if ( ( (_010 > 0) && self ) || ( (_010 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_1 );
	  roots.push_back( rootOnEdge_2 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_2 );
	  roots.push_back( rootOnEdge_1 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_1 && isRootOnEdge_5) {
	if ( ( (_010  > 0) && self ) || ( (_010 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_1 );
	  roots.push_back( rootOnEdge_5 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_5 );
	  roots.push_back( rootOnEdge_1 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_1 && isRootOnEdge_6) {
	if ( ( ( _010 > 0 ) && self ) || ( (_010 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_1 );
	  roots.push_back( rootOnEdge_6 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_6 );
	  roots.push_back( rootOnEdge_1 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_2 && isRootOnEdge_5) {
	if ( ( (_010  > 0) && self ) || ( (_010 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_5 );
	  roots.push_back( rootOnEdge_2 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_2 );
	  roots.push_back( rootOnEdge_5 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_2 && isRootOnEdge_6) {
	if ( ( ( _011 > 0 ) && self ) || ( (_011 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_6 );
	  roots.push_back( rootOnEdge_2 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_2 );
	  roots.push_back( rootOnEdge_6 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_5 && isRootOnEdge_6) {
	if ( ( (_000 > 0 ) && self ) || ( (_000 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_5 );
	  roots.push_back( rootOnEdge_6 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_6 );
	  roots.push_back( rootOnEdge_5 );
	  return true;
	}
      }
      
      return false;
    }

    /*!
      Get roots on the face number 3 with good orientation.
      These orientation is based on the sign of corners.
      Return false if there is no root, true otherwise.
      "roots" is the list of roots in correct order for a good
      orientation of triangles.
      This order is differents if roots are taken on the current cube
      or if there are taken on a neighbor.
      "self" is true if roots are taken from the current cube 
      false otherwise. 
    */
    bool getLinkRootsOnFace_3( std::vector< vertex_t * > &roots, bool self ) 
    {
      bool isRootOnEdge_2 = false,
      isRootOnEdge_4 = false,
      isRootOnEdge_11 = false,
      isRootOnEdge_12 = false;
      self = !self;
      
      if ( rootOnEdge_2  != NULL ) isRootOnEdge_2  = true;
      if ( rootOnEdge_4  != NULL ) isRootOnEdge_4  = true;
      if ( rootOnEdge_11 != NULL ) isRootOnEdge_11 = true;
      if ( rootOnEdge_12 != NULL ) isRootOnEdge_12 = true;
      
      if ( isRootOnEdge_2 && isRootOnEdge_4) {
	if ( ( ( _011  > 0) && self ) || ( (_011 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_2 );
	  roots.push_back( rootOnEdge_4 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_4 );
	  roots.push_back( rootOnEdge_2 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_2 && isRootOnEdge_11) {
	if ( ( (_011  > 0) && self ) || ( (_011 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_2 );
	  roots.push_back( rootOnEdge_11 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_11 );
	  roots.push_back( rootOnEdge_2 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_2 && isRootOnEdge_12) {
	if ( ( (_011 > 0 ) && self ) || ( (_011 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_2 );
	  roots.push_back( rootOnEdge_12 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_12 );
	  roots.push_back( rootOnEdge_2 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_4 && isRootOnEdge_11) {
	if ( ( (_011  > 0) && self ) || ( ( _011 < 0)  && !self ) ) {
	  roots.push_back( rootOnEdge_11 );
	  roots.push_back( rootOnEdge_4 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_4 );
	  roots.push_back( rootOnEdge_11 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_4 && isRootOnEdge_12) {
	if ( ( ( _111  > 0) && self ) || ( (_111 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_12 );
	  roots.push_back( rootOnEdge_4 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_4 );
	  roots.push_back( rootOnEdge_12 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_11 && isRootOnEdge_12) {
	if ( ( ( _001  > 0) && self ) || ( (_001 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_11 );
	  roots.push_back( rootOnEdge_12 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_12 );
	  roots.push_back( rootOnEdge_11 );
	  return true;
	}
      }
      
      return false;
    }
  
    /*!
      Get roots on the face number 4 with good orientation.
      These orientation is based on the sign of corners.
      Return false if there is no root, true otherwise.
      "roots" is the list of roots in correct order for a good
      orientation of triangles.
      This order is differents if roots are taken on the current cube
      or if there are taken on a neighbor.
      "self" is true if roots are taken from the current cube 
      false otherwise.
    */
    bool getLinkRootsOnFace_4( std::vector< vertex_t * > &roots, bool self ) 
    {
      bool isRootOnEdge_6 = false,
      isRootOnEdge_8 = false,
      isRootOnEdge_10 = false,
      isRootOnEdge_12 = false;
      self = !self;
      
      if ( rootOnEdge_6  != NULL ) isRootOnEdge_6  = true;
      if ( rootOnEdge_8  != NULL ) isRootOnEdge_8  = true;
      if ( rootOnEdge_10 != NULL ) isRootOnEdge_10 = true;
      if ( rootOnEdge_12 != NULL ) isRootOnEdge_12 = true;
      
      if ( isRootOnEdge_6 && isRootOnEdge_8 ) {
	if ( ( ( _010  > 0 ) && self ) || ( (_010 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_6 );
	  roots.push_back( rootOnEdge_8 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_8 );
	  roots.push_back( rootOnEdge_6 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_6 && isRootOnEdge_10) {
	if ( ( ( _010 > 0) && self ) || ( (_010 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_6 );
	  roots.push_back( rootOnEdge_10 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_10 );
	  roots.push_back( rootOnEdge_6 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_6 && isRootOnEdge_12) {
	if ( ( ( _010 > 0)  && self ) || ( (_010 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_6 );
	  roots.push_back( rootOnEdge_12 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_12 );
	  roots.push_back( rootOnEdge_6 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_8 && isRootOnEdge_10) {
	if ( ( ( _110 > 0 ) && self ) || ( (_110 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_10 );
	  roots.push_back( rootOnEdge_8 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_8 );
	  roots.push_back( rootOnEdge_10 );
	  return true;
	}
      }
    
      if ( isRootOnEdge_8 && isRootOnEdge_12) {
	if ( ( ( _110 > 0 ) && self ) || ( (_110 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_12 );
	  roots.push_back( rootOnEdge_8 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_8 );
	  roots.push_back( rootOnEdge_12 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_10 && isRootOnEdge_12) {
	if ( ( ( _110  > 0) && self ) || ( ( _110 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_10 );
	  roots.push_back( rootOnEdge_12 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_12 );
	  roots.push_back( rootOnEdge_10 );
	  return true;
	}
      }
      
      return false;
    }
  
    /*!
      Get roots on the face number 5 with good orientation.
      These orientation is based on the sign of corners.
      Return false if there is no root, true otherwise.
      "roots" is the list of roots in correct order for a good
      orientation of triangles.
      This order is differents if roots are taken on the current cube
      or if there are taken on a neighbor.
      "self" is true if roots are taken from the current cube 
      false otherwise. 
    */
    bool getLinkRootsOnFace_5( std::vector< vertex_t * > &roots, bool self ) 
    {
      bool isRootOnEdge_3 = false,
      isRootOnEdge_4 = false,
      isRootOnEdge_7 = false,
      isRootOnEdge_8 = false;
      self = !self;
      
      if ( rootOnEdge_3 !=NULL ) isRootOnEdge_3 = true;
      if ( rootOnEdge_4 !=NULL ) isRootOnEdge_4 = true;
      if ( rootOnEdge_7 !=NULL ) isRootOnEdge_7 = true;
      if ( rootOnEdge_8 !=NULL ) isRootOnEdge_8 = true;
      
      if ( isRootOnEdge_3 && isRootOnEdge_4 ) {
	if ( ( ( _100  > 0) && self ) || ( ( _100 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_3 );
	  roots.push_back( rootOnEdge_4 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_4 );
	  roots.push_back( rootOnEdge_3 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_3 && isRootOnEdge_7) {
	if ( ( (_100 > 0 ) && self ) || ( (_100 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_3 );
	  roots.push_back( rootOnEdge_7 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_7 );
	  roots.push_back( rootOnEdge_3 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_3 && isRootOnEdge_8) {
	if ( ( (_100 > 0 ) && self ) || ( (_100 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_3 );
	  roots.push_back( rootOnEdge_8 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_8 );
	  roots.push_back( rootOnEdge_3 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_4 && isRootOnEdge_7) {
	if ( ( (_101  > 0) && self ) || ( (_101 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_7 );
	  roots.push_back( rootOnEdge_4 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_4 );
	  roots.push_back( rootOnEdge_7 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_4 && isRootOnEdge_8) {
	if ( ( ( _101 > 0 ) && self ) || ( (_101 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_8 );
	  roots.push_back( rootOnEdge_4 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_4 );
	  roots.push_back( rootOnEdge_8 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_7 && isRootOnEdge_8) {
	if ( ( (_101  > 0)  && self ) || ( (_101 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_7 );
	  roots.push_back( rootOnEdge_8 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_8 );
	  roots.push_back( rootOnEdge_7 );
	  return true;
	}
      }
      
      return false;
    }
  
    /*!
      Get roots on the face number 6 with good orientation.
      These orientation is based on the sign of corners.
      Return false if there is no root, true otherwise.
      "roots" is the list of roots in correct order for a good
      orientation of triangles.
      This order is differents if roots are taken on the current cube
      or if there are taken on a neighbor.
      "self" is true if roots are taken from the current cube 
      false otherwise. 
    */
    bool getLinkRootsOnFace_6( std::vector< vertex_t * > &roots, bool self ) 
    {
      bool isRootOnEdge_5 = false,
      isRootOnEdge_7 = false,
      isRootOnEdge_9 = false,
      isRootOnEdge_11 = false;
      self = !self;
      
      if ( rootOnEdge_5  !=NULL ) isRootOnEdge_5  = true;
      if ( rootOnEdge_7  !=NULL ) isRootOnEdge_7  = true;
      if ( rootOnEdge_9  !=NULL ) isRootOnEdge_9  = true;
      if ( rootOnEdge_11 !=NULL ) isRootOnEdge_11 = true;
      
      if ( isRootOnEdge_5 && isRootOnEdge_7 ) {
	if ( ( (_001  > 0) && self ) || ( ( _001 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_5 );
	  roots.push_back( rootOnEdge_7 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_7 );
	  roots.push_back( rootOnEdge_5 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_5 && isRootOnEdge_9) {
	if ( ( ( _001  > 0) && self ) || ( (_001 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_5 );
	  roots.push_back( rootOnEdge_9 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_9 );
	  roots.push_back( rootOnEdge_5 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_5 && isRootOnEdge_11) {
	if ( ( ( _001  > 0) && self ) || ( (_001 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_5 );
	  roots.push_back( rootOnEdge_11 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_11 );
	  roots.push_back( rootOnEdge_5 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_7 && isRootOnEdge_9) {
	if ( ( (_101 > 0 ) && self ) || ( (_101 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_9 );
	  roots.push_back( rootOnEdge_7 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_7 );
	  roots.push_back( rootOnEdge_9 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_7 && isRootOnEdge_11) {
	if ( ( ( _101  > 0) && self ) || ( (_101 < 0 ) && !self ) ) {
	  roots.push_back( rootOnEdge_11 );
	  roots.push_back( rootOnEdge_7 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_7 );
	  roots.push_back( rootOnEdge_11 );
	  return true;
	}
      }
      
      if ( isRootOnEdge_9 && isRootOnEdge_11) {
	if ( ( ( _001 > 0 ) && self ) || ( (_001 < 0) && !self ) ) {
	  roots.push_back( rootOnEdge_9 );
	  roots.push_back( rootOnEdge_11 );
	  return true;
	}
	else {
	  roots.push_back( rootOnEdge_11 );
	  roots.push_back( rootOnEdge_9 );
	  return true;
	}
      }
      
      return false;
    }

 

    /*! 
      Intervention of boolean for the orientation of the triangles.
    */
    bool buildListOfRoots( std::vector< int > &listRootsOnEdges)
    {
      int firstRoot;
      int faceNumber;
      
      if ( rootOnEdge_1 != NULL ) {
	if ( _010 < 0 ) {
	  firstRoot = 1;
	  faceNumber = 2;
	}
	else {
	  firstRoot = 1;
	  faceNumber = 1;
	}
      }
      else if ( rootOnEdge_2 != NULL ) {
	if ( _011 < 0 ) {
	  firstRoot = 2;
	  faceNumber = 3;
	}
	else{
	  firstRoot = 2;
	  faceNumber = 2;
	}
      }
      else if ( rootOnEdge_3 != NULL ) {
	if ( _110 < 0 ) {
	  firstRoot = 3;
	  faceNumber = 1;
	}
	else {
	  firstRoot = 3;
	  faceNumber = 5;
	}
      }
      else if ( rootOnEdge_4 != NULL ) {
	if ( _111 < 0 ) {
	  firstRoot = 4;
	  faceNumber = 5;
	}
	else {
	  firstRoot = 4;
	  faceNumber = 3;
	}
      }
      else if ( rootOnEdge_5 != NULL ) {
	if ( _001 < 0 ) {
	  firstRoot = 5;
	  faceNumber = 6;
	}
	else { 
	  firstRoot = 5;
	  faceNumber = 2;
	}
      }
      else if ( rootOnEdge_6 != NULL ) {
	if ( _011 < 0 ) {
	  firstRoot = 6;
	  faceNumber = 2;
	}
	else {
	  firstRoot = 6;
	  faceNumber = 4;
	}
      }
      else if ( rootOnEdge_7 != NULL ) {
	if ( _101 < 0 ) {
	  firstRoot = 7;
	  faceNumber = 5;
	}
	else {
	  firstRoot = 7;
	  faceNumber = 6;
	}
      }
      else if ( rootOnEdge_8 != NULL ) {
	if ( _111 < 0 ) {
	  firstRoot = 8;
	  faceNumber = 4;
	}
	else {
	  firstRoot = 8;
	  faceNumber = 5;
	}
      }
      else if ( rootOnEdge_9 != NULL ) {
	if ( _000 < 0 ) {
	  firstRoot = 9;
	  faceNumber = 6;
	}
	else {
	  firstRoot = 9;
	  faceNumber = 1;
	}
      }
      else if ( rootOnEdge_10 != NULL ) {
	if ( _010 < 0 ) {
	  firstRoot = 10;
	  faceNumber = 1;
	}
	else {
	  firstRoot = 10;
	  faceNumber = 4;
	}
      }
      else if ( rootOnEdge_11 != NULL ) {
	if ( _001 < 0 ) {
	  firstRoot = 11;
	  faceNumber = 3;
	}
	else {
	  firstRoot = 11;
	  faceNumber = 6;
	}
      }
      else if ( rootOnEdge_12 != NULL ) {
	if ( _011 < 0 ) {
	  firstRoot = 12;
	  faceNumber = 4;
	}
	else {
	  firstRoot = 12;
	  faceNumber = 3;
	}
      }
      // Solution sur aucune des aretes.
      // Dans ce cas peut etre, ne pas faire de traitement.
      else {
	std::cerr << " Erreur : buildListOfRoots, pas de racines sut les aretes... \n";
	return false;
      }
      
      listRootsOnEdges.push_back( firstRoot );
    
      int lastRoot = 0;
      int nextRoot = firstRoot ;
      
      
      while ( firstRoot != lastRoot ) {
	
	switch ( nextRoot ) {
	  
	  // Racine sur l'arete 1.
	case 1 : 
	  if ( faceNumber == 1 ) {
	    nextRootOnFace_1( 1, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 2 ) {
	    nextRootOnFace_2( 1, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 2.
	case 2 : 
	  if ( faceNumber == 2 ) {
	    nextRootOnFace_2( 2, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 3 ) {
	    nextRootOnFace_3( 2, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 3.
	case 3 : 
	  if ( faceNumber == 1 ) {
	    nextRootOnFace_1( 3, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 5 ) {
	    nextRootOnFace_5( 3, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 4.
	case 4 : 
	  if ( faceNumber == 3 ) {
	    nextRootOnFace_3( 4, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 5 ) {
	    nextRootOnFace_5( 4, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 5.
	case 5 : 
	  if ( faceNumber == 2 ) {
	    nextRootOnFace_2( 5, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 6 ) {
	    nextRootOnFace_6( 5, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 6.
	case 6 : 
	  if ( faceNumber == 2 ) {
	    nextRootOnFace_2( 6, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 4 ) {
	    nextRootOnFace_4( 6, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 7.
	case 7 : 
	  if ( faceNumber == 6 ) {
	    nextRootOnFace_6( 7, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 5 ) {
	    nextRootOnFace_5( 7, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 8.
	case 8 : 
	  if ( faceNumber == 4 ) {
	    nextRootOnFace_4( 8, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 5 ) {
	    nextRootOnFace_5( 8, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 9.
	case 9 : 
	  if ( faceNumber == 1 ) {
	    nextRootOnFace_1( 9, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 6 ) {
	    nextRootOnFace_6( 9, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 10.
	case 10 : 
	  if ( faceNumber == 1) {
	    nextRootOnFace_1( 10, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 4 ) {
	    nextRootOnFace_4( 10, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 11.
	case 11 : 
	  if ( faceNumber == 3 ) {
	    nextRootOnFace_3( 11, nextRoot, faceNumber );
	    break;
	  }
	  if ( faceNumber == 6 ) {
	    nextRootOnFace_6( 11, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	  // Racine sur l'arete 12.
	case 12 : 
	  if ( faceNumber == 3 ) {
	    nextRootOnFace_3( 12, nextRoot, faceNumber );
	    break;
	  }
	  if (faceNumber==4) {
	    nextRootOnFace_4( 12, nextRoot, faceNumber );
	    break;
	  }
	  std::cerr << " Erreur :  buildListOfRoots boucle while \n";
	  break;
	  
	}
	
	lastRoot = nextRoot;
	// Insertion de la nouvelle racine dans l'arete. 
	listRootsOnEdges.push_back(nextRoot);
	
      }
      
      return true;
    }


    /*!
      Built the triangles representing surface.
    */
    bool buildTrianglesOnCube(scene<MP,REP3D> & scn) 
    {
      std::vector<int> listRootsOnEdges;
      
      std::vector< vertex_t * > roots;
      
      if ( rootOnEdge_1 != NULL )
      roots.push_back( rootOnEdge_1 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_2 != NULL )
      roots.push_back( rootOnEdge_2 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_3 != NULL )
      roots.push_back( rootOnEdge_3 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_4 != NULL)
      roots.push_back( rootOnEdge_4 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_5 != NULL )
      roots.push_back( rootOnEdge_5 );
      else
      roots.push_back( new vertex_t() );
    
      if ( rootOnEdge_6 != NULL )
      roots.push_back( rootOnEdge_6 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_7 != NULL )
      roots.push_back( rootOnEdge_7 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_8!= NULL )
      roots.push_back( rootOnEdge_8 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_9 != NULL )
      roots.push_back( rootOnEdge_9 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_10 != NULL )
      roots.push_back( rootOnEdge_10 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_11 != NULL )
      roots.push_back( rootOnEdge_11 );
      else
      roots.push_back( new vertex_t() );
      
      if ( rootOnEdge_12 != NULL )
      roots.push_back( rootOnEdge_12 );
      else
      roots.push_back( new vertex_t() );
      
      if ( !buildListOfRoots( listRootsOnEdges ) ){ 
	std::cerr <<  " Erreur : buildTrianglesOnCube, \n";
	return false;
      }
      
      int size = listRootsOnEdges.size();
      
      if ( !listRootsOnEdges.empty() ) {
	if ( listRootsOnEdges[ size - 1 ] != listRootsOnEdges[ 0 ] )
	  std::cerr << " Erreur : buildTrianglesOnCube, chaine non fermee\n";
      }
      
      
      std::vector< vertex_t * > listRoots;
      
      // Construit la liste des racines a partir de la liste des aretes.
      if ( !listRootsOnEdges.empty() )
      for ( int i = 0 ; i < size ; i++ )
      listRoots.push_back( roots[ listRootsOnEdges[ i ] - 1 ] );
      
      buildTriangle(scn, listRoots );
      
      return true;
    }

  
    /*!
      Construction of a whole of triangle from a list of points.
    */
    void buildTriangle(scene<MP,REP3D>& scn, std::vector< vertex_t * > &listRoots ) 
    {
      // Pour la construction d'un triangle 3 points sont necesaires.
      if ( listRoots.size() <= 3 ){
	std::cerr << " Erreur : buildTriangle\n"; 
      }
      
      else {    
	for ( int i = 2 ; i < listRoots.size() - 1 ; i++ ) {
	  
	  triangle_t trg( listRoots[0], listRoots[i-1], listRoots[i] );
	  scn.listTriangles.push_back( trg );
	  
	}
      }
    }
  
  
    /*!
      Indicate the next edge which contains a root, for, an edge 
      containing a root (nextRoot), and a given face (nextFace).
      origin is the number of the current root. 
      nextRoot the next edge on which a root is present.
      nextFace is the next face has to traverse. 
    */
    void nextRootOnFace_1( int origin, int &nextRoot, int &nextFace ) 
    {
      switch( origin ) { 
	case  1 :
	if ( rootOnEdge_3 != NULL ) {
	  nextRoot = 3;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_9 != NULL ) { 
	  nextRoot = 9;
	  nextFace = 6;
	  break;
	}
	if ( rootOnEdge_10 != NULL ) { 
	  nextRoot = 10;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_1 1\n" ;
	break;
	case 3 : 
	if ( rootOnEdge_1 != NULL ) {
	  nextRoot = 1;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_9 != NULL ) {
	  nextRoot = 9;
	  nextFace = 6;
	  break;
	}
	if ( rootOnEdge_10 != NULL ) {
	  nextRoot = 10;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_1 2\n" ;
	break;
	case 9 :
	if ( rootOnEdge_1 != NULL ) {
	  nextRoot = 1;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_3 != NULL) {
	  nextRoot = 3;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_10 != NULL) {
	  nextRoot = 10;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_1 3\n" ;
	break;
	case 10 :
	if ( rootOnEdge_1 != NULL){
	  nextRoot = 1;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_3 != NULL) {
	  nextRoot = 3;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_9 != NULL) {
	  nextRoot = 9;
	  nextFace = 6;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_1 4\n" ;
	break;
	default : std::cerr << " Erreur : nextRootOnFace_1 default\n" ;
      }    
    }
  
    /*!
      Indicate the next edge which contains a root, for, an edge 
      containing a root (nextRoot), and a given face (nextFace).
      origin is the number of the current root. 
      nextRoot the next edge on which a root is present.
      nextFace is the next face has to traverse. 
    */
    void nextRootOnFace_2( int origin, int &nextRoot, int &nextFace ) 
    {
      switch(origin) { 
	case  1 :
	if ( rootOnEdge_2 != NULL ) {
	  nextRoot = 2;
	  nextFace = 3;
	  break;
	}
	if ( rootOnEdge_5 != NULL ) {
	  nextRoot = 5;
	  nextFace = 6;
	  break;
	}
	if ( rootOnEdge_6 != NULL ) {
	  nextRoot = 6;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_2\n" ;
	break;
	case 2 : 
	if ( rootOnEdge_1 != NULL ) {
	  nextRoot = 1;
	  nextFace = 1;
	  break;
	}
	if ( rootOnEdge_5 != NULL ) {
	  nextRoot = 5;
	  nextFace = 6;
	  break;
	}
	if ( rootOnEdge_6 != NULL ) {
	  nextRoot = 6;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_2\n" ;
	break;
	case 5 :
	if ( rootOnEdge_1 != NULL ){
	  nextRoot = 1;
	  nextFace = 1;
	  break;
	} 
	if ( rootOnEdge_2 != NULL ) {
	  nextRoot = 2;
	  nextFace = 3;
	  break;
	} 
	if ( rootOnEdge_6 != NULL ){
	  nextRoot = 6;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_2\n" ;
	break;
	case 6 :
	if ( rootOnEdge_1 != NULL ){
	  nextRoot = 1;
	  nextFace = 1;
	  break;
	}
	if ( rootOnEdge_2 != NULL ) {
	  nextRoot = 2;
	  nextFace = 3;
	  break;
	}
	if ( rootOnEdge_5 != NULL ) {
	  nextRoot = 5;
	  nextFace = 6;
	  break;
	} 
	std::cerr << " Erreur : nextRootOnFace_2\n" ;
	break;
	default : std::cerr << " Erreur : nextRootOnFace_2 default\n" ;
      }    
    }
  
    /*!
      Indicate the next edge which contains a root, for, an edge 
      containing a root (nextRoot), and a given face (nextFace).
      origin is the number of the current root. 
      nextRoot the next edge on which a root is present.
      nextFace is the next face has to traverse. 
    */
    void nextRootOnFace_3( int origin, int &nextRoot, int &nextFace ) 
    {
      switch( origin ) { 
	case  2 :
	if ( rootOnEdge_4 != NULL ) {
	  nextRoot = 4;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_11 != NULL ) {
	  nextRoot = 11;
	  nextFace = 6;
	  break;
	}
	if ( rootOnEdge_12 != NULL ) {
	  nextRoot = 12;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_3 1\n" ;
	break;
	case 4 : 
	if ( rootOnEdge_2 != NULL) {
	  nextRoot = 2;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_11 != NULL ) {
	  nextRoot = 11;
	  nextFace = 6;
	  break;
	}
	if ( rootOnEdge_12 != NULL ) {
	  nextRoot = 12;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_3 2\n" ;
	break;
	case 11 :
	if ( rootOnEdge_2 != NULL ) {
	  nextRoot = 2;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_4 != NULL ){
	  nextRoot = 4;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_12 != NULL ){
	  nextRoot = 12;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_3 3 \n" ;
	break;
	case 12 :
	if ( rootOnEdge_2 != NULL ) {
	  nextRoot = 2;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_4 != NULL ) {
	  nextRoot = 4;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_11 != NULL ){
	  nextRoot = 11;
	  nextFace = 6;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_3 4\n" ;
	break;
	default : std::cerr << " Erreur : nextRootOnFace_3 default\n" ;
      }    
    }
  
    /*!
      Indicate the next edge which contains a root, for, an edge 
      containing a root (nextRoot), and a given face (nextFace).
      origin is the number of the current root. 
      nextRoot the next edge on which a root is present.
      nextFace is the next face has to traverse. 
    */
    void nextRootOnFace_4( int origin, int &nextRoot, int &nextFace ) 
    {
      switch(origin) { 
	case  6 :
	if ( rootOnEdge_8 != NULL ) {
	  nextRoot = 8;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_12 != NULL ) {
	  nextRoot = 12;
	  nextFace = 3;
	  break;
	}
	if ( rootOnEdge_10 != NULL ) {
	  nextRoot = 10;
	  nextFace = 1;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_4 1\n" ;
	break;
	case 8 : 
	if ( rootOnEdge_6 != NULL ) {
	  nextRoot = 6;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_12 != NULL ){
	  nextRoot = 12;
	  nextFace = 3;
	  break;
	} 
	if ( rootOnEdge_10 != NULL ){
	  nextRoot = 10;
	  nextFace = 1;
	  break;
	} 
	std::cerr << " Erreur : nextRootOnFace_4 2 \n" ;
	break;
	case 12 :
	if ( rootOnEdge_6 != NULL ) {
	  nextRoot = 6;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_8 != NULL ){
	  nextRoot = 8;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_10 != NULL ) {
	  nextRoot = 10;
	  nextFace = 1;
	  break;
	} 
	std::cerr << " Erreur : nextRootOnFace_4 3\n" ;
	break;
	case 10 :
	if ( rootOnEdge_6 != NULL ) {
	  nextRoot = 6;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_8 != NULL ){
	  nextRoot = 8;
	  nextFace = 5;
	  break;
	} 
	if ( rootOnEdge_12 != NULL ){
	  nextRoot = 12;
	  nextFace = 3;
	  break;
	}  
	std::cerr << " Erreur : nextRootOnFace_4 4\n" ;
	break;
	default : std::cerr << " Erreur : nextRootOnFace_4 default\n" ;
      }    
    }
  
    /*!
      Indicate the next edge which contains a root, for, an edge 
      containing a root (nextRoot), and a given face (nextFace).
      origin is the number of the current root. 
      nextRoot the next edge on which a root is present.
      nextFace is the next face has to traverse. */
    void nextRootOnFace_5( int origin, int &nextRoot, int &nextFace ) 
    {
      switch( origin ) { 
	case  3 :
	if ( rootOnEdge_4 != NULL) {
	  nextRoot = 4;
	  nextFace = 3;
	  break;
	}
	if ( rootOnEdge_7 != NULL ) {
	  nextRoot = 7;
	  nextFace = 6;
	  break;
	}
	if ( rootOnEdge_8 != NULL ) {
	  nextRoot = 8;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_5\n" ;
	break;
	case 4 : 
	if ( rootOnEdge_3 != NULL ) {
	  nextRoot = 3;
	  nextFace = 1;
	  break;
	}
	if ( rootOnEdge_7 != NULL ){
	  nextRoot = 7;
	  nextFace = 6;
	  break;
	} 
	if ( rootOnEdge_8 != NULL ){
	  nextRoot = 8;
	  nextFace = 4;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_5\n" ;
	break;
	case 7 :
	if ( rootOnEdge_3 != NULL ) {
	  nextRoot = 3;
	  nextFace = 1;
	  break;
	}
	if ( rootOnEdge_4 != NULL ){
	  nextRoot = 4;
	  nextFace = 3;
	  break;
	} 
	if ( rootOnEdge_8 != NULL ){
	  nextRoot = 8;
	  nextFace = 4;
	  break;
	} 
	std::cerr << " Erreur : nextRootOnFace_5\n" ;
	break;
	case 8 :
	if ( rootOnEdge_3 != NULL ){
	  nextRoot = 3;
	  nextFace = 1;
	  break;
	} 
	if ( rootOnEdge_4 != NULL ){
	  nextRoot = 4;
	  nextFace = 3;
	  break;
	}  
	if ( rootOnEdge_7 != NULL ){
	  nextRoot = 7;
	  nextFace = 6;
	  break;
	} 
	std::cerr << " Erreur : nextRootOnFace_5\n" ;
	break;
	default : std::cerr << " Erreur : nextRootOnFace_5 default\n" ;
      }
    }
  
    /*!
      Indicate the next edge which contains a root, for, an edge 
      containing a root (nextRoot), and a given face (nextFace).
      origin is the number of the current root. 
      nextRoot the next edge on which a root is present.
      nextFace is the next face has to traverse. 
    */
    void nextRootOnFace_6( int origin, int &nextRoot, int &nextFace ) 
    {
      switch( origin ) { 
	case  5 :
	if ( rootOnEdge_7 != NULL ) {
	  nextRoot = 7;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_11 != NULL ) {
	  nextRoot = 11;
	  nextFace = 3;
	  break;
	}
	if ( rootOnEdge_9 != NULL) {
	  nextRoot = 9;
	  nextFace = 1;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_6 1\n" ;
	break;
	case 7 : 
	if ( rootOnEdge_5 != NULL) {
	  nextRoot = 5;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_11 != NULL ) {
	  nextRoot = 11;
	  nextFace = 3;
	  break;
	}
	if ( rootOnEdge_9 != NULL ) {
	  nextRoot = 9;
	  nextFace = 1;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_6 2\n" ;
	break;
	case 11 :
	if ( rootOnEdge_5 != NULL ) {
	  nextRoot = 5;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_7 != NULL ) {
	  nextRoot = 7;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_9 != NULL ) {
	  nextRoot = 9;
	  nextFace = 1;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_6 3\n" ;
	break;
	case 9 :
	if ( rootOnEdge_5 != NULL ){
	  nextRoot = 5;
	  nextFace = 2;
	  break;
	}
	if ( rootOnEdge_7 != NULL ) {
	  nextRoot = 7;
	  nextFace = 5;
	  break;
	}
	if ( rootOnEdge_11 != NULL) {
	  nextRoot = 11;
	  nextFace = 3;
	  break;
	}
	std::cerr << " Erreur : nextRootOnFace_6 4\n";
	break;
	default : std::cerr << " Erreur : nextRootOnFace_6 default\n";
      }
    }
  
  };
  
} // end namespace topology

/*********************************************************************/

__END_NAMESPACE_SYNAPS

#endif // axel_mesh_cube3d_H
