/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
/** \file                                                           **/
#ifndef axel_viewer_geomview_mesh_H__
#define axel_viewer_geomview_mesh_H__

/********************************************************************/
#include <synaps/init.h>
#include <synaps/topology/mesh.h>
#include <synaps/base/io/geomview.h>
/*********************************************************************/
__BEGIN_NAMESPACE_SYNAPS 
namespace geomview
{

  /*! 
    \param out the file where we save the list of lines in geomview format.
    \param scn the structure where we have all the data for meshing
               the surface..
    \param with_box allows us to represents the box of
                    regular cells(if with_box == 1)
    \param r the value for the red color.
    \param g the value for the green color.
    \param b the value for the blue color.
    \param a the value for the transparency.

    Print in the file represented by \c out the result of the mesh 
    of an implicit surface.
  */
  template<class MP, class REP3D>
  void
  print( ostream & out, scene<MP,REP3D> & scn, 
	 double r = 0.314, double g = 0.976, double b = 1., 
	 double a = 1., bool with_box =false)
  {
    
    std::ofstream f;
    f.open (out.file, std::ofstream::out | std::ofstream::app);

    f << "{=COFF\n";

    // Number of points.
    f << scn.listVertices.size() <<  " ";
    
    // Number of triangles.
    std::cout << " Nb of triangles: " <<   scn.listTriangles.size() << std::endl;
    std::cout << " Nb of points   : " <<   scn.listVertices.size() << std::endl;
    f << scn.listTriangles.size() << std::endl;
    
    // Edge numbers.
    f << "0"  << std::endl;
    
    // List of points.
    for ( int i = 0 ; i < scn.listVertices.size() ; i++ )
      f <<  scn.x+scn.u*scn.listVertices[i]->point[0] << " "
	<<  scn.y+scn.v*scn.listVertices[i]->point[1] << " "
	<<  scn.z+scn.w*scn.listVertices[i]->point[2] << " "
	<< r << " " << g << " " << b << " " << a  <<"\n";
    
    // The triangles and their indices.
    for ( int j = 0 ; j < scn.listTriangles.size() ; j++ )
      f <<  "3" << " "
	<< scn.listTriangles[j].a->index << " "
	<< scn.listTriangles[j].b->index << " "
	<< scn.listTriangles[j].c->index << std::endl;
    
    f <<"}\n";

    if (with_box)
      {
	std::cout <<" Nb of regular  cells: "
		  <<scn.listRegCells.size()/2<<std::endl;
	printStudiedCells(f, scn.listRegCells,
			  scn.x,scn.y,scn.z,scn.u,scn.v,scn.w);
      }
    if ( scn.SingCells.size() != 0 ) {
      std::cout << " Nb of singular cells: "
		<<  scn.SingCells.size()/2  << std::endl;
      printNotStudiedCells(f, scn.SingCells, 1,0,0,
			   scn.x,scn.y,scn.z,scn.u,scn.v,scn.w);
    }
    if ( scn.XRegCells.size() != 0 ) {
      std::cout << " Nb of non-treated x-regular cells: "
		<<  scn.XRegCells.size()/2  << std::endl;
      printNotStudiedCells(f, scn.XRegCells, 0,1,1,
			   scn.x,scn.y,scn.z,scn.u,scn.v,scn.w);
    }
    if ( scn.YRegCells.size() != 0 ) {
      std::cout << " Nb of non-treated y-regular cells: "
		<<  scn.YRegCells.size()/2  << std::endl;
      printNotStudiedCells(f, scn.YRegCells, 1,0,1,
			   scn.x,scn.y,scn.z,scn.u,scn.v,scn.w);
    }
    if ( scn.ZRegCells.size() != 0 ) {
      std::cout << " Nb of non-treated z-regular cells: "
		<<  scn.ZRegCells.size()/2  << std::endl;
      printNotStudiedCells(f, scn.ZRegCells, 1,1,0,
			   scn.x,scn.y,scn.z,scn.u,scn.v,scn.w);
    }

    f.close(); 
  }

  //--------------------------------------------------------------------
  /*! 
    Safeguard in a file the cells which don't pose problem.
  */
  template < class C >
  void printStudiedCells( std::ofstream & f,
			  std::vector<topology::point<C> > &listCells,
			  double x=0, double y=0, double z=0,
			  double u=1, double v=1, double w=1)
  {
    
    if  (!listCells.empty() ) 
      {
	int size = (int) listCells.size();

	// Format du fichier pour Geomview.
	f << "\n{VECT\n";
	
	// Il y a 12 aretes par cellules.
	f << size*6 << " " 
	  << size*12  << " " << 1 << std::endl;
	
      // nombre de points constituant une suite de segments.
	for ( int l = 0 ; l < size*6 ; ++l ) 
	  f << "2 ";
	f << std::endl;
	
	// Pour la couleur.
	f << 1 << " ";
	for ( int m= 0 ; m < size*6 - 1 ; ++m ) 
	  f << "0 ";
	f << std::endl;
	
	for ( int i= 0 ; i < size-1 ; i = i + 2  ) 
	  {
	    C x0 = x+u*listCells[i][0] ;
	    C y0 = y+v*listCells[i][1] ;
	    C z0 = z+w*listCells[i][2] ;
	    
	    C x1 = x+u*listCells[i+1][0] ;
	    C y1 = y+v*listCells[i+1][1] ;
	    C z1 = z+w*listCells[i+1][2] ;
	    
	    // Construire la boite...
	    
	    f << x0 << " " << y0 << " " << z0 << "\t";
	    f << x0 << " " << y1 << " " << z0 << std::endl;
	    
	    f << x0 << " " << y0 << " " << z1 << "\t";
	    f << x0 << " " << y1 << " " << z1 << std::endl;
	    
	    f << x1 << " " << y0 << " " << z0 << "\t";
	    f << x1 << " " << y1 << " " << z0 << std::endl;
	    
	    f << x1 << " " << y0 << " " << z1 << "\t";
	    f << x1 << " " << y1 << " " << z1 << std::endl;
	    
	    f << x0 << " " << y0 << " " << z0 << "\t";
	    f << x0 << " " << y0 << " " << z1 << std::endl;
	    
	    f << x0 << " " << y1 << " " << z0 << "\t";
	    f << x0 << " " << y1 << " " << z1 << std::endl;
	  
	    f << x1 << " " << y0 << " " << z0 << "\t";
	    f << x1 << " " << y0 << " " << z1 << std::endl;
	    
	    f << x1 << " " << y1 << " " << z0 << "\t";
	    f << x1 << " " << y1 << " " << z1 << std::endl;
	    
	    f << x0 << " " << y0 << " " << z0 << "\t";
	    f << x1 << " " << y0 << " " << z0 << std::endl;
	  
	    f << x0 << " " << y1 << " " << z0 << "\t";
	    f << x1 << " " << y1 << " " << z0 << std::endl;
	    
	    f << x0 << " " << y0 << " " << z1 << "\t";
	    f << x1 << " " << y0 << " " << z1 << std::endl;
	    
	    f << x0 << " " << y1 << " " << z1 << "\t";
	    f << x1 << " " << y1 << " " << z1 << std::endl;
	    
	  }
	
	f << 0 << " " << 0.61 << " " << 0 << " " << 1 ;
      }
    f<<"}\n";
  }

  //--------------------------------------------------------------------
  /*!
    Safeguard in a file the cells which pose problem.
  */
  template < class C >
  void printNotStudiedCells( std::ofstream & f, 
			     std::vector<topology::point< C > > &listCells,
			     double red=1, double green=0, double blue=0,
			     double x=0, double y=0, double z=0,
			     double u=1, double v=1, double w=1
			     )
  {
    int size = (int) listCells.size();

    // Format du fichier pour Geomview.
    f << "\n{=VECT\n";
    
    // Il y a 12 aretes par cellules.
    f << size*6 << " " << size*12  << " " << 1 << std::endl;
    
    // nombre de points constituant une suite de segments.
    for ( int l = 0 ; l < size*6 ; ++l ) 
      f << "2 ";
    f << std::endl;
  
    // Pour la couleur.
    f << 1 << " ";
    for ( int m = 0 ; m < size*6 - 1 ; ++m ) 
      f << "0 ";
    f << std::endl;
  
    for ( int i = 0 ; i< size - 1 ; i = i + 2  ) 
      {
	C x0 = x+u*listCells[i][0] ;
	C y0 = y+v*listCells[i][1] ;
	C z0 = z+w*listCells[i][2] ;
	
	C x1 = x+u*listCells[i+1][0] ;
	C y1 = y+v*listCells[i+1][1] ;
	C z1 = z+w*listCells[i+1][2] ;	
	
	// Construire la boite...      
	f << x0 << " " << y0 << " " << z0 << "\t";
	f << x0 << " " << y1 << " " << z0 << std::endl;
	
	f << x0 << " " << y0 << " " << z1 << "\t";
	f << x0 << " " << y1 << " " << z1 << std::endl;
	
	f << x1 << " " << y0 << " " << z0 << "\t";
	f << x1 << " " << y1 << " " << z0 << std::endl;
	
	f << x1 << " " << y0 << " " << z1 << "\t";
	f << x1 << " " << y1 << " " << z1 << std::endl;
	
	f << x0 << " " << y0 << " " << z0 << "\t";
	f << x0 << " " << y0 << " " << z1 << std::endl;
	
	f << x0 << " " << y1 << " " << z0 << "\t";
	f << x0 << " " << y1 << " " << z1 << std::endl;
	
	f << x1 << " " << y0 << " " << z0 << "\t";
	f << x1 << " " << y0 << " " << z1 << std::endl;
	
	f << x1 << " " << y1 << " " << z0 << "\t";
	f << x1 << " " << y1 << " " << z1 << std::endl;
	
	f << x0 << " " << y0 << " " << z0 << "\t";
	f << x1 << " " << y0 << " " << z0 << std::endl;
	
	f << x0 << " " << y1 << " " << z0 << "\t";
	f << x1 << " " << y1 << " " << z0 << std::endl;
	
	f << x0 << " " << y0 << " " << z1 << "\t";
	f << x1 << " " << y0 << " " << z1 << std::endl;
	
	f << x0 << " " << y1 << " " << z1 << "\t";
	f << x1 << " " << y1 << " " << z1 << std::endl;
      }
    
    f << red << " " << green << " " << blue << " "<<1 ;
    f<<"}\n";
  }

}

/*********************************************************************/
__END_NAMESPACE_SYNAPS 
#endif // axel_viewer_geomview_mesh_H__
