/*********************************************************************
*      This file is part of the source code of SYNAPS library          *
*   Author(s): E. Tobis                                              *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/

#ifndef axel_mesh_miscellaneous_H__
#define axel_mesh_miscellaneous_H__

#include <synaps/init.h>

#include <vector>
#include <synaps/base/Seq.h>
#include <synaps/topology/mesh/cell3d.h>

__BEGIN_NAMESPACE_SYNAPS

template < class MP ,
           class REP3D>
struct cell3d;
// This function returns a vector which holds the indices of the edges
// corresponding to a face. They are return in clockwise, if we look at
// the face from inside the cube. The first edge of each face is relevant
// for the function nextRoot
inline void edgesOfAFace(unsigned face,std::vector<unsigned> & result)
{
  result.resize(4);

  switch (face) {
  case 1:
    result[0] = 0;result[1] = 9;result[2] = 2;result[3] = 8;
    break;
  case 2:
    result[0] = 4;result[1] = 1;result[2] = 6;result[3] = 0;
    break;
  case 3:
    result[0] = 10;result[1] = 3;result[2] = 11;result[3] = 1;
    break;
  case 4:
    result[0] = 6;result[1] = 11;result[2] = 7;result[3] = 9;
    break;
  case 5:
    result[0] = 2;result[1] = 7;result[2] = 3;result[3] = 5;
    break;
  case 6:
    result[0] = 8;result[1] = 5;result[2] = 10;result[3] = 4;
    break;
  }
}

// This function modifies edge and root, to make them point to the
// next root, in clockwise order, for the received face.
// We assume there is at least one root on the face.
// Edges run from 0-3 (index into edgesOfAFace) and faces from 1-6
template <class C>
void nextRoot(const Seq<Seq<C> > roots,unsigned & edge,
	      int & root,unsigned face)
{

  static int counter = 0;
  counter++;

  std::vector<unsigned> edges;
  edgesOfAFace(face,edges);

  switch (edge) {
  case 0:
    // We go up in the current edge, or move to the first root of the 
    // following one.
    
    if (root >= (int(roots[edges[edge]].size()) - 1)) {
      edge++;
      root = 0;
      if (roots[edges[edge]].size() == 0) 
	// if we don't have roots on the new edge
	nextRoot(roots,edge,root,face);
    } else 
      root++;
    break;
  case 1:
    // We go up in the current edge, or move to the last root of the
    // following one.
    
    if (root >= int(roots[edges[edge]].size()) - 1) {
      edge++;
      root = roots[edges[edge]].size() - 1;
      if (roots[edges[edge]].size() == 0)
	// if we don't have roots on the new edge
	nextRoot(roots,edge,root,face);
    } else
      root++;
    break;
  case 2:
    // We go down in the current edge, or move to the last root of the
    // following one.

    if (root <= 0) {
      edge++;
      root = roots[edges[edge]].size() - 1;
      if (roots[edges[edge]].size() == 0)
	// if we don't have roots on the new edge
	nextRoot(roots,edge,root,face);
    } else
      root--;
    break;
  case 3:
    if (root <= 0) {
      edge = 0;
      root = 0;
      if (roots[edges[edge]].size() == 0)
	// if we don't have roots on the new edge
	nextRoot(roots,edge,root,face);
    } else
      root--;
    break;
  default:
    std::cerr << "Wrong edge number: " << edge
	      << ". Must be 0-3." << std::endl;
    assert(edge >= 0 && edge <= 3);
    break;
  }
}
inline 
void variablesOfAFace(unsigned face,Seq<unsigned> & vars)
{
  vars.resize(2);

  switch (face) {
  case 1:case 3:
    vars[0] = 0;vars[1] = 1;
    break;
  case 2: case 5:
    vars[0] = 0;vars[1] = 2;
    break;
  case 4: case 6:
    vars[0] = 1;vars[1] = 2;
    break;
  default:
    std::cerr << "Wrong face number: " << face
	      << ". Must be 1-6." << std::endl;
  }
}

template <class C,class MP,class REP3D>
void convertRootToPoint(const cell3d<MP,REP3D> & cell,
			C root,unsigned edge,
			topology::point<C> & p)
{
  p.resize(3);
  switch (edge) {
  case 1:
    p[0]  = root;
    p[1] = cell.box.y0;
    p[2] = cell.box.z0;
    break;
  case 2:
    p[0] = root;
    p[1] = cell.box.y0;
    p[2] = cell.box.z1;
    break;
  case 3:
    p[0] = root;
    p[1] = cell.box.y1;
    p[2] = cell.box.z0;
    break;
  case 4:
    p[0] = root;
    p[1] = cell.box.y1;
    p[2] = cell.box.z1;
    break;
  case 5:
    p[0] = cell.box.x0;
    p[1] = cell.box.y0;
    p[2] = root;
    break;
  case 6:
    p[0] = cell.box.x0;
    p[1] = cell.box.y1;
    p[2] = root;
    break;
  case 7:
    p[0] = cell.box.x1;
    p[1] = cell.box.y0;
    p[2] = root;
    break;
  case 8:
    p[0] = cell.box.x1;
    p[1] = cell.box.y1;
    p[2] = root;
    break;
  case 9:
    p[0] = cell.box.x0;
    p[1] = root;
    p[2] = cell.box.z0;
    break;
  case 10:
    p[0] = cell.box.x1;
    p[1] = root;
    p[2] = cell.box.z0;
    break;
  case 11:
    p[0] = cell.box.x0;
    p[1] = root;
    p[2] = cell.box.z1;
    break;
  case 12:
    p[0] = cell.box.x1;
    p[1] = root;
    p[2] = cell.box.z1;
    break;
  default:
    std::cerr << "Wrong edge number: " << edge << ". Must be 1-12"
	      << std::endl;
    assert(edge >= 1 && edge <= 12);
  }
}

// This function takes a 3d point, and returns a "stripped" point:
// we loose the coordinate not relevant for the face we are processing
template <class C>
void dropExtraCoordinate(const topology::point<C> & p,
			 unsigned face,
			 topology::point<C> & result)
{
  result.resize(2);

  switch (face) {
  case 1:case 3: // We drop z
    result[0] = p[0];
    result[1] = p[1];
    break;
  case 2:case 5: // We drop y
    result[0] = p[0];
    result[1] = p[2];
    break;
  case 4:case 6: // We drop x
    result[0] = p[1];
    result[1] = p[2];
    break;
  default:
    std::cerr << "Wrong face number: " << face << ". Must be 1-6."
	      << std::endl;
    assert(face >= 1 && face <= 6);
  }
}

__END_NAMESPACE_SYNAPS

#endif
