/*********************************************************************
*      This file is part of the source code of SYNAPS library          *
*   Author(s): G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
/** \file                                                           **/
#ifndef synaps_mesh_scene_H
#define synaps_mesh_scene_H

/*********************************************************************/

#include <synaps/init.h>

#include <synaps/topology/point.h>
#include <synaps/topology/vertex.h>
#include <synaps/topology/triangle.h>
#include <synaps/topology/Octree.h>
#include <synaps/topology/mesh/cell3d.h>

#include <synaps/mpol.h>
#include <synaps/mpol/MPOLDST.m>
#include <synaps/linalg/rep1d.h>
#include <synaps/linalg/rep3d.h>
#include <synaps/usolve/bezier/Std.h>

#include <vector>

__BEGIN_NAMESPACE_SYNAPS

/*********************************************************************/
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
template<class MP, class REP3D > struct cell3d;

namespace topology 
{ 
  template<class C > struct vertex; 
  template<class C > struct triangle; 
}
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////


/*********************************************************************/


/*!
 */
template<class MP = MPol<double> , 
         class REP3D = linalg::rep3d< typename MP::coeff_t > >
struct scene 
{
  typedef typename MP::coeff_t  coeff_t;

  coeff_t x,y,z,u,v,w;

  /*! The Octree. */
  OCTREE::Octree< cell3d< MP, REP3D > > octree;

  std::vector< OCTREE::Node < cell3d< MP, REP3D > > * > listOfNode;

  /*! List of all the vertex constituting the mesh. */
  std::vector< topology::vertex< topology::point<coeff_t> >* > listVertices;

  /*! List of triangles that define a mesh. */
  std::vector< topology::triangle< topology::point<coeff_t> >  > listTriangles;

  std::vector< topology::point< coeff_t > > SingCells;
  std::vector< topology::point< coeff_t > > XRegCells;
  std::vector< topology::point< coeff_t > > YRegCells;
  std::vector< topology::point< coeff_t > > ZRegCells;
  std::vector< topology::point< coeff_t > > listRegCells;

  scene():
    x(0),y(0),z(0),u(1),v(1),w(1), 
    octree(), listOfNode(), listVertices(), listTriangles(),
    SingCells(), XRegCells(), YRegCells(), ZRegCells(),
    listRegCells() {}

};

/*!
  Function that inserts a vertex in the list of vertices of 
  the mesh.
*/
template <class MP, class REP3D>
void 
insertVertexInList( scene<MP,REP3D> & scn,
		    topology::vertex< topology::point<typename MP::coeff_t> > * Vertex )
{
  if ( scn.listVertices.empty() )
    Vertex->index = 0;
  else 
    Vertex->index = scn.listVertices.size();
  scn.listVertices.push_back( Vertex );
}

/*!
  Build a triangle with vertices given as parameter. 
*/
template <class MP, class REP3D>
void makeTriangle(
		  scene<MP,REP3D> & scn,
		  topology::vertex< topology::point<typename MP::coeff_t> > * v1,  
		  topology::vertex< topology::point<typename MP::coeff_t> > * v2, 
		  topology::vertex< topology::point<typename MP::coeff_t> > * v3 ) 
{
  topology::triangle< topology::point<typename MP::coeff_t> > triangle( v1, v2, v3 );
  scn.listTriangles.push_back( triangle );
}

/*********************************************************************/
__END_NAMESPACE_SYNAPS
#endif // synaps_mesh_scene_H
