/*********************************************************************
*      This file is part of the source code of SYNAPS library          *
*   Author(s): G. Gatellier                                          *
*              A. Labrouzy                                           *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef synaps_topology_point_H
#define synaps_topology_point_H
/*********************************************************************/
#include <vector>
#include <synaps/init.h>
#include <synaps/base/shared_object.h>
//#include <synaps/arithm/QQ.h>
//#include <synaps/arithm/RR.h>
/*********************************************************************/
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
//! Container for the topology domain
/*!
  Here, we have implemented structures which are useful for the 
  algorithms, in order to return a result. In this case, we can reuse
  the result with another algorithm.
  For instance, in this namespace, we have the structure points and
  the structure graph of points which are useful in the topology
  algorithms.
*/
namespace topology
{
  //! The topology point structure
  /*!
    An topology point is represented by its coordinates.

    \li \c C is the type of coordinates.\n
    \li \c R is a type storing all the coordinates of the point.\n

    \remarks This structure doesn't depend of the dimension.\n
    \remarks By default, in order to store the coordinates in all 
    dimensions, we use \c shared_object<std::vector<C> >.
    \htmlonly <A HREF="point.xml"> For further explanations, 
    we can see the page for this structure in the
    documentation.</A> \endhtmlonly
  */
  template<class C, class R = shared_object<std::vector<C> > >
  struct point 
  {
    typedef unsigned size_type;
    typedef typename ReferTo<R>::value_type::iterator        iterator;
    typedef typename ReferTo<R>::value_type::const_iterator  const_iterator;

    /*! The type of the coefficients. */
    typedef typename ReferTo<C>::value_type coeff_t;

    /*!
      The variable which represents the coordinates of the point.
     */ 
    R data;

    /*! The default constructor. */
    point():data() {}

    /*! The default destructor. */
    ~point() {}

    /*!
      \param r it represents all the coordinates of the point.

      Construct an topology point by giving all its coordinates.
    */
    point(const R & r): data(r) {}
    
    /*! The copy constructor. */
    point(const point<C,R> & P): data(P.data) {}

    /*!
      \param d the dimension of the point.
      \param t it represents all the coordinates of the point.

      Construct an topology point by giving its dimensional \c d, and 
      a table \c t with its coordinates.
    */
    point(unsigned d, C* t):data() 
      {
	rep(data).resize(d);
	iterator it = rep(data).begin();
	C* v=t;
	for(unsigned i = 0; i < d; ++i) *it++=*v++;
      }

    /*!
      Construct an topology point by giving its dimensional \c d ,
      and a char* \c str with its coordinates.
    */
    point(unsigned d, char *str):data() 
      {
	rep(data).resize(d);
	std::istringstream ins(str);
	for(unsigned i = 0; i < d; ++i) ins >> rep(data)[i];
      }

    /*! Construction of a point of size \c d. */
    point(unsigned d):data() 
      {
	rep(data).resize(d);
      }

    /*!
      \param b the first element in the table.
      \param e the last element in the table.

      Construct an topology point by giving its first element \c b
      and the last element \c e in a table.
    */
    point(const C* b, const C* e):data() 
      {
	rep(data).resize(std::distance(b,e));
	iterator it = rep(data).begin(); 
	const C* tt = b;
	for(;tt != e; ) *it++ = *tt++;
      }

    /*!
      Construct an topology point in 2-dimensional by giving
      its 2 coordinates.
    */
    point(const coeff_t & x, const coeff_t & y)
      {
	rep(data).resize(2);
	rep(data)[0] = x;
	rep(data)[1] = y;
      }
    
    /*!
      Construct an topology point in 3-dimensional by giving
      its 3 coordinates.
    */
    point(const coeff_t & x, const coeff_t & y, const coeff_t & z)
      {
	rep(data).resize(3);
	rep(data)[0] = x;
	rep(data)[1] = y;
	rep(data)[2] = z;
      }
    
    /*!
      \return The dimension of the point.
    */
    typename ReferTo<R>::value_type::size_type size() const   
      { return rep(data).size(); }


    iterator       begin()   {return rep(data).begin();}
    const_iterator begin() const  {return rep(data).begin();}
    iterator       end()   {return rep(data).end();}
    const_iterator end() const  {return rep(data).end();}


    /*!
      \param d the new size of the point.

      Resize the size of the point.
    */
    void resize(int d)   
      { rep(data).resize(d); }
    
    /*! Access to the \c operator[] */
    C   operator[](unsigned i) const 
      {if(i<size()) return (rep(data))[i]; else return C(0); }

    /*! The \c operator[] */
    C & operator[](unsigned i)       {assert(i<size()); return (rep(data))[i];}
    
    /*! The \c operator== */
    bool operator==(const point<C,R>& p) const {return this == &p;}

    /*! The \c operator!= */
    bool operator!=(const point<C,R>& p) {return !(rep(data)==rep(p.data));}

    /*! The \c operator= */
    point<C,R> operator=(const point<C,R>& p)
      {
	rep(data).resize(rep(p.data).size());
	for(int i = 0; i < (int)rep(data).size(); ++i)
	  rep(data)[i] = rep(p.data)[i];	
	return (*this);
      }

    /*! The \c operator+ */
    point<C,R> operator+(point<C,R>& p) const 
      {
	assert(rep(data).size() == rep(p.data).size());
	C* tab = new C[rep(data).size()];
	for(int i = 0; i < rep(data).size(); ++i)
	  rep(tab)[i] = rep(data)[i]+rep(p.data)[i];
	return point(rep(data).size(),rep(tab));
      }

    /*! The \c operator+= */
    point<C,R> operator+=(point<C,R>& p)
      {
	assert(rep(data).size() == rep(p.data).size());
	for(int i = 0; i < rep(data).size(); ++i)
	  rep(data)[i] += rep(p.data)[i];
	return *this;
      }

    /*! The \c operator- */
    point<C,R> operator-(point<C,R>& p) const 
      {	
	assert(rep(data).size() == rep(p.data).size());
	C* tab = new C[rep(data).size()];
	for(int i = 0; i < rep(data).size(); ++i)
	  rep(tab)[i] = rep(data)[i]-rep(p.data)[i];
	return point(rep(data).size(),rep(tab));
      }

    /*! The \c operator-= */
    point<C,R> operator-=(point<C,R>& p)
      {
	assert(rep(data).size() == rep(p.data).size());
	for(int i = 0; i < rep(data).size(); ++i)
	  rep(data)[i] -= rep(p.data)[i];
	return *this;
      }

    /*! The unary \c operator- */
    point<C,R> operator-() const 
      {
	C* tab = new C[rep(data).size()];
	for(int i = 0; i < rep(data).size(); ++i)
	  rep(tab)[i] = -rep(data)[i];
	return point(rep(data).size(),rep(tab));
      }

    /*! The unary \c operator+ */
    point<C,R> operator+() const {return this;}

    /*! The \c operator* */
    point<C,R> operator*(C cte) const 
      {
	C* tab = new C[rep(data).size()];
	for(int i = 0; i < rep(data).size(); ++i)
	  rep(tab)[i] = rep(data)[i]*cte;
	return point(rep(data).size(),rep(tab));
      }

    /*! The \c operator/ */
    point<C,R> operator/(C cte) const 
      {
	C* tab = new C[rep(data).size()];
	for(int i = 0; i < rep(data).size(); ++i)
	  rep(tab)[i] = rep(data)[i]/cte;
	return point(rep(data).size(),rep(tab));
      }

    /*! The left \c operator* */
    friend point<C,R> operator*(C cte, point<C,R>& p) {return p*cte;}

    /*! The left \c operator/ */
    friend point<C,R> operator/(C cte, point<C,R>& p) {return p/cte;}

    /*!
      \param p a point.

      The \c operator< allows us to sort a list of points
      (p1<p2  according the lexicographic order).
    */
    bool operator<(const point<C,R>& p) const
      {                              
	C eps = 1e-5;
	for(int j = 0; j < (int)rep(data).size(); ++j)
	  {
	    if (rep(p.data)[j]>rep(data)[j]+eps)
	      return true;
	    else if(rep(p.data)[j]<rep(data)[j]-eps)
	      return false;
	  }
	return false;
      }
    
  };

  /*!
    \param p1 a point.
    \param p2 a point.
    
    Compare the position of two points.
    \return
    \li -1 if p1 is under p2
    \li  0 if p1 == p2
    \li  1 otherwise
  */
  template<class C,class R>
  int 
  ptcmp(const topology::point<C,R> & p1, const topology::point<C,R> & p2)
  {
    assert( p1.size() == p2.size() );
    for(int i = 0; i < (int)p1.size(); ++i)
      {
	if ( (rep(p1.data)[i] > rep(p2.data)[i]))
	  return 1;
	else if ( (rep(p1.data)[i] < rep(p2.data)[i]))
	  return -1;
      }
    return 0;
  }

#ifdef synaps_HAVE_LIBGMP
  /*!
    The \c operator \c << .\n
    This operator allows us to display a point.
  */
  template<class R>
  std::ostream & 
  operator<<(std::ostream & os, const topology::point<QQ,R>& P)
  {
    os <<"Point(";
    if((int)rep(P.data).size())
      {
	int i=0;
	os<<P[0].get_d();
	for(int i = 1; i < (int)rep(P.data).size(); ++i)
	  os <<","<< P[i].get_d();
      }
    os <<")";
   return os;
  }

  /*!
    The \c operator \c << .\n
    This operator allows us to display a point.
  */
  template<class R>
  std::ostream & 
  operator<<(std::ostream & os, const topology::point<RR,R>& P)
  {
    os <<"Point(";
    if((int)rep(P.data).size())
      {
	int i=0;
	os<<P[0].get_d();
	for(int i = 1; i < (int)rep(P.data).size(); ++i)
	  os <<","<< P[i].get_d();
      }
    os <<")";
   return os;
  }

#endif // synaps_HAVE_LIBGMP

  /*!
    The \c operator \c << .\n
    This operator allows us to display a point.
  */
  template<class C, class R>
  std::ostream & 
  operator<<(std::ostream & os, const topology::point<C,R>& P)
  {
    os <<"Point(";
    if((int)rep(P.data).size())
      {
	os<<P[0];
	for(int i = 1; i < (int)rep(P.data).size(); ++i)
	  os <<","<< P[i];
      }
    os <<")";
    return os;
  }
  
  /*!
    The \c operator \c >> .\n
    This operator allows us to read a point.
  */
  template<class C, class R>
  std::istream &
  operator>>(std::istream & is, topology::point<C,R>& P)
  {
    rep(P.data).resize(3);
    is>>P[0]>>P[1]>>P[2];
    return is;
  }

}


//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
/*********************************************************************/
#endif // synaps_topology_point_H
