


#include <synaps/arithm/gmp.h>
//#include <synaps/usolve/sturm.h>
#include <synaps/base/Rand.h>
#include <synaps/base/Seq.h>
#include <synaps/upol/UPolDse.h>
#include <synaps/upol/xml.h>

#include <fstream>

using std::cout;
using std::cin;
using std::endl;

using SYNAPS::ZZ;
using SYNAPS::QQ;

typedef SYNAPS::UPolDse<ZZ>   Poly;
typedef SYNAPS::Seq<Poly>     Seq;


template < typename T >
std::string to_string(T num)
{
  std::stringstream ss;
  ss << num;
  return ss.str();
}



__BEGIN_NAMESPACE_SYNAPS

size_t bitsize( const ZZ& num)
{
  return mpz_sizeinbase( num.get_mpz_t(), 2);
}



class RandomPoly {
public:
  typedef UPolDse<ZZ>    Poly;

protected:
  ZZ m;
  ZZ M;
  Rand<ZZ>  RND;  

public:
  RandomPoly( unsigned int nb_bits = 10)
    : m( SYNAPS::pow( ZZ(2), nb_bits - 1) ), 
      M( m * 2),
      RND()
  {
  }

  RandomPoly( unsigned int nb_bits, unsigned long seed)
    : m( SYNAPS::pow( ZZ(2), nb_bits - 1) ), 
      M( m * 2),
      RND( m, M, seed)
  {
  }


  size_t get_size() const
  {
    return mpz_sizeinbase( M.get_mpz_t(), 2);
  }
  
  void set_size( unsigned int nb_bits)
  {
    m = SYNAPS::pow( ZZ(2), nb_bits - 2);
    M = m * 2; 
  }

  virtual std::string description() const 
  {
    return "Polynomials with random coefficients";
  }

  virtual std::string code() const 
  {
    return "random";
  }
  

  virtual Poly produce( unsigned int degree)
  {
    if ( degree  == 0) return Poly();

    Poly f( degree+1, AsSize());
      
    for (int i = 0; i <= degree; ++i) {
      f[i] = RND( m, M) * SYNAPS::pow( -1, RND.get_rand_bool());
    }
    // make sure that lcoeff(f) <> 0
    while (f[degree] == 0)
      f[degree] = RND( m, M);
    // make sure lcoeff(f) > 0
    if (f[degree] < 0) f = -f;
    return f;
  }
  
};


class MultipleRandomPoly : public RandomPoly 
{
public:
  MultipleRandomPoly( unsigned int nb_bits = 10)
    : RandomPoly( nb_bits / 2)
  {
  }

  MultipleRandomPoly( unsigned int nb_bits, unsigned long seed)
    : RandomPoly( nb_bits / 2, seed)
  {
  }

  std::string code() const 
  {
    return "random_2";
  }

  std::string description() const 
  {
    return "Polynomials with random coefficients and multiple roots";
  }
  

  Poly produce( unsigned int degree)
  {
    
    int d = static_cast<int>( floor( degree / 2.));
    
    Poly f = this->RandomPoly::produce( d);


    if ( degree % 2 == 0) {
      return f * f;
    } else {
      return f * f * Poly("7*x-22");
    }
  }
};



class RandomIntRootsPoly : public RandomPoly 
{
public:
  RandomIntRootsPoly( unsigned int nb_bits = 10)
    : RandomPoly( nb_bits)
  {
  }

  RandomIntRootsPoly( unsigned int nb_bits, unsigned long seed)
    : RandomPoly( nb_bits, seed)
  {
  }

  std::string code() const 
  {
    return "introots";
  }

  std::string description() const 
  {
    return "Polynomials with integer real roots";
  }
  

  Poly produce( unsigned int degree)
  {
    if (degree == 0) return Poly();
    
    size_t sz = get_size();
    size_t new_sz = (sz / (degree)) +1;
    if (new_sz < 2) new_sz = 2;
    set_size( new_sz);

    Poly f(1, AsSize());
    f[0] = 1;
    for (unsigned i = 0; i < degree; ++i) {
      Poly t(2, AsSize());
      t[0] = RND( m, M) * SYNAPS::pow( -1, RND.get_rand_bool());
      t[1] = 1;
      f *= t;
    }
      
    f = RND( m, M) * f;
      
    if (f[degree] < 0) f = -f;
    set_size(sz);
    return f;
  }
};




class RandomRatRootsPoly : public RandomPoly 
{
public:
  RandomRatRootsPoly( unsigned int nb_bits = 10)
    : RandomPoly( nb_bits)
  {
  }

  RandomRatRootsPoly( unsigned int nb_bits, unsigned long seed)
    : RandomPoly( nb_bits, seed)
  {
  }

  std::string code() const 
  {
    return "ratroots";
  }

  std::string description() const 
  {
    return "Polynomials with rational real roots";
  }
  
  Poly produce( unsigned int degree)
  {
    if (degree == 0) return Poly();
      
    size_t sz = bitsize(M);
    size_t new_sz = (sz / (degree)) +1;
    if (new_sz < 2) new_sz = 2;
    set_size( new_sz);

    Poly f(1, AsSize());
    f[0] = 1;
    for (unsigned i = 0; i < degree; ++i) {
      Poly t(2, AsSize());
      t[0] = RND( m, M) * SYNAPS::pow( -1, RND.get_rand_bool());
      t[1] = RND( m, M);
      f *= t;
    }
      
    f = RND( m, M) * f;
      
    if (f[degree] < 0) f = -f;
    set_size(sz);
      
    return f;
  }
};


class MignottePoly : public RandomPoly 
{
public:
  MignottePoly( unsigned int nb_bits = 10)
    : RandomPoly( nb_bits)
  {
  }

  MignottePoly( unsigned int nb_bits, unsigned long seed)
    : RandomPoly( nb_bits, seed)
  {
  }

  std::string code() const 
  {
    return "mignotte";
  }

  std::string description() const 
  {
    return "Mignotte Polynomials";
  }
  

  virtual Poly produce( unsigned int degree)
  {
    if (degree < 3) return Poly();

    Poly f(1, degree);
    Poly t(2, AsSize());
    ZZ a = RND( m, M);
    t[0] = ZZ(-1);
    t[1] = a;
    f = f - t * t * 2;
      
    f = RND( m, M) * f;
      
    if (f[degree] < 0) f = -f;
    return f;
  }
};



class MignottePoly_2 : public MignottePoly 
{
public:
  MignottePoly_2( unsigned int nb_bits = 10)
    : MignottePoly( nb_bits / 2)
  {
  }

  MignottePoly_2( unsigned int nb_bits, unsigned long seed)
    : MignottePoly( nb_bits / 2, seed)
  {
  }

  std::string code() const 
  {
    return "mignotte_2";
  }

  std::string description() const 
  {
    return "Two Mignotte Polynomials";
  }
  

  virtual Poly produce( unsigned int degree)
  {

    if (degree < 6) {
      return this->MignottePoly::produce( degree);
    }
  
    int d1 = static_cast<int>( floor( degree / 2.));
    int d2 = static_cast<int>( ceil( degree / 2.));
    
    Poly f1 = this->MignottePoly::produce( d1);
    Poly f2 = this->MignottePoly::produce( d2);

    return f1 * f2;
  }
};


class MultipleMignottePoly : public MignottePoly 
{
public:
  MultipleMignottePoly( unsigned int nb_bits = 10)
    : MignottePoly( nb_bits / 2)
  {
  }

  MultipleMignottePoly( unsigned int nb_bits, unsigned long seed)
    : MignottePoly( nb_bits / 2, seed)
  {
  }

  std::string code() const 
  {
    return "mignotte_3";
  }

  std::string description() const 
  {
    return "Mignotte Polynomials with multiple roots";
  }
  

  virtual Poly produce( unsigned int degree)
  {

    if (degree <= 4) {
      return this->MignottePoly::produce( degree);
    }

    Poly tf("2*x-1");
    
    if (degree == 5) {
      return tf * this->MignottePoly::produce( 3);
    }
    
    int d = static_cast<int>( floor( degree / 2.));
    Poly f = this->MignottePoly::produce( d);

    if ( degree % 2 == 0) {
      return f * f;
    } else {
      return f * f * tf;
    }

  }
};


__END_NAMESPACE_SYNAPS



enum PolyType  { RANDOM = 0, RANDOM2, INTROOTS, RATROOTS, 
		 MIGNOTTE, MIGNOTTE2, MIGNOTTE3 };




int main( int argc, char** argv)
{
  
  int iters = 100;
  int  whatType = RANDOM;
  std::string path("generic/");
  int degree = 3;
  int bits = 10;
  int seed = 101;

  if (argc == 2) {
    if (std::string(argv[1]) == "-help") {
      std::cout << "Program usage: program_name option value \n"
		<< "where options value pair is: \n"
		<< "    -type [0 1 2 3 4] \n"
		<< "    -path the_path \n"
		<< "    -bits nb_of_bits\n"
		<< "    -degree the_degree\n"
		<< "    -iters iterations \n"
		<< "    -seed  seed" << std::endl;;
      return 0;
    }
  }
  
  
  
  for (int i = 1; i < argc; ++i) {
    if (std::string(argv[i]) == "-type") {
      ++i;
      int tmp = atoi( argv[i]);
      if (tmp <= 6) whatType = tmp;
      else whatType = RANDOM;
      continue;
    }
    
    if (std::string(argv[i]) == "-path") {
      ++i;
      path = argv[i];
      continue;
    }
    
    if (std::string(argv[i]) == "-bits") {
      ++i;
      bits = atoi( argv[i]);
      continue;
    }

    if (std::string(argv[i]) == "-degree") {
      ++i;
      degree = atoi( argv[i]);
      continue;
    }
    
    if (std::string(argv[i]) == "-iters") {
      ++i;
      iters = atoi( argv[i]);
      continue;
    }

    
    if (std::string(argv[i]) == "-seed") {
      ++i;
      seed = atoi( argv[i]);
      continue;
    }

    std::cout << "Wrong parameter (" << argv[i] << ") \n"
	      << "type: programe_name -help" << std::endl; 
    return -1;
  }

  
  SYNAPS::RandomPoly* RND;
  switch (whatType) 
  {
      case RANDOM:
	RND = new SYNAPS::RandomPoly(bits, seed);
	break;
      case RANDOM2:
	RND = new SYNAPS::MultipleRandomPoly(bits, seed);
	break;
      case INTROOTS:
	RND = new SYNAPS::RandomIntRootsPoly(bits, seed);
	break;
      case RATROOTS:
	RND = new SYNAPS::RandomRatRootsPoly(bits, seed);
	break;
      case MIGNOTTE:
	RND = new SYNAPS::MignottePoly(bits, seed);
	break;
      case MIGNOTTE2:
	RND = new SYNAPS::MignottePoly_2(bits, seed);
	break;
      case MIGNOTTE3:
	RND = new SYNAPS::MultipleMignottePoly(bits, seed);
	break;
	
      default:
	RND = new SYNAPS::RandomPoly(bits, seed);
	break;
  }


  //  typedef SYNAPS::Sturm<ZZ> SOLVER;  
  
  std::string fname = path + "/d";
  if (degree < 10) fname += "0";
  fname += to_string( degree);
  fname += "_n" + to_string( bits) + ".dat";
      
  std::ofstream outfile(fname.c_str(), std::ios::out);

  outfile << "<data> \n<var> x </var> \n";

  outfile << "<code> ";
  outfile << RND->code();
  outfile << " </code>" << std::endl; 

  outfile << "<description> ";
  outfile << RND->description();
  outfile << "</description> " << std::endl;

  outfile << "<degree> " << degree << " </degree> " << std::endl;
  outfile << "<nbbits> " << bits << "</nbbits>" << std::endl; 
      
  for (int i = 0; i < iters; ++i) {
    Poly f = RND->produce( degree);
    
    SYNAPS::xml::print( outfile,  f);
    outfile << std::endl; 
    //    outfile << "<nb_sols> " << Count_real_roots(f, SOLVER())<< " </nb_sols>" << std::endl;
  }
      
  outfile << "</data>" << std::endl; 
  outfile.close();

  return 0;
}

