/*********************************************************************
*   This file is part of the source code of the SYNAPS kernel.
*   Author(s): B. Mourrain, GALAAD, INRIA
**********************************************************************/
#ifndef synaps_upol_bezier_BEZIER_h_
#define synaps_upol_bezier_BEZIER_h_


#include <algorithm>
#include <vector>
#include <utility>
#include <synaps/init.h>
#include <synaps/linalg/rep1d.h>
#include <synaps/arithm/binomial.h>
#include <synaps/arithm/sign.h>
#include <synaps/upol/UPOLDAR.m>
/********************************************************************/
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
struct  TheRow {};
struct  TheCol {};
//--------------------------------------------------------------------
/// Module for univariate Bernstein representation.
/**
   It contains generic functions for univariate polynomials
   represented in the Bernstein basis.
**/
namespace BEZIER
{

  //--------------------------------------------------------------------
  /// Print the vector in terms of the Bernstein polynomials @B[d,i]@.
  template <class OS, class R>
  inline OS & print(OS & os, const R & v, char* x="x")
  {
    typedef typename R::size_type size_type;
    typename R::const_iterator it=v.begin();
    if ( v.size()==0) 
      os << '(' << 0 << ')';
    else 
    {
      bool plus=false;
      size_type i=0;
      int d=v.size()-1;
      for(; it != v.end(); ++it, ++i)
	if(*it!=0) 
	  { 
	    if(plus) os<<'+'; else plus=true;
	    os << "(" << *it << ")*B["<<i<<","<<d<<"]("<<x<<")";
	  }
    }
    return os;
  }
  //--------------------------------------------------------------------
  template <class R, class C>
  inline R& add_cst(R & r, const C & c)
  {
    for(unsigned i=0;i<r.size();i++) r[i]+=c; return r;
  }
  //--------------------------------------------------------------------
  template <class R>
  inline R& diff(R & r, const R & p)
  {
    int d=p.size()-1;
    for(int i=0;i<p.size()-1;i++) 
      {
	r[i]=p[i+1]-p[i];
	r[i]*=d;
      }
  }
  //--------------------------------------------------------------------
  /** Apply the de Casteljau evaluation algorithm in place, on @A@
      at ratio @x@.
  **/
  template<class A,class X>
  void subdivide_size( A& right, unsigned size, const X& x )
  {
    unsigned k,i;
    for ( k = 1; k < size; k++ )
      for ( i = 0; i < size-k; i++ )
	  right[i] =(1-x)*right[i]+x*right[i+1];
  };

  //--------------------------------------------------------------------
  /** Apply the de Casteljau evaluation algorithm on @B@ at value
      $0.5$. The result is put in @A@.
  **/
  template<class A,class B>
  void subdivide_size( A& left, B& right, unsigned size )
  {
    unsigned k,i;
    for ( k = 1; k < size; k++ )
      {
	left[k-1] = right[0];
	for ( i = 0; i < size-k; i++ )
	  {
	    right[i] +=right[i+1];
	    right[i] /=2;
	  };
      };
    left[size-1] = right[0];
  };


  /** Apply de Casteljau method and clear denominators to work with integers.
   **/
  template<class C> inline
  void subdivide_integer( C* r, C* l, unsigned sz )
    {
      //      VECTOR::print(std::cout,r,sz)<<std::endl;
      C * er, * p, * el;
      el = l+ (sz-2); 
      C * bl; bl=l;
      er = r+ (sz-1);
      for ( er = r + (sz-1); er != r; er--, l++ )
	{
	  *l = *r;
	  for ( p = r; p != er; p ++ )
	    *p = (*p+*(p+1));
	};
      *l = *r;
      C lb=1;
      //      VECTOR::print(std::cout,bl,sz)<<std::endl;
      //      VECTOR::print(std::cout,r,sz)<<std::endl;
      er=r+sz;
      for(p=r+1;p != er; el--,p++)
	{
	  lb  *=2;
	  *el *=lb;
	  *p  *=lb;
	}
    };

  //--------------------------------------------------------------------
  /** Apply the de Casteljau evaluation algorithm on @B@ at value
      @x@. The result is put in @A@.
  **/
  template<class A,class B,class X>
  void subdivide_size( A& left, B& right, unsigned size, const X& x )
  {
    unsigned k,i;
    for ( k = 1; k < size; k++ )
      {
	left[k-1] = right[0];
	for ( i = 0; i < size-k; i++ )
	  right[i] =(1-x)*right[i]+x*right[i+1];
      };
    left[size-1] = right[0];
  };

  //--------------------------------------------------------------------
  /** conversion from bernstein basis to the scaled bernstein basis */
  template<typename C, typename T >
  void scale(T& bzc, size_t sz, int st = 1 )
  {
    int p = 0, d= sz-1;
    for ( size_t i = 0; i < sz; i++, p += st ) bzc[p] *= binomial<C>(d,i);
  };
  //--------------------------------------------------------------------
  /** conversion from scaled bernstein basis to bernstein basis    */
  template<typename C, typename T >
  void uscale(T& bzc, size_t sz, int st = 1 )
  {
    int p = 0, d= sz-1;
    for ( int i = 0; i < sz; i++, p+= st ) bzc[p] /= binomial<C>(d,i);
  };

  //--------------------------------------------------------------------
  template <class R>
  R& elevate_deg(R & a, unsigned d=1)
  {
    if(d !=0)
      {
	R b; 
	scale<typename R::value_type>(a,a.size());
	b.resize(d+1);
	for(unsigned i=0; i<=d; i++) b[i]=binomial<typename R::value_type>(d,i);
	UPOLDAR::mul(a,b);
	uscale<typename R::value_type>(a,a.size());
      }
  }
  //--------------------------------------------------------------------
  template <class R>
  R& elevate_deg(R & r, const R & p, unsigned d=1)
  {
    if(d !=0)
      {
	R a(p),b; 
	scale<typename R::value_type>(a,a.size());
	b.resize(d+1);
	for(unsigned i=0; i<=d; i++) b[i]=binomial<typename R::value_type>(d,i);
	r.resize(a.size()+b.size()-1);
	UPOLDAR::mul(r,a,b);
	uscale<typename R::value_type>(r,r.size());
      }
    else 
      r=p;
  }
  //--------------------------------------------------------------------
  template <class R>
  R& add(R & r, const R & a, const R & b)
  {
    unsigned sa=a.size(), sb=b.size(), sz = std::max(sa,sb);
    if(sa !=sz)
      {
	R ta(a);
	elevate_deg(ta,sz-a.size());
	r.resize(sz);
	VECTOR::add(r,ta,b);
      }
    else if(sb !=sz)
      {
	R tb(b);
	elevate_deg(tb,sz-b.size());
	r.resize(sz);
	VECTOR::add(r,a,tb);
      }
    else
      {
	r.resize(sz);
	VECTOR::add(r,a,b);
      }
    return r;
  }

  //--------------------------------------------------------------------
  template <class R>
  R& add(R & a, const R & b)
  {
    unsigned sa=a.size(), sb=b.size(), sz = std::max(sa,sb);
    if(sa !=sz)
      {
	elevate_deg(a,sz-a.size());
	VECTOR::add(a,b);
      }
    else if(sb !=sz)
      {
	R tb(b);
	elevate_deg(tb,sz-b.size());
	VECTOR::add(a,tb);
      }
    else
      {
	VECTOR::add(a,b);
      }
    return a;
  }

  //--------------------------------------------------------------------
  template <class R>
  R& sub(R & r, const R & a, const R & b)
  {
    unsigned sa=a.size(), sb=b.size(), sz = std::max(sa,sb);
    if(sa !=sz)
      {
	R ta(a);
	elevate_deg(ta,sz-a.size());
	r.resize(sz);
	VECTOR::sub(r,ta,b);
      }
    else if(sb !=sz)
      {
	R tb(b);
	elevate_deg(tb,sz-b.size());
	r.resize(sz);
	VECTOR::sub(r,a,tb);
      }
    else
      {
	r.resize(sz);
	VECTOR::sub(r,a,b);
      }
    return r;
  }

  //--------------------------------------------------------------------
  template <class R>
  R& sub(R & a, const R & b)
  {
    unsigned sa=a.size(), sb=b.size(), sz = std::max(sa,sb);
    if(sa !=sz)
      {
	elevate_deg(a,sz-a.size());
	VECTOR::sub(a,b);
      }
    else if(sb !=sz)
      {
	R tb(b);
	elevate_deg(tb,sz-b.size());
	VECTOR::sub(a,tb);
      }
    else
      {
	VECTOR::sub(a,b);
      }
    return a;
  }

  //--------------------------------------------------------------------
  template <class R>
  R& mul(R& r, const R & a, const R&  b)
  {
    R ta(a), tb(b);
    scale<typename R::value_type>(ta,ta.size());
    scale<typename R::value_type>(tb,tb.size());
    UPOLDAR::mul(r,ta,tb);
    uscale<typename R::value_type>(r,r.size());
    return r;
  }

  //--------------------------------------------------------------------
  // calcul de toutes les elements de la base de bernstein
  // en t. (pour remplir les lignes des matrices d'interpolation par exemple)
  template<class B,class X>
  void BernsteinBasis_loop( B& bcoffs, const X& t, unsigned degree )
  {
    bcoffs[0] = 1;
    unsigned i,k;
    X s,tmp;
    for ( i = 1; i <= degree; i++ )
      {
	s = 0;
	for ( k = 0; k < i; k++ )
	  {
	    tmp = bcoffs[k];
	    bcoffs[k] = s+(1-t)*tmp;
	    s = t*tmp;
	  };
	bcoffs[i] = s;
      };
  };

  //--------------------------------------------------------------------
  /** Apply the de Casteljau subdivision algorithm to @t1@, at the middle of
      the intervall. At the end of the procedure, @t0@ represents the control
      polygon on the first half of the interval and @t1@ corresponds to the
      second half.
  **/
  template<class T>
  inline void subdivide (T& t0,T& t1)
  {
    /*
    int d=(t1.size())-1;
    for (int j=d;j>=0;j--)
      {
	t0[d-j]= t1[0];
	for(int i=0;i<j;i++) t1[i]=(t1[i]+t1[i+1])/2;
      }
    */
    subdivide_size( t0, t1, t1.size() );
  }

  //--------------------------------------------------------------------
  /** Apply the de Casteljau subdivision algorithm, at the point of
      barycentric coordinates $(1-x,x)$.  intervall. At the end of the
      procedure, @t0@ represents the control polygon on the first half
      of the interval and @t1@ corresponds to the second half.
  **/
  template<class T, class C>
  inline void subdivide (T& t0,T& t1, const C& x)
  {
	/*
	   int d=(t1.size())-1;
	   for (int j=d;j>=0;j--)
	   {
	   t0[d-j]= t1[0];
	   for(int i=0;i<j;i++) t1[i]=(1-x)*t1[i]+x*t1[i+1];
	   }
	   */
	subdivide_size(t0,t1,t1.size(),x);
  }
  //--------------------------------------------------------------------
  /** Apply the de Casteljau subdivision algorithm, at the point of
      barycentric coordinates $(1-x,x)$. At the end of the procedure,
      @t1@ corresponds to the representation on $[x,1]$.
  **/
  template<class C, class T>
  inline void subdivide (T& t1, const C& x)
  {
    /*
    int d=(t1.size())-1;
    for (int j=d;j>=0;j--)
      {
	for(int i=0;i<j;i++) t1[i]=(1-x)*t1[i]+x*t1[i+1];
      }
    */
    subdivide_size(t1,t1.size(),x);
  }


  //--------------------------------------------------------------------
  template<class M>
  void subdivide (M& m0,M& m1, const TheRow & x)
  {
    int d=(m1.nbcol())-1;
    for(unsigned k=0;k<m1.nbrow();++k)
      for(int j=d;j>=0;j--)
	{
	  m0(k,d-j)= m1(k,0);
	  for(int i=0;i<j;i++) m1(k,i)=(m1(k,i)+m1(k,i+1))/2;
	}
  }

  //--------------------------------------------------------------------
  template<class M>
  void subdivide (M& m0,M& m1, const TheCol & x)
  {
    int d=(m1.nbrow())-1;
    for(unsigned k=0;k<m1.nbcol();++k)
      for(int j=d;j>=0;j--)
	{
	  m0(d-j,k)= m1(0,k);
	  for(int i=0;i<j;i++) m1(i,k)=(m1(i,k)+m1(i+1,k))/2;
	}
  }
  //--------------------------------------------------------------------
  /** Convert in place the representation in the Bezier basis on $[0,1]$ in
      the monomial basis **/
  template<class C,class T>
  void  bezier2monomials(T& p, int size )
  {
    int i,k;
    for ( i = 1; i < size; i++ )
      for ( k = size-1; k >= i; k-- )
	p[k] -= p[k-1];
    //    cout <<"Pol "<<p<<endl;
    for ( k = 0; k < size; k++ )
      p[k] *= binomial<C>(size-1,k);
  }
  
  //--------------------------------------------------------------------
  /** Convert the representation in the monomial basis to the representation
      in the Bezier basis of $[a,b]$. The matrix of the change of bases is
      $(c_{i,j})_{0\le i,j \le d}$ with
      $c_{i,j}= {{i}\choose{j}}/{{d}\choose{j}}$ if $j\le i$ and $0$ otherwise. **/
  template<class T,class P, class C>
  void  monomials2bezier(T& bz, const P& p, unsigned size,
			 const C& a, const C& b)
  {
    T pps(p.size()); 
    VECTOR::assign(pps,p);
    if(a !=0) 
      UPOLDAR::shift(pps,pps,a);
    UPOLDAR::scale(pps, pps, C(b-a));

    VECTOR::init(bz,pps[0]);
    C dn=1, nm=1;

    for(unsigned j=1; j<size; j++)
      {
	dn *= (size-j); dn/=j;
	nm=1;
	bz[j] += pps[j]/dn;
	for(unsigned i=j+1; i<size; i++)
	  {
	    nm *= i; nm/= (i-j);
	    bz[i] +=nm/dn*pps[j];
	  }
      }
  }

  /** Transform  the Bernstein representation of @p@ on $[a,b]$ into
      its representation on $[c,d]$ and store it in @bz@.
   **/
  template<class T, class C> 
  void  bezier2bezier(T& bz, const T& p,
		      const C& a, const C& b,
		      const C& c, const C& d)
  {
    C t=(c-a)/(b-a);
    bz=p;
    T tmp=p;
    subdivide(bz,tmp,t);
    t=(d-c)/(b-c);
    subdivide(bz,tmp,t);
  }
  //--------------------------------------------------------------------
  template<class C,class T>
  C shrinkinf(const T & p, const C&  a, const C&  b)
  {
    C u0=0, u1=1;
    int s0=sign(p[0]), s1=sign(p[p.degree()]);
    if(b==1)
      while(u1 ==1)
      {
	int sm= sign(eval(p,C((u0+u1)/C(2))));
	if(sm != s0)
	  { s1=sm; u1=(u0+u1)/C(2); }
	else if(sm != s1)
	  { s0=sm; u0=(u0+u1)/C(2); }
	else if(sm!=0)
	  {
	    std::cerr <<"Error in sign evaluation"<< std::endl;
	    return 100000;//	  exit(1);
	  }
      }
    return C(a+ u1*(b-a));
  }
  //----------------------------------------------------------------------
  /**
     Compute of the value of
     $\sum p_{i}*{d\choose i}\, x^i\, (1-x)^{d-i}, i=0,..,d.$ at @x@.
  **/
  template<class Y,class R, class C>
  Y eval(const R & p, const C & x)
  {
    linalg::rep1d<Y> v(p.size());
    VECTOR::assign(v,p);
    Y xx, val = Y(1), b=Y(1);
    int d=p.degree();

    xx = val - (Y)x;

    for(int i=d;i>=0;i--)
      { v[i] = p[i]*b; b*=Y(i);b/=Y(d-i+1); v[i] *= val; val *= xx; }

    val = v[d];

    for(int i=d;i>0;i--) { val *= x; val += v[i-1]; }

    return val;
  }

} //CLOSE BEZIER

__END_NAMESPACE_SYNAPS
/*********************************************************************/
#endif // SYNAPS_UPOL_BEZIER_H

