#ifndef SYNAPS_ARITHM_ALGEBRAIC_ALGEBRAIC_CONFIG_H
#define SYNAPS_ARITHM_ALGEBRAIC_ALGEBRAIC_CONFIG_H
#include <iostream>
#include <string>
#include <vector>
#include <map>

#include <synaps/init.h>
#include <synaps/base/comparison.h>
#include <synaps/base/Debug_util.h>
#include <synaps/usolve/NumberTraits.h>


#define __ALGEBRAIC_BEGIN_NAMESPACE   __BEGIN_NAMESPACE_SYNAPS \
                                      namespace ALGEBRAIC {

#define __ALGEBRAIC_END_NAMESPACE     __END_NAMESPACE_SYNAPS \
									  }

__ALGEBRAIC_BEGIN_NAMESPACE
//====================================================================
/*! \file Algebraic_config.H

	\brief Various definitions for the ALGEBRAIC_NAMESPACE.


	We define the ALGEBRAIC namespace in this file. 
	Also, there several inline functions that we need for the computations.
*/

using ::SYNAPS::sign;
using ::SYNAPS::to_double;

enum Root_mult
{
	// For the Root_of_1
	ROOT_ONE,
	// For the Root_of _2
	ROOT_DOUBLE,
	ROOT_ONE_ONE,
	ROOT_COMPLEX_COMPLEX,
	// For the Root_of_3
	ROOT_ONE_ONE_ONE,
	ROOT_ONE_COMPLEX_COMPLEX,
	ROOT_TRIPLE,
	ROOT_ONE_DOUBLE,
	ROOT_DOUBLE_ONE,
	// For the Root_of_4
	ROOT_ONE_ONE_ONE_ONE,
	ROOT_DOUBLE_ONE_ONE,
	ROOT_ONE_DOUBLE_ONE,
	ROOT_ONE_ONE_DOUBLE,
	ROOT_DOUBLE_DOUBLE,
	ROOT_DOUBLE_COMPLEX_COMPLEX,
	ROOT_ONE_ONE_COMPLEX_COMPLEX,
	ROOT_TRIPLE_ONE,
	ROOT_ONE_TRIPLE,
	ROOT_QUADRUPLE,
	// All the root complex
	ROOT_ALL_COMPLEX,

	// Other
	ROOT_UNKNOWN
};

const int POS_MIDDLE     = 0;
const int POS_FIRST      = 1;
const int POS_LAST       = 2;
const int POS_FIRST_LAST = 3;
const int POS_UNKNOWN    = 4;



// Print for debug reasons
template < typename STR > inline
void print(const STR& name)
{
#ifdef DO_DEBUG
	std::cout << name << std::endl;
#endif
	return;
}

// Print for debug reasons
template < typename STR,
           typename RT >   inline
void print(const STR& name, const RT& value)
{
#ifdef DO_DEBUG
	std::cout << "  " << name << " = " << value << std::endl;
#endif
	return;
}

inline std::string my_pretty_function (const std::string& str_i)
{
	int     int1 = (str_i.substr (0, str_i.find ("::"))).find_last_of (" ") + 1;
	std::string  str1 = str_i.substr (int1);
	std::string str_j = str_i.substr (int1, str1.find ("["));

	unsigned int loc;
	std::string str_t("SYNAPS::ALGEBRAIC::");
	for(;;) {
		loc = str_j.find(str_t, 0);
		if (loc == std::string::npos) break;
		str_j.replace(loc, str_t.length(), str_t.length(), 0);
	}

	return str_j;
	//	int     int1 = (str_i.substr (0, str_i.find ("::"))).find_last_of (" ") + 1;
	//	std::string  str1 = str_i.substr (int1);
	//	return str_i.substr (int1, str1.find ("("));
}
#define __MY_PRETTY_FUNCTION__ my_pretty_function (__PRETTY_FUNCTION__)


#ifndef ERR_LOG_LEVEL_1
	#if defined(NDEBUG) || !defined(DO_DEBUG)
		#define ERR_LOG_LEVEL_1
	#else
		#define ERR_LOG_LEVEL_1 std::cerr << "ALGEBRAIC::" << __MY_PRETTY_FUNCTION__ << std::endl
	#endif
#endif


#ifndef SYNAPS_ASSERT
	#define SYNAPS_ASSERT(EX) \
     (EX ? cout << "Ok" << endl : cout <<  #EX << "HELLAS" << endl)
#endif


inline int opposite(const int& a)
{ return -a; }



template< class RT >
inline
RT SQR(const RT& A)
{ return A * A; }

template< class RT >
inline
RT MUL(const RT& A, const RT& B)
{ return A * B; }



template< class RT >
inline
RT ADD(const RT& A, const RT& B)
{ return A + B; }

template< class RT >
inline
RT SUBS(const RT& A, const RT& B)
{ return A - B; }



template< class RT >
inline
RT calc_brackets(const RT& A1, const RT& B2, const RT& A2, const RT& B1)
{ return A1 * B2 - A2 * B1; }


template< class RT >
inline
RT calc_D2(const RT& A, const RT& B, const RT& C)
{ return B * B - A * C; }


template< class RT >
inline
RT calc_P(const RT& C3, const RT& W, const RT& C2, const RT& D2)
{ return -C3 * W + RT(3) * C2 * D2; }



template< class RT >
inline
RT calc_F3_BA(const RT& A, const RT& B, const RT& W, const RT& D2x2)
{ return A * W - B * D2x2; }


template< class RT >
inline
RT calc_F4_BA(const RT& a, const RT& A, const RT& D2)
{ return SQR(a) * A + RT(-3) * SQR(D2); }


template< class RT >
inline
RT calc_F4_WD(const RT a, const RT& B, const RT& D2, const RT& D3, const RT& W1)
{
	RT D2_3 = SQR(D2);
	D2_3 = D2_3 * D2;
	RT tmp = SQR(W1) - RT(4) * D2 * D3;
	return RT(-16) * B * D2_3 + a * SQR(tmp);
}





///   inline functions for poly of degree 1
template< class RT >
inline
RT calc_one(
	    const RT& A, const RT& B,
		const RT& P
		)
{
	return  A * P + B;
}

template< class RT >
inline
RT calc_one(
	    const RT& A, const RT& B,
		const RT& P, const RT& Q
		)
{
	return  A * P + B * Q;
}



/// Inline functions for poly of degree 2
template< class RT >
inline
RT calc_quadratic(
	const RT& A, const RT& B, const RT& C,
	const RT& P
	)
{
	return  (A * P + B) * P + C;
}


template< class RT >
inline
RT calc_quadratic(
	const RT& A, const RT& B, const RT& C,
	const RT& P, const RT& Q
	)
{
	RT Q_2 = SQR(Q);
	return  (A * P + B * Q) * P + C * Q_2;
}


template< class RT >
inline
RT calc_quadratic(
	const RT& A, const RT& B, const RT& C,
	const RT& P, const RT& Q, const RT& Q_2
	)
{
	return  (A * P + B * Q) * P + C * Q_2;
}





/// inline functions for poly of degree 3


template< class RT >
inline
RT calc_cubic(
	const RT& A, const RT& B, const RT& C,   const RT& D,
	const RT& P
	)
{
	return  ( ( A * P + B ) * P + C ) * P + D;
}



template< class RT >
inline
RT calc_cubic(
	const RT& A, const RT& B, const RT& C,   const RT& D,
	const RT& P, const RT& Q)
{
	RT Q_2 = SQR(Q);
	RT Q_3 = Q * Q_2;
	return ( ( A * P + B * Q ) * P + C * Q_2 ) * P + D * Q_3;
}


template< class RT >
inline
RT calc_cubic(
	const RT& A, const RT& B, const RT& C,   const RT& D,
	const RT& P, const RT& Q, const RT& Q_2, const RT& Q_3)
{
	return ((A*P + B*Q)*P + C*Q_2)*P + D*Q_3;
}




/// Inline functions for poly of degree 4


template< class RT >
inline
RT calc_quartic(
		const RT& A, const RT& B, const RT& C, const RT& D, const RT& E,
		const RT& P)
{
	return  ( ( ( A * P + B ) * P + C ) * P + D ) * P + E;
}


template< class RT >
inline
RT calc_quartic(
		const RT& A, const RT& B, const RT& C, const RT& D, const RT& E,
		const RT& P, const RT& Q)
{
	RT Q_2 = SQR(Q);
	RT Q_3 = Q * Q_2;
	RT Q_4 = Q_2 * Q_2;
	return  (((A*P + B*Q)*P + C*Q_2)*P + D*Q_3)*P + E*Q_4;
}



template< class RT >
inline
RT calc_quartic(
		const RT& A, const RT& B, const RT& C, const RT& D, const RT& E,
		const RT& P, const RT& Q, const RT& Q_2, const RT& Q_3, const RT& Q_4)
{
	return  ( ( ( A * P + B * Q ) * P + C * Q_2 ) * P + D * Q_3 ) * P + E * Q_4;
}




// the coefficients are considered normalized!
template< class RT >
inline
RT calc_quadratic_normal(
	const RT& A, const RT& B, const RT& C,
	const RT& P, const RT& Q, const RT& Q_2
	)
{
	return  (A * P - RT(2) * B * Q) * P + C * Q_2;
}



// the coefficients are considered normalized!
template< class RT >
inline
RT calc_cubic_normal(
	const RT& A, const RT& B, const RT& C,   const RT& D,
	const RT& P, const RT& Q, const RT& Q_2, const RT& Q_3)
{
	return  ( ( A * P - RT(3) * B * Q ) * P + RT(3) * C * Q_2 ) * P - D * Q_3;
}


// the coefficients are considered normalized!
template< class RT >
inline
RT calc_cubic_normal(
	const RT& A, const RT& B, const RT& C,   const RT& D,
	const RT& P, const RT& Q)
{
	RT Q_2 = SQR(Q);
	RT Q_3 = MUL(Q, Q_2);
	return  ( ( A * P - RT(3) * B * Q ) * P + RT(3) * C * Q_2 ) * P - D * Q_3;
}


// the coefficients are considered normalized!
template< class RT >
inline
RT calc_quartic_normal(
		const RT& A, const RT& B, const RT& C, const RT& D, const RT& E,
		const RT& P, const RT& Q, const RT& Q_2, const RT& Q_3, const RT& Q_4)
{
	return  ( ( ( A * P - RT(4) * B * Q ) * P + RT(6) * C * Q_2 ) * P - RT(4) * D * Q_3 ) * P + E * Q_4;
}





template< class RT >
inline
RT calc_A(const RT& W3_, const RT& D3_)
{ return W3_ + RT(3) * D3_; }


template< class RT >
inline
RT calc_B(const RT& d, const RT& W1_, const RT& e, const RT& D2_, const RT& c, const RT D3_)
{ return -d * W1_ - e * D2_ - c * D3_; }


// Calc the discriminant of the quartic
template< class RT >
inline
RT calc_D4(const RT& A, const RT& B)
		// ATTENTION
		// ERROR in Yang's paper
{ return SQR(A) * A - RT(27) * SQR(B); }


template< class RT >
inline
RT calc_T1(const RT& W1_, const RT& D2_, const RT& D3_, const RT& W3_)
{ return RT(-3) * SQR(W1_) + RT(9) * D2_ * D3_ - W3_ * D2_; }


template< class RT >
inline
RT calc_T2(const RT& A, const RT& W1_, const RT& b, const RT& B)
{ return RT(3) * A * W1_ - RT(9) * b * B; }

template< class RT >
inline
RT calc_R(const RT& a, const RT& W1_, const RT& b, const RT& D2_x2)
{ return a * W1_ + b * D2_x2; }


template< class RT >
inline
RT calc_triple_p(const RT& a, const RT& W1_, const RT& b, const RT& D2_)
{ return RT(3) * a * W1_ + RT(8) * b * D2_; }


template< class RT >
inline
RT calc_triple_q(const RT& a, const RT& D2_)
{ return RT(2) * a * D2_; }





// REMOVE them from here!!!!!

template< class RT >
inline
RT calc_isol_1(const RT& D2_, const RT& W1_, const RT& W3_)
{
//	RT radical = RT(9)*SQR(W1_) + RT(12)*D2_*W3_;
	return RT(-3) * W1_ - sqrt( RT(  RT(9)*SQR(W1_) + RT(12)*D2_*W3_ ) );//;radical );
}

template< class RT >
inline
RT calc_isol_2(const RT& D2_, const RT& W1_, const RT& W3_)
{
//	RT radical = RT(9)*SQR(W1_) + RT(12)*D2_*W3_;
	return RT(-3) * W1_ + sqrt( RT( RT(9)*SQR(W1_) + RT(12)*D2_*W3_ ) );//radical );
}

template< class RT >
inline
RT calc_isol_denom(const RT& D2_)
{
	return RT(6)*D2_;
}


__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_ALGEBRAIC_CONFIG_H

