/*! \file Comparison.H

\brief Definitions needed for the comparison of algebraic numbers.

We define equality and inequality operators for the algebraic numbers.

*/

#ifndef SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_H
#define SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_H


#include <map>

#include "Algebraic_config.h"

#include "comparison/auto/Auto_Results.h"
#include "comparison/Comparison_1x.h"
#include "comparison/Comparison_22.h"
#include "comparison/Comparison_23.h"
#include "comparison/Comparison_24.h"
#include "comparison/Comparison_33.h"
#include "comparison/Comparison_34.h"
#include "comparison/Comparison_44.h"
#include "comparison/Sturm_comparison.h"


__ALGEBRAIC_BEGIN_NAMESPACE

// Forward declaration
template < typename RT, typename POLY > struct root_of;
//-------------------------------------------------------------------------------

template < typename RT,
	   typename P >  inline
int
compare(const root_of<RT, P>& a, const RT& x)
{
  return compare(a, to_FT(x));
}
//-------------------------------------------------------------------------------

template < typename RT,
	   typename P >  inline
int
compare(const RT& x, const root_of<RT, P>& a)
{
  return opposite(compare(a, x));
}

//-------------------------------------------------------------------------------

template < typename RT,
	   typename P >  inline
int
compare(const root_of<RT, P>& a, const typename NumberTraits<RT>::FT& x)
{
  if (a.degree() == 1) {
    if (a.interval() > x) return LARGER;
    if (a.interval() < x) return SMALLER;
    return EQUAL;
  }

  // a.degree() > 1
  if (a.interval() >= x) return LARGER;
  if (a.interval() <= x) return SMALLER;

  int sgn_on_x = sign( UPOLDAR::eval_horner(a.poly(), x) );
  if (sgn_on_x == 0) return EQUAL;

  if (sgn_on_x == a.sign_low()) return LARGER;
  return SMALLER;
}
//-------------------------------------------------------------------------------

template < typename RT,
	   typename P >  inline
int
compare(const typename NumberTraits<RT>::FT& x, const root_of<RT, P>& a)
{
  return opposite(compare(a, x));
}
//-------------------------------------------------------------------------------

template < typename RT,
	   typename P >
int
compare(const root_of<RT, P>& a, const root_of<RT, P>& b)
{
  typedef root_of<RT, P>    RO_t;

  typedef int (* CompFuncPtr)(const RO_t&, const RO_t&);

  static const CompFuncPtr compareTable[5][5] = {
    { compare_11<RO_t>,
      compare_1x<RO_t>,
      compare_1x<RO_t>,
      compare_1x<RO_t>,
      compare_1x<RO_t> },

    { compare_x1<RO_t>,
      compare_22<RO_t>,
      compare_23<RO_t>,
      compare_24<RO_t>,
      sturmcompare<RO_t> },

    { compare_x1<RO_t>,
      compare_32<RO_t>,
      compare_33<RO_t>,
      compare_34<RO_t>,
      sturmcompare<RO_t> },

    { compare_x1<RO_t>,
      compare_42<RO_t>,
      compare_43<RO_t>,
      compare_44<RO_t>,
      sturmcompare<RO_t> },

    { sturmcompare<RO_t>,
      sturmcompare<RO_t>,
      sturmcompare<RO_t>,
      sturmcompare<RO_t>,
      sturmcompare<RO_t> }

  };

  a.refine_interval();
  b.refine_interval();
  
  int i = (a.degree() <= 4 ? a.degree() - 1 : 4);
  int j = (b.degree() <= 4 ? b.degree() - 1 : 4);

  return compareTable[i][j](a, b);
}


__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_H


