// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)


#ifndef SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_22_HH
#define SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_22_HH

#include <synaps/usolve/algebraic/Algebraic_config.h>


__ALGEBRAIC_BEGIN_NAMESPACE


// TODO: TODO:
//    Put them in internal namespace !!!

// Forward declaration
template < typename RT, typename P > struct root_of;
//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
compare_22_11(const RO_t& a, const RO_t& b)
{
	ERR_LOG_LEVEL_1;

	typedef typename RO_t::NTR      NTR;
	typedef typename NTR::RT        RT;
	typedef typename NTR::FT        FT;

	RT a1 = RT( 2) * a[2];
	RT b1 = RT(-1) * a[1];
	RT c1 = RT( 2) * a[0];
	RT a2 = RT( 2) * b[2];
	RT b2 = RT(-1) * b[1];
	RT c2 = RT( 2) * b[0];

	RT a1xb2 = (a1 * b2);
	RT a2xb1 = (a2 * b1);

	RT J = (a1xb2 - a2xb1);
	RT K = calcK(a1, b1, c1, a2, b2, c2);
	RT D1 = a.discriminant();
	RT D2 = b.discriminant();
	if (J > 0){
		if (K > 0) return SMALLER;
		RT D = calcD(a1,  D1, a2, D2);
		if (D > 0) return SMALLER;
		RT Jp = calcJp(b1, c1, b2, c2);
		if (Jp < 0) return LARGER;
		RT P4 = calcP4(D1, D2, K);
		return sign(P4);
	}
	else{ // J<0
		if (K > 0) return LARGER;
		RT D = calcD(a1, D1, a2, D2);
		if (D < 0) return LARGER;
		RT Jp = calcJp(b1, c1, b2, c2);
		if (Jp > 0) return SMALLER;
		RT P4 = calcP4(D1, D2, K);
		return sign(-P4);
	}

}
//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
compare_22_22(const RO_t& a, const RO_t& b)
{
	RO_t ta(a);
	RO_t tb(b);
	ta.negate();
	tb.negate();
	int result = compare_22_11(ta, tb);

	return opposite(result);
}
//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
compare_22_21(const RO_t& a, const RO_t& b)
{
	typedef typename RO_t::NTR      NTR;
	typedef typename NTR::RT        RT;
	typedef typename NTR::FT        FT;

	RT a1 = RT( 2) * a[2];
	RT b1 = RT(-1) * a[1];
	RT c1 = RT( 2) * a[0];
	RT a2 = RT( 2) * b[2];
	RT b2 = RT(-1) * b[1];
	RT c2 = RT( 2) * b[0];
//	RT a1 = a.N2(); RT b1 = a.N1();  RT c1 = a.N0();
//	RT a2 = b.N2(); RT b2 = b.N1();  RT c2 = b.N0();

	RT a1xb2 = MUL(a1, b2);
	RT a2xb1 = MUL(a2, b1);
	RT J = SUBS(a1xb2, a2xb1);
	if (J <= RT(0)) return LARGER;


	RT K = calc_K(a1, b1, c1, a2, b2, c2);
	if (K < RT(0)) return LARGER;

	RT Jp = calc_brackets(b1, c2, b2, c1);
	if (Jp < RT(0)) return SMALLER;

	RT D1 = a.discriminant();
	RT D2 = b.discriminant();
	RT S3 = -calc_22_S3(K, D1, D2);
	if (S3 > RT(0)) return SMALLER;
	if (S3 < RT(0)) return LARGER;

	return EQUAL;
}
//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
compare_22_12(const RO_t& a, const RO_t& b)
{
	RO_t ta(a);
	RO_t tb(b);
	ta.negate();
	tb.negate();
	int result = compare_22_21(ta, tb);
	return opposite(result);
}
//-------------------------------------------------------------------------------


template < typename RO_t >  inline
int
compare_22(const RO_t& a, const RO_t& b)
{
	assert(a.degree() == 2);
	assert(b.degree() == 2);

	if (a.which() == 0) {
		if (b.which() == 0) return compare_22_11(a, b);
		return compare_22_12(a, b);
	}
	if (b.which() == 0) return compare_22_21(a, b);
	return compare_22_22(a, b);
}


template < typename RT >
inline RT calcK(const RT& A1, const RT& B1, const RT& C1,
                const RT& A2, const RT& B2, const RT& C2)
{ return C1 * A2 + A1 * C2 - RT(2) * B1 * B2; }

template < typename RT >
inline RT calcJp(const RT& B1, const RT& C1, const RT& B2, const RT& C2)
{ return B1 * C2 - C1 * B2; }

template < typename RT >
inline RT calcP4(const RT& I1, const RT& I2, const RT& K)
{ return SQR(K) - RT(4) * I1 * I2;}

template < typename RT >
inline RT calcD(const RT& A1, const RT& I1, const RT& A2, const RT& I2)
{ return I1 * SQR(A2) - I2 * SQR(A1);}

template< typename RT >
inline
RT calc_K(const RT& A1, const RT& B1, const RT& C1,
const RT& A2, const RT& B2, const RT& C2)
{ return A1 * C2 + A2 * C1 - RT(2) * B1 * B2; }

template< typename RT >
inline
RT calc_22_F_BA(const RT& t1, const RT& t2, const RT& B, const RT& A, const RT& C)
{ return ( t1 - RT(2) * t2 ) * B + C * SQR(A); }

template< typename RT >
inline
RT calc_22_S2(const RT& a1, const RT& K, const RT& a2, const RT& D)
{ return a1 * K + RT(2) * a2 * D; }

template< typename RT >
inline
RT calc_22_S3(const RT& K, const RT& D1, const RT& D2)
{ return RT(4) * D1 * D2 - K * K; }




__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_22_HH

