// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)


#ifndef SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_33_HH
#define SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_33_HH

#include <synaps/usolve/algebraic/Algebraic_config.h>
#include <synaps/usolve/algebraic/comparison/Common.h>


__ALGEBRAIC_BEGIN_NAMESPACE

// Forward declaration
template < typename RT, typename P > struct root_of;
//-------------------------------------------------------------------------------


template< class RT >
inline
RT calc_33_F_PQ(const RT& A, const RT& B, const RT& C,   const RT& D,
                const RT& P, const RT& Q, const RT& Q_2, const RT& Q_3)
{ return  ( ( A * P - RT(3) * B * Q ) * P + RT(3) * C * Q_2) * P - D * Q_3; }

template< class RT >
inline
RT calc_33_S2_PQ(const RT& J, const RT& G, const RT& M,
		         const RT& P, const RT& Q, const RT& Q_2)
{ return RT(-3) * ( J * P - G * Q ) * P - M * Q_2; }

template< class RT >
inline
RT calc_33_Z1(const RT& J, const RT& M, const RT& J1, const RT& G)
{ return J * ( M + RT(9) * J1 ) - RT(3) * G * G; }

template< class RT >
inline
RT calc_33_Z2(const RT& G, const RT& M, const RT& J, const RT& M1)
{ return G * M - RT(3) * J * M1; }

template< class RT >
inline
RT calc_33_S3_PQ(const RT& Z1, const RT& Z2, const RT& P, const RT& Q)
{ return Z1 * P + Z2 * Q; }

template< class RT >
inline
RT calc_33_S4(const RT& M, const RT& J, const RT& M1, const RT& M2,
		      const RT& Z1, const RT& Z2)
{ return M * ( SQR(M) - RT(9) * J * M2 )- RT(9) * ( M1 * Z2 + M2 * Z1 ); }


template< class RT >
inline
RT calc_33_deg_S4(const RT& M, const RT& G, const RT& M1, const RT& M2)
{ return M * SQR(M) - RT(9) * G * ( M * M1 - RT(3) * G * M2 ); }




//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
eval_full_Sturm_33_pq(
		const RO_t& a,
		const RO_t& b,
		const typename RO_t::FT& isol_1,
		const typename RO_t::FT& isol_2,
		int& where
		)
{
	typedef typename RO_t::NTR    NTR;
	typedef typename RO_t::RT     RT;
	typedef typename RO_t::FT     FT;

	RT a1 = RT( 3) * a[3];
	RT b1 = RT(-1) * a[2];
	RT c1 = RT( 1) * a[1];
	RT d1 = RT(-3) * a[0];
	RT a2 = RT( 3) * b[3];
	RT b2 = RT(-1) * b[2];
	RT c2 = RT( 1) * b[1];
	RT d2 = RT(-3) * b[0];
//	RT a1 = a.N3(); RT b1 = a.N2();  RT c1 = a.N1();  RT d1 = a.N0();
//	RT a2 = b.N3(); RT b2 = b.N2();  RT c2 = b.N1();  RT d2 = b.N0();

	RT p1 = numerator(isol_1);
	RT q1 = denominator(isol_1);
	RT p2 = numerator(isol_2);
	RT q2 = denominator(isol_2);
	RT q1_2 = SQR(q1);
	RT q2_2 = SQR(q2);

	int result;
	RT J = calc_brackets(a1, b2, a2, b1);
	RT S2_inf = MUL(RT(-3), J);
	if (S2_inf == RT(0)){
		return eval_degJ_Sturm_33_pq(
				a1, b1, c1, d1,
				a2, b2, c2, d2,
				p1, q1,
				p2, q2,
				where);
	}

	RT G = calc_brackets(a1, c2, a2, c1);
	RT M = calc_brackets(a1, d2, a2, d1);
	RT S2_p2q2 = calc_33_S2_PQ(J, G, M, p2, q2, q2_2);
	if (S2_p2q2 > RT(0)) where ^= 16;
	RT S2_p1q1 = calc_33_S2_PQ(J, G, M, p1, q1, q1_2);
	if (S2_p1q1 >= RT(0)) where ^= 32;

	RT J1 = calc_brackets(b1, c2, b2, c1);
	RT Z1 = calc_33_Z1(J, M, J1, G);
	RT S3_inf = Z1;
	// Degeneracy S3_inf = 0
	if (S3_inf == RT(0)){
		RT M1 = calc_brackets(b1, d2, b2, d1);
		RT Z2 = calc_33_Z2(G, M, J, M1);
		S3_inf = Z2;
		if (S3_inf == RT(0))
			return Result_deg_3::on_end(where);
		if (S3_inf > RT(0)) { where ^= 64; where ^= 128;}
			return Result_deg_4::on_end(where);
	}
	RT M1 = calc_brackets(b1, d2, b2, d1);
	RT Z2 = calc_33_Z2(G, M, J, M1);

	RT S3_p2q2 = calc_33_S3_PQ(Z1, Z2, p2, q2);
	if (S3_p2q2 > RT(0)) where ^= 64;
	result = Static_Result::Result_5.on_S3_p2q2(where);
	if (result != UNKNOWN) {
		return result;
	}
	RT S3_p1q1 = calc_33_S3_PQ(Z1, Z2, p1, q1);
	if (S3_p1q1 >= RT(0)) where ^= 128;
	result = Static_Result::Result_5.on_S3_p1q1(where);
	if (result != UNKNOWN) {
		return result;
	}

	RT M2 = calc_brackets(c1, d2, c2, d1);
	RT S4 = calc_33_S4(M, J, M1, M2, Z1, Z2);
	if (S4 ==  RT(0)) return EQUAL;
	if (S4 > RT(0)) where ^= 256;
	result = Static_Result::Result_5.on_S4(where);
	if (result != UNKNOWN) {
		return result;
	}
	return EQUAL;
}
//-------------------------------------------------------------------------------


template < typename RT >  inline
int
eval_degJ_Sturm_33_pq(
		const RT &a1, const RT &b1, const RT &c1, const RT &d1,
		const RT &a2, const RT &b2, const RT &c2, const RT &d2,
		const RT &p1,  const RT &q1,
		const RT &p2,  const RT &q2,
		int& where
		)
{
	int result;
	RT G = calc_brackets(a1, c2, a2, c1);
	RT S2_inf = MUL(RT(3), G);
	if (S2_inf == RT(0)){
		RT M = calc_brackets(a1, d2, a2, d1);
		S2_inf = -M;
		if (S2_inf == RT(0)) return EQUAL;
		if (S2_inf > RT(0)) where ^= 16;
		return Static_Result::Result_3.on_S2(where);
	}
	RT M = calc_brackets(a1, d2, a2, d1);

	RT S2_p2q2 = calc_33_S3_PQ(S2_inf, RT(-M), p2, q2);
	if (S2_p2q2 > RT(0)) where ^= 16;
	RT S2_p1q1 = calc_33_S3_PQ(S2_inf, RT(-M), p1, q1);

	if (S2_p1q1 >= RT(0)) where ^= 32;
	result = Static_Result::Result_4.on_S2_p1q1(where);
	if (result != UNKNOWN) {
		return result;
	}

	RT J = RT(0);
	RT J1 = calc_brackets(b1, c2, b2, c1);
	RT Z1 = calc_33_Z1(J, M, J1, G);
	RT M1 = calc_brackets(b1, d2, b2, d1);
	RT Z2 = calc_33_Z2(G, M, J, M1);
	RT M2 = calc_brackets(c1, d2, c2, d1);
	RT S3 = calc_33_S4(M, J, M1, M2, Z1, Z2);
	if (G >= RT(0)) S3 = -S3;
	if (S3 == RT(0)) return EQUAL;
	if (S3 > RT(0)) where ^= 64;
	return Static_Result::Result_4.on_S3(where);
}
//-------------------------------------------------------------------------------


template < typename RO_t >  inline
int
compare_33(const RO_t& a, const RO_t& b)
{
	typedef typename RO_t::NTR    NTR;
	typedef typename RO_t::RT     RT;
	typedef typename RO_t::FT     FT;
	typedef typename RO_t::FIT    FIT;

	int where = 0;
	FIT I;
	int result = decide_common_interval(a, b, I, where);
	if (result != UNKNOWN)
		return result;

	// cout << "TOTO" << endl;
	// cout << "where = " << where << endl;
	// cout << "common = " << I << endl;
	return eval_full_Sturm_33_pq(a, b, I.lower(), I.upper(), where) * b.sign_up();
}
//-------------------------------------------------------------------------------


__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_33_HH

