/********************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): B. Mourrain, GALAAD, INRIA
 *   Olivier Devillers - Sylvain Pion - Monique Teillaud
 *   $Id: root2.H,v 1.1 2005/07/11 10:39:51 mourrain Exp $
 ********************************************************************/
#ifndef SYNAPS_ARITHM_ALGEBRAIC2_H
#define SYNAPS_ARITHM_ALGEBRAIC2_H

/********************************************************************/
#include <synaps/init.h>
#include <synaps/arithm/algebraic.h>


__BEGIN_NAMESPACE_SYNAPS

//--------------------------------------------------------------------
template<class Coef> 
struct algebraic2_rep : public algebraic_rep<Coef>
{
  //protected:
  Coef A,B,C;  //A*x^2+B*x+C;
  bool left; 

public:
  algebraic2_rep() : A(1),B(0),C(-1),left(true) {}
  algebraic2_rep(const algebraic_rep<Coef> & x) 
    : A(x.A),B(x.B),C(x.C),left(x.left) {}
  algebraic2_rep(const Coef & a, const Coef & b, const Coef & c, bool f)
    : algebraic_rep<Coef>(), A(a), B(b), C(c), left(f) 
    {
      // should this assertion be left ?
      // now we have polynomials of degree 1 and 0...
      assert(a != Coef(0) ); 
      assert(b*b-a*c*Coef(4)>=Coef(0));
      if (a < Coef(0)) 
	{
	  A=-a; B=-b; C=-c;
	}
    };

  virtual ~algebraic2_rep<Coef>() {} ;

  unsigned improve()    { return 0; }

  int compare(algebraic_rep<Coef>* a);

  int degree() const
  {
    if (A != Coef(0))  return 2;
    if (B != Coef(0))  return 1;
    if (C != Coef(0))  return 0;
    return -1;
  }

  void print(std::ostream & os) const
    {
      if(left) os<<"Left"; else os<<"Right";
      os<<" root of "<<A<<"*x^2";
      if(B>Coef(0)) os<<"+"; if(B !=Coef(0)) os<<B<<"*x";
      if(C>Coef(0)) os<<"+"; if(C !=Coef(0)) os<<C;
    }
};
//--------------------------------------------------------------------
//Class for the representation of square roots of algebraic numbers.
template<class Coef>
struct algebraic2 
{
protected:
  algebraic2_rep<Coef>* _rep;
public:  
  algebraic2<Coef>() : _rep(NULL) {}
  algebraic2<Coef>(const algebraic2<Coef> & x)
    : _rep(x._rep) {}
  algebraic2<Coef>(const Coef & a, const Coef & b, const Coef & c, bool f)
    : _rep(new algebraic2_rep<Coef>(a,b,c,f)) {}
  algebraic2<Coef>(const Coef & r)
    : _rep(new algebraic2_rep<Coef>(Coef(1),Coef(0),-(r*r),r>Coef(0)?false:true)) {}
  // degree 1
  algebraic2<Coef>(const Coef& a, const Coef& b)
    : _rep(new algebraic2_rep<Coef>(Coef(0),a>Coef(0)?a:-a,a>Coef(0)?b:-b,true)) {}

  algebraic2_rep<Coef>* rep      () {return _rep;}
  algebraic2_rep<Coef>* operator-> () const  {return _rep;}
  inline operator algebraic<Coef> () { return algebraic<Coef>(_rep); }


  algebraic2<Coef> operator-() const
  {
    return algebraic2<Coef>(_rep->A, -_rep->B, _rep->C, !_rep->left);
  }
  bool which() const {return _rep->left;}
};

//--------------------------------------------------------------------
template<typename Coef>
int algebraic2_rep<Coef>::compare(algebraic_rep<Coef>* a)
{
  assert(has_type<algebraic2_rep<Coef> >(a));
  algebraic2_rep<Coef>* a2 = dynamic_cast< algebraic2_rep<Coef>* >(a);
  
  // degree 1
  if ( A == Coef(0) )
  {
    if ( a2->A == Coef(0) )
      return compare_11_11(this,a2);
    if(a2->left)
      return compare_12_11(this,a2);
    return compare_12_12(this,a2);
  }

  if ( a2->A == Coef(0) )
  {
     if (left)
       return compare_21_11(this,a2);
     return compare_21_21(this,a2);
  }

  // degree 2
  if(left)
    if(a2->left)
     return compare_l_l(this,a2);
    else
     return compare_l_r(this,a2);
  else
    if(a2->left)
     return compare_r_l(this,a2);
    else
     return compare_r_r(this,a2);
}
//--------------------------------------------------------------------
#ifndef CGAL_NTS
#define CGAL_NTS 
#endif
template <class FT>
inline FT square(const FT  & A)
{ return CGAL_NTS A*A; }

template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcA(const FT& p, const FT& q)
{return CGAL_NTS square(p) + CGAL_NTS square(q);}

template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcB(const FT& p, const FT& q, const FT& s,
                const FT& alpha, const FT& beta)
{return CGAL_NTS square(q)*alpha-p*(s+q*beta);}
 
template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcC(const FT& q, const FT& s,
                const FT& alpha, const FT& beta, const FT& gamma)
{return s*(s+FT(2)*q*beta)+
   (CGAL_NTS square(q))*(CGAL_NTS square(alpha)+CGAL_NTS square(beta)-gamma);}
 
template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcI(const FT& A, const FT& B, const FT& C)
{ return CGAL_NTS square(B)-A*C; }
 
template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcJ(const FT& A1, const FT& B1, const FT& A2, const FT& B2)
{ return A1*B2-A2*B1; }
 
template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcK(const FT& A1, const FT& B1, const FT& C1,
                const FT& A2, const FT& B2, const FT& C2)
{ return C1*A2+A1*C2-FT(2)*B1*B2; }
 
template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcJp(const FT& B1, const FT& C1, const FT& B2, const FT& C2)
{ return B1*C2-C1*B2; }
 
template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcP4(const FT& J,  const FT& Jp,
                 const FT& A1, const FT& C1,
                 const FT& A2, const FT& C2)
{ return CGAL_NTS square(FT(A1*C2-C1*A2))-FT(4)*J*Jp;}
 
template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcP4(const FT& I1, const FT& I2, const FT& K)
{ return CGAL_NTS square(K)-I1*I2*4;}
 
template <class FT>
/*CGAL_NO_FILTER*/
inline FT calcD(const FT& A1, const FT& I1, const FT& A2, const FT& I2)
{ return I1*CGAL_NTS square(A2)-I2*CGAL_NTS square(A1);}
//--------------------------------------------------------------------
#ifndef CGAL_ENUM_H
namespace CGAL {
 typedef enum {NEGATIVE = -1,ZERO,POSITIVE } Sign;
 typedef enum {LARGER =1, EQUAL=0, SMALLER=-1} Comparison_result;
}
#endif

// degree 1
template <class Coef>
int compare_11_11(algebraic2_rep<Coef>* a, algebraic2_rep<Coef>* b)
{
  // Compares roots of (A1 X + B1) and (A2 X + B2).
  assert( a->B > Coef(0) && b->B > Coef(0) );
  return CGAL_NTS compare(b->C*a->B,a->C*b->B);
}

//degree 1
template <class Coef>
int compare_12_11(algebraic2_rep<Coef>* a, algebraic2_rep<Coef>* b)
{
   // First, we compare the root of a to the root of the derivative of b.

  int cmp = CGAL_NTS compare(b->B*a->B,a->C*b->A*Coef(2));
  if (cmp == CGAL::LARGER)
    return CGAL::LARGER;

  // If it doesn't work, we evaluate the sign of P2 at the root of P1.

  Coef p2 = a->C * (a->B*b->B - b->A*a->C) - b->C * CGAL_NTS square(a->B);

  return (int)CGAL_NTS sign(p2);
}

//degree 1
template <class Coef>
int compare_12_12(algebraic2_rep<Coef>* a, algebraic2_rep<Coef>* b)
{
   // boils down to the previous case by doing X <-> -X,
   // and changing the sign of the result.

  int cmp = CGAL_NTS compare(-(b->B*a->B),-(a->C*b->A*Coef(2)));
  if (cmp == CGAL::LARGER)
    return CGAL::SMALLER;

  // If it doesn't work, we evaluate the sign of P2 at the root of P1.

  Coef p2 = (a->C) * (a->B*b->B - b->A*a->C) - b->C * CGAL_NTS square(a->B);

  return -((int)CGAL_NTS sign(p2));
}

//degree 1
template <class Coef>
inline
int compare_21_11(algebraic2_rep<Coef>* a, algebraic2_rep<Coef>* b)
{
  return -compare_12_11(b,a);
}

//degree 1
template <class Coef>
inline
int compare_21_21(algebraic2_rep<Coef>* a, algebraic2_rep<Coef>* b)
{
  return -compare_12_12(b,a);
}

template<typename FT>
int compare_l_l(algebraic2_rep<FT>* a, algebraic2_rep<FT>* b)
{
  //  cout <<"Compare l l"<<endl;
  FT A1=a->A*2, B1=-a->B, C1=a->C*2;
  FT A2=b->A*2, B2=-b->B, C2=b->C*2;

  
  // Compares the smaller root of (A1 X^2 -2B1 X + C1)
  //       to the smaller root of (A2 X^2 -2B2 X + C2)
  FT J = calcJ(A1,B1,A2,B2);
  FT K = calcK(A1,B1,C1,A2,B2,C2);
 
  if (J > FT(0)){
        if (K > FT(0)) return CGAL::SMALLER;  // l1 < l2
 
        FT I1= calcI(A1,B1,C1);
        FT I2= calcI(A2,B2,C2);
        FT D = calcD(A1,I1,A2,I2);
 
        if (D > FT(0)) return CGAL::SMALLER;  // l1 < l2
 
        FT Jp = calcJp(B1,C1,B2,C2);
 
        if (Jp < FT(0)) return CGAL::LARGER;   // l1 > l2
 
        FT P4 = calcP4(I1,I2,K);

// 	if (P4<0)return CGAL::LARGER;   // l1 > l2
// 	if (P4>0)return CGAL::SMALLER;  // l1 < l2
// 	return CGAL::EQUAL;
	return ((int) CGAL_NTS sign(P4));
    }
    else{ // J<0
        if (K > FT(0)) return CGAL::LARGER;   // l1 > l2
 
        FT I1= calcI(A1,B1,C1);
        FT I2= calcI(A2,B2,C2);
        FT D = calcD(A1,I1,A2,I2);
 
        if (D < FT(0)) return CGAL::LARGER;   // l1 > l2
 
        FT Jp = calcJp(B1,C1,B2,C2);
 
        if (Jp > FT(0)) return CGAL::SMALLER;  // l1 < l2
 
        FT P4 = calcP4(I1,I2,K);
 
//     if (P4<0)return CGAL::LARGER;   // l1 > l2
//     if (P4>0)return CGAL::SMALLER;  // l1 < l2
//     return CGAL::EQUAL;
	return -((int) CGAL_NTS sign(P4));
    }
}



template<typename FT>
int compare_r_l(algebraic2_rep<FT>* a, algebraic2_rep<FT>* b)
{
  //  cout <<"Compare r l"<<endl;
  FT A1=a->A*2, B1=-a->B, C1=a->C*2;
  FT A2=b->A*2, B2=-b->B, C2=b->C*2;

  // Now compares the larger root of (A1 X^2 -2B1 X + C1)
  //          to the smaller root of (A2 X^2 -2B2 X + C2)
  FT J = calcJ(A1,B1,A2,B2);
  //  cout <<"J  "<<J<<endl;
  if ( J < FT(0) ) return CGAL::LARGER;   // r1 > l2
  
  FT K = calcK(A1,B1,C1,A2,B2,C2);
  //  cout <<"K  "<<K<<endl;

  if ( K < FT(0) ) return CGAL::LARGER;   // r1 > l2
 
  FT Jp = calcJp(B1,C1,B2,C2);
  //  cout <<"Jp "<<Jp<<endl;

  if ( Jp < FT(0) ) return CGAL::SMALLER;  // r1 < l2
 
  FT P4 = calcP4(J,Jp,A1,C1,A2,C2);
  //  cout <<"P4 "<<P4<<endl;

  return -((int) CGAL_NTS sign(P4));
//   if ( P4< FT(0) ) return CGAL::LARGER;   // r1 > l2
//   if ( P4> FT(0) ) return CGAL::SMALLER;  // r1 < l2
//   return CGAL::EQUAL;
}

template<typename FT>
int compare_l_r(algebraic2_rep<FT>* a, algebraic2_rep<FT>* b)
{ 
  //  cout <<"Compare l r"<<endl;
  return -compare_r_l(b,a);
}

template<typename FT>
int compare_r_r(algebraic2_rep<FT>* a, algebraic2_rep<FT>* b)
{
  //   cout <<"Compare r r"<<endl;
  FT A1=a->A*FT(2), B1=-a->B, C1=a->C*FT(2);
  FT A2=b->A*FT(2), B2=-b->B, C2=b->C*FT(2);

  FT J = calcJ(A1,B1,A2,B2);
  FT K = calcK(A1,B1,C1,A2,B2,C2);
    
  if (J>0){
    if (K>0) return CGAL::SMALLER;   // r1 < r2   1,2
	
    FT I1= calcI(A1,B1,C1);
    FT I2= calcI(A2,B2,C2);
    FT D = calcD(A1,I1,A2,I2);

    if (D<0) return CGAL::SMALLER;   // r1 < r2   2,3b

    FT Jp = calcJp(B1,C1,B2,C2);

    if (Jp<0) return CGAL::LARGER;    // r1 > r2   3a

    FT P4= calcP4(I1,I2,K);

//     if (P4<0)return CGAL::SMALLER;    // r1 > r2   3a
//     if (P4>0)return CGAL::LARGER;   // r1 < r2   2
//     return CGAL::EQUAL;
    return ((int) CGAL_NTS sign(P4));
  }
  else{ // J<0
    if (K>0) return CGAL::LARGER;   // r1 > r2    4,5
    
    FT I1= calcI(A1,B1,C1);
    FT I2= calcI(A2,B2,C2);
    FT D = calcD(A1,I1,A2,I2);

    if (D>0) return CGAL::LARGER;   // r1 > r2    3a,4

    FT Jp = calcJp(B1,C1,B2,C2);

    if (Jp>0)return CGAL::SMALLER;   // r1 < r2     3b

    FT P4= calcP4(I1,I2,K);
	
//     if (P4<0)return CGAL::LARGER;   // r1 > r2     4
//     if (P4>0)return CGAL::SMALLER;  // r1 < r2     3b
//     return CGAL::EQUAL; 
    return -((int) CGAL_NTS sign(P4));
  }
}
//--------------------------------------------------------------------
template<class Coef>
inline
CGAL::Comparison_result compare(algebraic2<Coef> a,algebraic2<Coef> b)
{
  algebraic<Coef> B= b;
  return static_cast<CGAL::Comparison_result> (a.rep()->compare(B.rep()));
}

template<class Coef>
inline
CGAL::Comparison_result compare(algebraic2<Coef> a,const Coef &b)
{
  return compare(a, algebraic2<Coef>(b));
}
//--------------------------------------------------------------------
template<class Coef>
inline
CGAL::Sign sign(algebraic2<Coef> a)
{
  return static_cast<CGAL::Sign>((int) compare(a,Coef(0)));
}
//--------------------------------------------------------------------
template<class NT> 
bool operator<(algebraic2<NT>  a, algebraic2<NT> b)
{
  algebraic<NT> B= b;
  return (a.rep()->compare(B.rep())<0);
}
//--------------------------------------------------------------------
template<class NT> 
std::ostream & operator<<(std::ostream & os,  algebraic2<NT> a)
{
  a.rep()->print(os); return os;
}
//--------------------------------------------------------------------

template < typename NT >
algebraic2<NT>
square(const algebraic2<NT> &a)
{
  // It's easy to get the explicit formulas for the square of the two roots.
  // Then it's easy to compute their sum and their product, which gives the
  // coefficients of the polynomial (X^2 - Sum X + Product).

  NT a_sqr = CGAL_NTS square(a->A);
  NT b_sqr = CGAL_NTS square(a->B);
  NT c_sqr = CGAL_NTS square(a->C);
  NT s = NT(2) * a->C * a->A - b_sqr;

  bool b = (a->B>NT(0) ? !a.which() : a.which());

  return algebraic2<NT>(a_sqr, s, c_sqr, b);
}
//--------------------------------------------------------------------
template < typename NT >
algebraic2<NT>
operator-(const algebraic2<NT> a, const NT& b)
{
  if (a->degree() == 1)
    return algebraic2<NT>(a->B, a->C + b * a->B);

  assert(a->degree() == 2);

  // It's easy to see it using the formula (X^2 - sum X + prod).
  NT p = a->C + b * a->B + a->A * CGAL_NTS square(b);
  NT s = a->B + b * a->A*2;

  return algebraic2<NT>(a->A, s, p, a.which());
}
//--------------------------------------------------------------------
template < typename NT >
inline
algebraic2<NT>
operator-(const NT &a, const algebraic2<NT> b)
{
  return -(b-a);
}
//--------------------------------------------------------------------
template < typename NT >
inline
algebraic2<NT>
operator+(const algebraic2<NT> a, const NT& b)
{
  return a - NT(-b);
}
//-------------------------------------------------------------------- 
template < typename NT >
inline
algebraic2<NT>
operator+(const NT &a, const algebraic2<NT> b)
{
  return b - NT(-a);
}
//-------------------------------------------------------------------- 
template <class Coef>
algebraic2<Coef> operator*(algebraic2<Coef> a,const Coef& b)
{
  int degree = a->degree();

  if (degree == 1)
    return algebraic2<Coef>(a->B, a->C*b);

  assert(a->degree() == 2);

  return algebraic2<Coef>(a->A, a->B * b,a->C*CGAL_NTS square(b), b<Coef(0)?!a.which():a.which());
}

//thAAAnos added for CGAL compatibility
template <class Coef>
inline
algebraic2<Coef> operator*(const Coef& b,algebraic2<Coef> a)
{
  return a*b;
}
//-------------------------------------------------------------------- 
template < typename NT >
bool
is_valid(algebraic2<NT> r)
{
  int d = r->degree();
  if (d == 1)
    return r->B > NT(0);
  if (d == 2)
    return (r->B*r->B-r->A*r->C*NT(4)>=NT(0));
  return false;
}
//-------------------------------------------------------------------- 
template<class Coef>
double to_double(algebraic2<Coef> x)
{
  assert(is_valid(x));

  int i = x->degree();
  if (i == 2)
  {
    double a = CGAL::to_double(x->A);
    double b = CGAL::to_double(x->B);

    double d = std::sqrt(CGAL::to_double(calcP4(x->A,x->C,x->B)));

    assert(CGAL_NTS sign(a) == CGAL::POSITIVE);
    if (x.which())
      d = -d;

    return (d-b)/(2.0*a);
  }
  assert(i == 1);
  double a = CGAL::to_double(x->A);
  double b = CGAL::to_double(x->B);
  return -b/a;
}


__END_NAMESPACE_SYNAPS

/********************************************************************/
#endif // SYNAPS_ARITHM_ALGEBRAIC2_H

