// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>
//           Athanasios Kakargias <grad0460@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)
//

#ifndef SYNAPS_ARITHM_ALGEBRAIC_ROOT_OF_H
#define SYNAPS_ARITHM_ALGEBRAIC_ROOT_OF_H

#include "Algebraic_config.h"
#include <synaps/upol/UPolDse.h>
#include <synaps/usolve/Dichotomy.h>

 
__ALGEBRAIC_BEGIN_NAMESPACE
/**! Representation of a real algebraic number.\\
 * The algebraic number is a root of a polynomial
 * $P(X) = \sum_{i=0}^{n}{c_i X^i}$.
 *
 * It is represented by a square-free polynomial and an isolating interval.
 *
 *
 * \todo We have to make all the constructors private!
 * \todo Add Polynomial traits.
 *
 * \sa NumberTraits, Interval
 *
 **/
template< typename RT_,
          typename POLY_ = UPolDse< RT_ > > 
class root_of
{
public:
  typedef NumberTraits<RT_>           NTR;
  typedef typename NTR::RT             RT;
  typedef typename NTR::FT             FT;
  typedef typename NTR::XT             XT;
  typedef typename NTR::FIT            FIT;
  typedef typename NTR::XIT            XIT;

  typedef POLY_                        POLY;
  typedef unsigned                     size_type;
  typedef RT                           value_type;
  typedef root_of<RT_, POLY>           self_t;


private:
  mutable POLY        C_;            ///< Coefficients
  mutable int         rootnb_;       ///< root index
  mutable RT          D_;            ///< discriminant
  mutable FIT         interval_;     ///< Isolating interval

  int                 sign_left_;    ///< Sign at the left endpoint
  int                 sign_right_;   ///< Sign at the right endpoint
  unsigned int        mu_;           ///< The multiplicity in the original equation
  bool                valid_;        ///< Is it a valid real algebraic number?


  void compute_endpoints_signs()
  {
    sign_left_  = sign(UPOLDAR::eval_horner(C_, left()));
    sign_right_ = opposite(sign_left_);
    //assert(sign_right_ == sign(UPOLDAR::eval_horner(C_, right())));
  }


  //! The leading coeff of the polynomial must always be positive
  void make_lcoeff_positive()
  {
    if (UPOLDAR::lcoeff(C_) < RT_(0)) {
      std::transform(C_.begin(), C_.end(), C_.begin(), std::negate<RT_>());
    }
    return;
  }

  //! Check if we can change one endpoint to zero
  void refine_isolating_interval()
  {
    if (sign(left()) * sign(right()) > 0) return;
    int sign_at_zero = sign(C_[0]);
    if (sign_at_zero == ZERO) {
      unsigned old_mu = mu_;
      *this = root_of();
      this->mu_ = old_mu;
      return;
    }
    if (sign_at_zero == sign_low()) {
      set_interval(0, right());
      return;
    }
    set_interval(left(), 0);
    return;
  }

  void check_for_rational() const
  {
    using namespace let;
    if (!singleton(interval_) && (UPOLDAR::degree(C_) == 1)) {
      FT x = to_FT<FT>( RT_(-C_[0]), C_[1] );
      set_interval(x, x);
      set_index(0);
      return;
    }
    if (singleton(interval_) && (UPOLDAR::degree(C_) != 1)) {
      POLY f (2, AsSize());
      assign(f[1],denominator(left()));
      assign(f[0], FT(-numerator(left())));
      C_ = f;
      set_index(0);
    }
  }

public:
  //------------- Constructors --------------------------------//
  root_of()
    : C_(2, AsSize()),
      rootnb_(0),
      D_(RT(1)),
      interval_(),
      sign_left_(ZERO),
      sign_right_(ZERO),
      mu_(1),
      valid_(true)
  {
    C_[0] = RT_(0);
    C_[1] = RT_(1);
    assert(is_valid());
  }


  root_of( const POLY& C,
	   const int& rootnb,
	   const RT& D,
	   const FT& left,
	   const FT& right,
	   const unsigned& mu = 1)
    : C_(C),
      rootnb_(rootnb),
      D_(D),
      interval_(left, right),
      mu_(mu),
      valid_(true)
  {
    make_lcoeff_positive();
    compute_endpoints_signs();
    refine_isolating_interval();
    check_for_rational();

    assert(is_valid());
  }

  root_of(const POLY& C, const FT& left, const FT& right, int rootnb = 0)
    : C_(C),
      rootnb_(rootnb),
      D_((UPOLDAR::degree(C) == 2) ? RT(C[1] * C[1] - 4 * C[0] * C[2]) : 1),
      interval_(left, right),
      mu_(1),
      valid_(true)
  {
    make_lcoeff_positive();
    compute_endpoints_signs();
    refine_isolating_interval();
    check_for_rational();

    assert(is_valid());
  }

  //-----------------------------------------------------------------------//

  root_of& negate()
  {
    for(size_type i = 1; i < C_.size(); i += 2) C_[i]= -C_[i];

    make_lcoeff_positive();
    set_interval(-right(), -left());
    if (degree() == 2) {
      rootnb_ = (rootnb_ == 0 ? 1: 0);
    }
    assert(is_valid());
    return *this;
  }

  //--------- Operators with RT ----------------------------------

  root_of& operator +=( const RT& a)
  {
    UPOLDAR::shift(C_, RT(-a));
    interval_ += a;
    return *this;
  }

  root_of& operator -=( const RT& a)
  {
    UPOLDAR::shift(C_, a);
    interval_ -= a;
    return *this;
  }

  root_of& operator *=( const RT& a)
  {
    if (a == RT(0)) {
      *this = root_of();
      return *this;
    }
    RT t(1);
    for (int i = degree(); i >= 0; --i) {
      C_[i] *= t;
      t *= a;
    }
    interval_ *= a;

    if ((a < RT(0)) && (degree() == 2)) {
      rootnb_ = (rootnb_ == 0 ? 1: 0);
    }
    return *this;
  }

  root_of& operator /= (const RT& a)
  {
    assert(a != 0);
    RT t(1);
    for (int i = 0; i <= degree(); ++i) {
      C_[i] *= t;
      t *= a;
    }
    interval_ /= a;
    if ((a < RT(0)) && (degree() == 2)) {
      rootnb_ = (rootnb_ == 0 ? 1: 0);
    }
    return *this;
  }

  //-------------- Operators with FT ------------------------------

  root_of& operator +=( const FT& a)
  {
    C_ = translate(C_, FT(-a));
    interval_ += a;
    return *this;  
  }

  root_of& operator -=( const FT& a)
  {
    C_ = translate(C_, a);
    interval_ -= a;
    return *this;  
  }

  root_of& operator *=( const FT& a)
  {
    RT n = numerator(a);
    RT d = denominator(a);
    
    RT t(1);
    for (int i = 0; i <= degree(); ++i) {
      C_[i] *= t;
      t *= d;
    }
    t = RT(1);
    for (int i = degree(); i >= 0; --i) {
      C_[i] *= t;
      t *= n;
    }

    interval_ *= a;
    if ((a < 0) && (degree() == 2)) {
      rootnb_ = (rootnb_ == 0 ? 1: 0);
    }
    return *this;
  }
  

  /**
   * @return The polynomial of the algebraic number
   */
  POLY  poly()             const { return C_; }

  /**
   * @brief Returns the @i@ coefficient of the
   * polynomial of the algebraic number.
   *
   * @param i
   */
  RT    operator[](int i)  const { return C_[i]; }

  /**
   * @brief Computes the degree of the algebaric number
   *
   */
  int  degree()               const { return UPOLDAR::degree(poly()); }

  /**
   * @brief Computes the size of the polynomial of the alg number.
   */
  size_type  size()           const { return degree() + 1; }

  /**
   * @brief Return the root index.
   */
  int  which()                const { return rootnb_; }

  /**
   * @brief  Computes (returns) the disciminant of the polynomial
   *
   * Actually we need the discriminant only for algebraic numbers
   * of degree 2. We will eliminate this member function in the future.
   */
  RT  discriminant()          const { return D_; }

  /**
   * @brief The isolating interval
   */
  FIT interval()  const { return interval_; }

  /**
   * @brief The left endpoint of the isolating interval
   */
  FT  left()      const { return interval_.lower(); }
  /**
   * @brief The right endpoint of the isolating interval
   */
  FT  right()     const { return interval_.upper(); }

  /**
   * @brief Change the isolating interval
   *
   * @param l
   * @param r
   */
  void set_interval(const FT& l, const FT& r) const
  {
    interval_.assign(l, r);
    check_for_rational();
  }

  /**
   * @brief Change the isolating interval
   *
   * @param a
   */
  void set_interval(const FIT& a) const
  { interval_.assign(a); }


  /**
   * @brief The sign of the polynomial evaluated ove the right endpoint
   * of the isolating interval.
   *
   */
  int sign_up() const
  { return sign_right_; }

  /**
   * @brief The sign of the polynomial evaluated over the left endpoint
   * of the isolating interval.
   *
   */
  int sign_low() const
  { return sign_left_; }
  ////

  /**
   * @brief Is the algebraic number a rational?
   *
   */
  bool is_rational()  const
  {
    return left() == right();
  }


  /// \brief Return the multiplicity of the algebraic number.
  /// Note that the multiplicity is computed in the original polynomial.
  unsigned multiplicity() const { return mu_; }

  void set_multiplicity(unsigned mu) { mu_ = mu; }

  void refine_interval() const
  {
    if (singleton(interval())) {
	if ( degree() == 1 ) return;
	POLY f( 2, AsSize());
	f[1] = denominator( left());
	f[0] = -numerator( left());
	C_ = f;
	assert( UPOLDAR::degree(C_) == 1);
	
	return;
    }
    
    FT m = median(interval());
    int sgn = sign(UPOLDAR::eval_horner(poly(), m));
    if (sgn == ZERO) {
      set_interval(m, m);
      POLY f( 2, AsSize());
      f[1] =  denominator( left());
      f[0] =  -numerator( left());
      C_ = f;
      assert( UPOLDAR::degree(C_) == 1);
      return;
    }
    if (sgn == sign_low()) set_interval(m, right());
    else set_interval(left(), m);
  }


  int get_index() const { return rootnb_; }

  void set_index( int idx) const { rootnb_ = idx; }

  /**
   * @brief Do we have a valid alegraic number ?
   *
   * \todo Eliminate this function
   */
  bool is_valid()
  {
    valid_ = ( (UPOLDAR::lcoeff(C_) > RT(0)) && 1
//	       (rootnb_ < degree())
	       );
    return valid_;
  }

  virtual ~root_of() { }
};


template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator +(const root_of<RT, POLY>& r, const RT& a)
{
  return root_of<RT, POLY>(r) += a;
}


template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator +(const RT& a, const root_of<RT, POLY>& r)
{
  return root_of<RT, POLY>(r) += a;
}



template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator -(const root_of<RT, POLY>& r, const RT& a)
{
  return root_of<RT, POLY>(r) -= a;
}


template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator -(const RT& a, const root_of<RT, POLY>& r)
{
  return root_of<RT, POLY>(r).negate() += a;
}



template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator *(const root_of<RT, POLY>& r, const RT& a)
{
  return root_of<RT, POLY>(r) *= a;
}


template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator *(const RT& a, const root_of<RT, POLY>& r)
{
  return root_of<RT, POLY>(r) *= a;
}


template < typename RT,
	  typename POLY >
root_of<RT, POLY>
operator /(const root_of<RT, POLY>& r, const RT& a)
{
  return root_of<RT, POLY>(r) /= a;
}


//--------- FT Operators---------------

template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator +(const root_of<RT, POLY>& r, const typename root_of<RT>::FT& a)
{
  return root_of<RT, POLY>(r) += a;
}

template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator +(const typename root_of<RT>::FT& a, const root_of<RT, POLY>& r)
{
  return root_of<RT, POLY>(r) += a;
}


template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator -(const root_of<RT, POLY>& r, const typename root_of<RT>::FT& a)
{
  return root_of<RT, POLY>(r) -= a;
}


template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator -(const typename root_of<RT>::FT& a, const root_of<RT, POLY>& r)
{
  return root_of<RT, POLY>(r).negate() += a;
}


template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator *(const root_of<RT, POLY>& r, const typename root_of<RT>::FT& a)
{
  return root_of<RT, POLY>(r) *= a;
}


template < typename RT,
	   typename POLY >
root_of<RT, POLY>
operator *(const typename root_of<RT>::FT& a, const root_of<RT, POLY>& r)
{
  return root_of<RT, POLY>(r) *= a;
}



/**
 * @brief Double approximation of the algebraic number
 *
 * @param a
 *
 * \sa Dichotomy
 */
template < typename RT,
	   typename POLY >
double
to_double(const root_of<RT, POLY>& r)
{
  typedef typename root_of<RT, POLY>::FT    FT;

  UPolDse< double > f(r.size(), AsSize());
  for (int i = 0; i <= r.degree(); i++)   f[i] = to_double(r[i]);

  std::vector< double > s;
  s.reserve(r.degree());
    
  int d = r.degree();
  if ( d == 1 ) {
    return - f[0] / f[1];
  }

  if ( d == 2 ) {    
    double D = std::sqrt(to_double(r.discriminant()));

    assert(f[2] > 0);

    if ( ! r.which() )
      D = -D;

    return ( D - f[1] ) / ( f[2] * 2 );
  }

  double left = to_double( r.left() );
  double right = to_double( r.right() );
  double appx_sol =  RootOf(f, left, right, Dichotomy(1e-6));
  return ( appx_sol );
}



/**
 * @brief Print the algebraic number
 *
 */
template < typename RT, typename POLY>
std::ostream& print(std::ostream& os, const root_of<RT, POLY>& r)
{
  os << "root_of( ";
  UPOLDAR::print(os, r.poly());
  os << ", " << r.which() << " )";
  os << " in " << r.interval();
  return os;
}

template < typename RT, typename POLY>
std::ostream& operator<<(std::ostream& output, const root_of<RT, POLY>& r)
{
  print(output, r);
  output << ", approx = " << to_double(r);
  return output;
  //return output << to_double(r);
}


__ALGEBRAIC_END_NAMESPACE

__BEGIN_NAMESPACE_SYNAPS
namespace let
{
  template<class C, class P> inline
  void assign( double & d, const ALGEBRAIC::root_of<C,P> & r )
  {
    d = to_double(r);
  };
};
__END_NAMESPACE_SYNAPS

#endif // SYNAPS_ARITHM_ALGEBRAIC_ROOT_OF_H


