// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>
//           Athanasios Kakargias <grad0460@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)
//
// 27/11/2003: Modification mourrain@sophia.inria.fr to have only
//             one class for the algebraic numbers.

/*! \file sign_fct.H
 *
 * Functions that compute the sign of an algebraic number, and
 * the sign of a polynomial evaluated over an algebraic number.
 *
 * \sa root_of
 *
*/



#ifndef SYNAPS_USOLVE_ALGEBRAIC_SIGN_FCT_H
#define SYNAPS_USOLVE_ALGEBRAIC_SIGN_FCT_H

#include <synaps/usolve/algebraic/root_of.h>
#include <synaps/usolve/algebraic/Comparison.h>
#include <synaps/usolve/algebraic/sign_22.h>
#include <synaps/usolve/algebraic/sign_33.h>
#include <synaps/usolve/algebraic/sign_34.h>
#include <synaps/usolve/algebraic/sign_43.h>
#include <synaps/usolve/algebraic/sign_44.h>
#include <synaps/upol/SturmSeq.h>

__ALGEBRAIC_BEGIN_NAMESPACE

template < typename RT,
	       typename P >  inline
int sign(const root_of<RT, P>& r)
{
	int dg = r.degree();
	if (dg == 1) return sign(-r[0]);
	if (r.left() < RT(0))
		return NEGATIVE;
	return POSITIVE;
}


template < typename POLY,
           typename RO_t >  inline
void
in_isolating_interval(
		const POLY& f,
		const RO_t& x,
		int& where
		)
{
  typedef typename RO_t::NTR   NTR;
  typedef typename RO_t::RT    RT;
  typedef typename RO_t::FT    FT;

  print(__FUNCTION__);
#if 0
  FT I2 = x.right_isolating_point();
  FT I1 = x.left_isolating_point();
  //	cout << a.which() << " " << b.which() << endl;
  int sgn_x_I2 = ((x.degree()-1-x.which()) % 2 ? NEGATIVE : POSITIVE);
  int sgn_x_I1 = opposite(sgn_x_I2);
  if (sgn_x_I2 != NEGATIVE) where ^= 1;
  if (sgn_x_I1 != NEGATIVE) where ^= 2;

  int sgn_f_I2 = sign(eval(f, I2));
  if (sgn_f_I2 != NEGATIVE) where ^= 4;
  int sgn_f_I1 = sign(eval(f, I1));
  if (sgn_f_I1 != NEGATIVE) where ^= 8;
#endif

  if (x.sign_up() != NEGATIVE)  where ^= 1;
  if (x.sign_low() != NEGATIVE) where ^= 2;
  if (sign( UPOLDAR::eval_horner(f, x.right()) ) != NEGATIVE) where ^= 4;
  if (sign( UPOLDAR::eval_horner(f, x.left()) )  != NEGATIVE) where ^= 8;
//  cout << sgn_x_I1 << " " << sgn_x_I2 << endl;
//  cout << sgn_f_I1 << " " << sgn_f_I2 << endl;

}

// Compute the sign of evaluation of a polynomial over an algebraic number




template < typename POLY,
           typename RO_t >  inline
int
sign_at_x1(const POLY& f, const RO_t& x)
{
  typedef typename RO_t::NTR    NTR;

  print(__FUNCTION__);
  return sign(UPOLDAR::eval_horner(f, x.left()));
}



template < typename POLY,
           typename RO_t >  inline
int
sign_at_1x(const POLY& f, const RO_t& x)
{
  typedef typename RO_t::NTR   NTR;
  print(__FUNCTION__);

  return opposite(compare(to_FT(-f[0], f[1]), x));
}



/** 
	* @brief Compute the sign of @f@ evaluated over @x@
	* 
	* @param f Univariate polynomial
	* @param x Algebraic number
	*/
template < typename POLY,
           typename RO_t >  inline
int
sign_at(const POLY& f, const RO_t& x)
{
	typedef typename RO_t::NTR     NTR;
	typedef typename RO_t::RT      RT;

	typedef int (* signFunc)(const POLY& f, const RO_t& x);

	const signFunc sgntab_[4][4] = {
		{ sign_at_x1<POLY, RO_t>, sign_at_1x< POLY, RO_t >, sign_at_1x< POLY, RO_t >, sign_at_1x< POLY, RO_t > },
		{ sign_at_x1<POLY, RO_t>, sign_at_22< POLY, RO_t >, sign_at_23< POLY, RO_t >, sign_at_24< POLY, RO_t > },
		{ sign_at_x1<POLY, RO_t>, sign_at_32< POLY, RO_t >, sign_at_33< POLY, RO_t >, sign_at_34< POLY, RO_t > },
		{ sign_at_x1<POLY, RO_t>, sign_at_42< POLY, RO_t >, sign_at_43< POLY, RO_t >, sign_at_44< POLY, RO_t > }
	};
	assert(x.degree() > 0);
//	assert(x.degree() < 5);

	int dgf = degree(f);
	if (dgf < 0) return ZERO;
	if (dgf == 0) return sign(f[0]);
	// The leading coeff must always be positive
	POLY G(f);
	make_lcoeff_positive(G);
	POLY P(G);
	int tmp1 = sign(f[degree(f)] );


	if (x.degree() > 4) {
	  //int sa = sign( UPOLDAR::eval_horner(f, x.left()) );
	  //int sb = sign( UPOLDAR::eval_horner(f, x.right()) );
	  //if (sa == sb) return sa;

	  SturmSeq<RT> s(x.poly(), G, SUBRES());
	  int v0 = variation(s, x.left()) - variation(s, x.right());
	  return tmp1 * v0 * x.sign_up();
	}

	if (degree(G) > 4) 
	  {
	    P = EUCLIDIAN::prem(G, x.poly());
	    if ((degree(P) <= 0) && (P[0] == 0)) return ZERO;

	    tmp1 *= sign(P[degree(P)]);
	    make_lcoeff_positive(P);
	    if (degree(P) == 0) 
	      return tmp1 * sign( P[degree(P)] );
	    return tmp1 * sgntab_[degree(P)-1][x.degree()-1](P, x);
	}

	return tmp1 * sgntab_[dgf-1][x.degree()-1](P, x);
}

__ALGEBRAIC_END_NAMESPACE
#endif // SYNAPS_ARITHM_ALGEBRAIC_SIGN_SIGN_FCT_H

