// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>
//           Athanasios Kakargias <grad0460@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for MPolys and Surfaces)
//
// 27/11/2003: Modification mourrain@sophia.inria.fr to have only
//             one typename for the algebraic numbers.
#ifndef SYNAPS_USOLVE_ALGEBRAIC_SIGN_SIGN_FCT_2_H
#define SYNAPS_USOLVE_ALGEBRAIC_SIGN_SIGN_FCT_2_H

#include <synaps/usolve/algebraic/Algebraic_config.h>
#include <synaps/usolve/algebraic/root_of.h>
#include <synaps/usolve/algebraic/sign_fct.h>
#include <synaps/usolve/algebraic/MPoly_2_2.h>

//----- Help functions
__BEGIN_NAMESPACE_SYNAPS

const unsigned POLY_wrt_Y = 1;
const unsigned POLY_wrt_X = 0;

namespace let {

  // if idx = 1 then result in (F[x])[y]
  // If idx = 0 then result in (F[y])[x]
  template< typename RT, typename O, typename R >
  UPolDse< UPolDse<RT> >
  convert_to_upoly_2(const MPol<RT, O, R>& p, unsigned idx = POLY_wrt_Y)
  {
    int d = MPOLDST::degree(p, idx);
    
    UPolDse< UPolDse<RT> > u(d+1,AsSize());
    for(typename MPol<RT, O, R>::const_iterator it = p.begin();
	it != p.end();
	++it)
      {
	u[ ((*it)[idx]) ] = u[ ((*it)[idx]) ] +  UPolDse<RT>(it->coeff(), (*it)[!idx]);
      }
    
    if (idx) {
      Variable< UPolDse<RT> >() = "x";
      Variable< UPolDse< UPolDse<RT> > >() = "Y";
    } else {
      Variable< UPolDse<RT> >() = "Y";
      Variable< UPolDse< UPolDse<RT> > >() = "x";
    }
    return u;
  }

  template < typename RT >
  UPolDse< UPolDse<RT> >
  convert_upoly_to_upoly_2(const UPolDse<RT>& f)
  {
    if (f.get_degree() == -1) {
      return UPolDse< UPolDse<RT> >();
    }
    UPolDse< UPolDse<RT> > result(f.get_degree() + 1, AsSize());
    for (unsigned i = 0; i <= f.get_degree(); ++i) {
      UPolDse<RT> tmp(1, AsSize());
      tmp[0] = f[i];
      result[i] = tmp;
    }
    return result;
  }

} // namespace let

//------End of Help Functions

__END_NAMESPACE_SYNAPS

//--------------------------------------------------------------------




/*! \file sign_fct_2.H
 *
 * Functions that compute the sign of bivariate polynomial of
 * total degree up to 2, evaluated over two algebraic numbers,
 * of degree up to 4.
 *
 */


__ALGEBRAIC_BEGIN_NAMESPACE



template < typename CONT >
int
Wvar(const CONT& C)
{
  int V = 0;
  typename CONT::const_iterator ci = C.begin();
  typename CONT::const_iterator pi = ci;
  ++ci;
  for (; ci != C.end(); ++ci) {
    if (*ci == 0) {
      if ( ci+1 == C.end() ) continue;
      if ( *(ci+1) != 0 )	continue;
      ++ci;
      if ( (ci+1) == C.end() ) continue;
      ++ci;
      if ( ((*ci) * (*pi)) < 0) {
	++V;
      }
      if ( ((*ci) * (*pi)) > 0) {
	V += 2;
      }
      pi = ci;
      continue;
    }
    if ( ((*ci) * (*pi)) < 0) {
      ++V;
    }
    pi = ci;
  }
  return V;
}




template < typename MPoly,
           typename RO_t >
int
sign_at_x1_cx2y2(const MPoly& Fxy, const RO_t& x, const RO_t& y)
{
  typedef typename RO_t::POLY   POLY;
  typedef typename RO_t::RT     RT;

  print(__FUNCTION__);

  //  cout << "x = "; print(cout, x); cout << x << endl;
  POLY f(3, AsSize());
  RT x1_2 = SQR(x[1]);
  f[2] = Fxy[4] * x1_2;
  f[1] = x[1] * (Fxy[1] * x[1] - Fxy[3] * x[0]);
  f[0] = (Fxy[5] * x[0] - Fxy[2] * x[1] ) * x[0] + Fxy[0] * x1_2;

  //  cout << "f = " << f << endl;
  return sign_at(f, y);
}


template < typename MPoly,
           typename RO_t>
int
sign_at_x2_cx2y2(const MPoly& Fxy, const RO_t& x, const RO_t& y)
{
  typedef typename RO_t::POLY   POLY;
  typedef typename RO_t::NTR    NTR;
  typedef typename NTR::RT      RT;
  typedef typename NTR::FT      FT;

  print(__FUNCTION__);

  int sgn_a = opposite((x.degree()-1-x.which()) % 2 ? NEGATIVE : POSITIVE);
  FT isol = x.left();
  int Va = StHa_x2_cx2y2(
			 Fxy,
			 x.poly(),
			 sgn_a,
			 numerator(isol),
			 denominator(isol),
			 y
			 );

  int sgn_b = opposite(sgn_a);
  isol = x.right();
  int Vb = StHa_x2_cx2y2(
			 Fxy,
			 x.poly(),
			 sgn_b,
			 numerator(isol),
			 denominator(isol),
			 y
			 );
  return (sgn_b - sgn_a)*(Va - Vb);
}


template < typename MPoly,
           typename RO_t >
int
sign_at_x3_cx2y2(const MPoly& Fxy, const RO_t& x, const RO_t& y)
{
  typedef typename RO_t::POLY   POLY;
  typedef typename RO_t::NTR    NTR;
  typedef typename NTR::RT      RT;
  typedef typename NTR::FT      FT;

  print(__FUNCTION__);

  int sgn_a = opposite((x.degree()-1-x.which()) % 2 ? NEGATIVE : POSITIVE);
  FT isol = x.left();
  int Va = StHa_x3_cx2y2(
			 Fxy,
			 x.poly(),
			 sgn_a,
			 numerator(isol),
			 denominator(isol),
			 y
			 );

  int sgn_b = opposite(sgn_a);
  isol = x.right();
  int Vb = StHa_x3_cx2y2(
			 Fxy,
			 x.poly(),
			 sgn_b,
			 numerator(isol),
			 denominator(isol),
			 y
			 );
  return (sgn_b - sgn_a)*(Va - Vb);
}


template < typename MPoly,
           typename RO_t >
int
sign_at_x4_cx2y2(const MPoly& Fxy, const RO_t& x, const RO_t& y)
{
  typedef typename RO_t::POLY   POLY;
  typedef typename RO_t::NTR    NTR;
  typedef typename NTR::RT      RT;
  typedef typename NTR::FT      FT;

  print(__FUNCTION__);

  int sgn_a = opposite((x.degree()-1-x.which()) % 2 ? NEGATIVE : POSITIVE);
  FT isol = x.left();
  // cout << "left isol = " << isol <<  " = " << to_double(isol) << endl;
  int Va = StHa_x4_cx2y2(
			 Fxy,
			 x.poly(),
			 sgn_a,
			 numerator(isol),
			 denominator(isol),
			 y
			 );

  int sgn_b = opposite(sgn_a);
  isol = x.right();
  //cout << "right isol = " << isol <<  " = " << to_double(isol) << endl;
  int Vb = StHa_x4_cx2y2(
			 Fxy,
			 x.poly(),
			 sgn_b,
			 numerator(isol),
			 denominator(isol),
			 y
			 );
  // cout << "Va - Vb = " << Va  << " - " << Vb << "=" << Va - Vb << endl;
  return (sgn_b - sgn_a)*(Va - Vb);
}




//----------------------------------------------------------------------------
template < typename MPoly,
           typename RT,
	   typename POLY >
std::vector<POLY>
compute_StHa_x1_cx2y2(
		      const MPoly& c,
		      const POLY& fx,
		      const RT& A, const RT& B)
{
  // The conic is of degree 2 in X
  // fx is of degree 2 in X
  //	RT c[6]  = {-2, 0, 0, 0, 1, 1};
  //  int sgn2, sgn3, sgn4;
  std::vector<POLY> p(2);
  p[0] = fx;
  p[1].resize(3);

  RT fx1_2 = SQR(fx[1]);
  p[1][2] = c[4] * fx1_2;
  p[1][1] = fx[1] * (c[1] * fx[1] - c[3] * fx[0]);
  p[1][0] = (c[5] * fx[0] - c[2] * fx[1] ) * fx[0] + c[0] * fx1_2;

  return p;
}

//----------------------------------------------------------------------------
template < typename MPoly,
           typename RT,
	   typename POLY >
std::vector<POLY>
compute_StHa_x2_cx2y2(
		      const MPoly& c,
		      const POLY& fx,
		      const RT& A, const RT& B)
{
  // The conic is of degree 2 in X
  // fx is of degree 2 in X
  //	RT c[6]  = {-2, 0, 0, 0, 1, 1};
  //  int sgn2, sgn3, sgn4;
  std::vector<POLY> p(4);
  p[0] = fx;
  p[1].resize(3);
  p[2].resize(3);
  p[3].resize(5);

  RT t23 = c[3] * A;
  RT t2 = fx[2] * fx[2];
  RT t22 = t2 * c[1];
  RT t21 = c[0] * fx[2];
  RT t20 = c[2] * fx[1];
  RT t19 = fx[1] * fx[2];
  RT t18 = fx[2] * fx[0];
  RT t17 = t2 * c[0];
  RT t16 = c[5] * fx[0];
  RT t15 = c[5] * fx[1] * fx[1];
  RT t14 = c[3] * t19;
  RT t13 = RT(-2) * t17 - t15 + (RT(2) * t16 + t20) * fx[2];
  RT t12 = B * B;

  p[1][2] = c[4] * t12;
  p[1][1] = B * t23 + c[1] * t12;
  p[1][0] = t12 * c[0] + (B * c[2] + c[5] * A) * A;

  p[2][2] = fx[2] * c[4] * B;
  p[2][1] = (c[1] * B + t23) * fx[2];
  p[2][0] = (t21 - t16) * B + (-c[5] * fx[1] + c[2] * fx[2]) * A;

  p[3][4] = -t2 * c[4] * c[4];
  p[3][3] = (t14 - RT(2) * t22) * c[4];
  p[3][2] = -c[3] * c[3] * t18 + (t14 - t22) * c[1] + t13 * c[4];
  p[3][1] = (RT(-2) * c[2] * t18 + (t16 + t21) * fx[1]) * c[3] + t13 * c[1];
  p[3][0] = (c[2] * t19 - t15 - t17) * c[0]
    + (-c[2] * c[2] * fx[2] + (RT(2) * t21 + t20 - t16) * c[5]) * fx[0];

  return p;
}




template < typename MPoly,
           typename RT,
	   typename POLY,
	   typename RO_t >
int StHa_x2_cx2y2(
		  const MPoly& c,
		  const POLY& fx, const int& sgn, const RT& A, const RT& B,
		  const RO_t& y)
{
  // The conic is of degree 2 in X
  // fx is of degree 2 in X
  //	RT c[6]  = {-2, 0, 0, 0, 1, 1};
  //  int sgn2, sgn3, sgn4;
  std::vector<POLY> p = compute_StHa_x2_cx2y2(c, fx, A, B);

  Seq<int> V;
  V.push_back( sgn );
  if (UPOLDAR::degree(p[1]) != -1)  V.push_back( sign_at(p[1], y) );
  if (UPOLDAR::degree(p[2]) != -1)  V.push_back( sign_at(p[2], y) );
  if (UPOLDAR::degree(p[3]) != -1)  V.push_back( sign_at(p[3], y) );
#if 0
  cout << "=====================================" << endl;
  cout << "c = " << c  << endl;
  cout << "fx = " << fx << endl;
  cout << "isol = " << A << "/" << B << endl;
  cout << "P2 = " << p[1] << endl;
  cout << "P3 = " << p[2] << endl;
  cout << "P4 = " << p[3] << endl;

  cout <<  sgn << endl;
  cout <<  sign_at(p[1], y)  << endl;
  cout <<  sign_at(p[2], y)  << endl;
  cout <<  sign_at(p[3], y)  << endl;
  cout << "Var = " << Wvar(V) << endl;
  //  cout << sgn << " " << sgn2 << " " << sgn3 << " " << sgn4 <<  endl;
  cout << "=====================================" << endl;
#endif
  return Wvar(V);
}




//----------------------------------------------------------------------------
template< typename MPoly,
          typename RT,
	  typename POLY >
std::vector<POLY>
compute_StHa_x3_cx2y2(
		      const MPoly& c,
		      const POLY& fx,
		      const RT& A, const RT& B)
{
  // The conic is of degree 2 in X
  // fx is of degree 3 in X
  //	RT c[6]  = {-2, 0, 0, 0, 1, 1};
  //	int sgn1, sgn2, sgn3, sgn4;
  std::vector<POLY> p(4);
  p[0] = fx;
  p[1].resize(3);
  p[2].resize(4);
  p[3].resize(7);

  RT t13 = fx[3];
  RT t3 = t13 * t13;
  RT t58 = RT(-3) * t3;
  RT t12 = fx[2];
  RT t2 = t12 * t12;
  RT t57 = RT(-2) * t2;
  RT t16 = c[5];
  RT t56 = RT(4) * t16;
  RT t55 = RT(2) * t12;
  RT t15 = c[4];
  RT t54 = t15 * B;
  RT t19 = c[2];
  RT t53 = t19 * B;
  RT t52 = t16 * A;
  RT t14 = c[3];
  RT t51 = t14 * A;
  RT t7 = t19 * t19;
  RT t50 = t7 * t13;
  RT t8 = t16 * t16;
  RT t11 = fx[1];
  RT t49 = t8 * t11;
  RT t9 = t14 * t14;
  RT t48 = t9 * t11;
  RT t10 = fx[0];
  RT t47 = t10 * t9;
  RT t18 = c[1];
  RT t46 = t3 * t18;
  RT t17 = c[0];
  RT t45 = t17 * t3;
  RT t44 = t11 * t13;
  RT t43 = t19 * t12;
  RT t42 = t12 * t13;
  RT t41 = t19 * t11;
  RT t40 = t10 * t16;
  RT t39 = t18 * t14;
  RT t38 = t13 * t17;
  RT t37 = t13 * t19;
  RT t36 = t16 * t12;
  RT t35 = RT(-3) * t45;
  RT t34 = t17 * t55;
  RT t33 = RT(-3) * t38;
  RT t32 = RT(2) * t43;
  RT t31 = t11 * t39;
  RT t30 = -t53 + t52;
  RT t29 = t34 - RT(3) * t40;
  RT t4 = t17 * t17;
  RT t1 = t11 * t11;
  RT t28 = (t10 * t55 - t1) * t8 + t4 * t58;
  RT t27 = t12 * t37 + (RT(2) * t44 - t2) * t16;
  RT t20 = B * B;
  RT t6 = t15 * t15;
  RT t5 = t18 * t18;

  p[1][2] = t15 * t20;
  p[1][1] = B * t51 + t20 * t18;
  p[1][0] = t17 * t20 + (t53 + t52) * A;

  p[2][3] = -t13 * t14 * t54;
  p[2][2] = t36 * t54 + (-B * t39 - t9 * A + t30 * t15) * t13;
  p[2][1] = (t18 * B + t51) * t36 + (t30 * t18 + (-2 * A * t19 - t17 * B) * t14) * t13;
  p[2][0] = (-t8 * t10 + (-t37 + t36) * t17) * B + (-t49 - t50 + (t38 + t43) * t16) * A;

  p[3][6] = -t15 * t6 * t3;
  p[3][5] = (-3 * t46 + t14 * t42) * t6;
  p[3][4] = (t5 * t58 + (t39 * t55 - t48) * t13) * t15 + (t35 + t27) * t6;
  p[3][3] = -t5 * t46 + (-t18 * t48 + (t47 + t5 * t12) * t14) * t13
    + (t11 * t14 * t36 + (RT(-6) * t45 + t16 * t57) * t18
       + ((t32 + t11 * t56) * t18 + (RT(-2) * t41 + t29) * t14) * t13) * t15;
  p[3][2] = t5 * t35 + (-t5 * t2 + (t31 - t47) * t12) * t16
    + ((t12 * t41 + t17 * t57) * t16 + t28) * t15
    + (t29 * t39 + (-t17 * t9 + RT(2) * t5 * t16 + (t17 * t56 - t7) * t15) * t11
       + (RT(-2) * t31 + (RT(2) * t17 * t15 + t5) * t12
          + RT(3) * (-t16 * t15 + t9) * t10) * t19) * t13;
  p[3][1] = (t4 * t42 + (t36 - RT(2) * t37) * t11 * t17
	     + (RT(3) * t50 + t49 + (RT(-2) * t43 + t33) * t16) * t10) * t14
    + ((t17 * t32 - t7 * t11) * t13 + ((t57 + RT(4) * t44) * t17
				       + (RT(-3) * t13 * t10 + t12 * t11) * t19) * t16 + t28) * t18;
  p[3][0] = (-t1 * t8 + (-t50 + t19 * t36) * t11) * t17
    + (t7 * t37 + (t19 * t33 - t7 * t12) * t16
       + (-t40 + t41 + t34) * t8) * t10 + (-t45 + t27) * t4;

  return p;
}


template < typename MPoly,
           typename RT,
	   typename POLY,
	   typename RO_t >
int StHa_x3_cx2y2(
		  const MPoly& c,
		  const POLY& fx, const int& sgn, const RT& A, const RT& B,
		  const RO_t& y)
{
  // The conic is of degree 2 in X
  // fx is of degree 3 in X
  //	RT c[6]  = {-2, 0, 0, 0, 1, 1};
  //	int sgn1, sgn2, sgn3, sgn4;
  std::vector<POLY> p = compute_StHa_x3_cx2y2(c, fx, A, B);

  Seq<int> V;
  V.push_back( sgn );
  if (UPOLDAR::degree(p[1]) != -1)  V.push_back( sign_at(p[1], y) );
  if (UPOLDAR::degree(p[2]) != -1)  V.push_back( sign_at(p[2], y) );
  if (UPOLDAR::degree(p[3]) != -1)  V.push_back( sign_at(p[3], y) );

  //  cout << "P2 = " << p2 << endl;
  //  cout << "P3 = " << p3 << endl;
  //  cout << "P4 = " << p4 << endl;
  //  cout << sgn1 << " " << sgn2 << " " << sgn3 << " " << sgn4 <<  endl;
  return Wvar(V);
}





//----------------------------------------------------------------------------
template < typename MPoly,
           typename RT,
	   typename POLY >
std::vector<POLY>
compute_StHa_x4_cx2y2(
		      const MPoly& c,
		      const POLY& fx,
		      const RT& A, const RT& B )
{

  // The conic is of degree 2 in X
  // fx is of degree 4 in X
  //  RT c[6]  = {-10, 0, 0, 0, 1, 1};

  std::vector<POLY> p(5);
  p[0] = fx;
  p[1].resize(3);
  p[2].resize(3);
  p[3].resize(5);
  p[4].resize(9);

  RT t24 = c[0];
  RT t10 = t24 * t24;
  RT t23 = fx[4];
  RT t5 = t23 * t23;
  RT t124 = t10 * t5;
  RT t29 = c[5];
  RT t15 = t29 * t29;
  RT t19 = fx[0];
  RT t116 = t15 * t19;
  RT t25 = c[1];
  RT t13 = t25 * t25;
  RT t131 = t13 * t5;
  RT t27 = c[3];
  RT t17 = t27 * t27;
  RT t115 = t17 * t19;
  RT t26 = c[2];
  RT t14 = t26 * t26;
  RT t130 = t14 * t23;
  RT t21 = fx[2];
  RT t4 = t21 * t21;
  RT t129 = RT(2) * t4;
  RT t128 = RT(4) * t5;
  RT t127 = RT(3) * t13;
  RT t126 = RT(-2) * t21;
  RT t22 = fx[3];
  RT t125 = RT(-3) * t22;
  RT t123 = t5 * t24;
  RT t6 = t22 * t22;
  RT t122 = t24 * t6;
  RT t121 = t29 * t6;
  RT t20 = fx[1];
  RT t120 = t26 * t20;
  RT t119 = t24 * t21;
  RT t28 = c[4];
  RT t118 = t28 * t14;
  RT t117 = t17 * t24;
  RT t114 = t27 * t25;
  RT t113 = t26 * t19;
  RT t112 = t24 * t20;
  RT t111 = t29 * t21;
  RT t110 = t29 * t20;
  RT t109 = t10 * t23;
  RT t108 = t15 * t21;
  RT t107 = t26 * t21;
  RT t106 = t14 * t20;
  RT t105 = t17 * t20;
  RT t16 = t27 * t17;
  RT t104 = t16 * t19;
  RT t103 = t17 * t21;
  RT t12 = t15 * t29;
  RT t102 = t12 * t19;
  RT t101 = t29 * t22;
  RT t100 = t28 * t29;
  RT t99 = t20 * t21;
  RT t98 = t14 * t19;
  RT t97 = t10 * t21;
  RT t96 = t20 * t27;
  RT t95 = t14 * t21;
  RT t94 = t13 * t21;
  RT t30 = B * B;
  RT t93 = t28 * t30;
  RT t92 = t26 * t22;
  RT t91 = t22 * t20;
  RT t90 = t28 * t24;
  RT t89 = t29 * t24;
  RT t88 = t16 * t20;
  RT t87 = t29 * t26;
  RT t86 = t24 * t22;
  RT t85 = t26 * t27;
  RT t84 = t22 * t27;
  RT t83 = RT(3) * t122;
  RT t82 = RT(4) * t123;
  RT t81 = RT(3) * t10 * t6;
  RT t80 = t24 * t129;
  RT t79 = RT(6) * t124;
  RT t78 = RT(3) * t116;
  RT t77 = t13 * t125;
  RT t76 = RT(2) * t89;
  RT t75 = RT(-4) * t115;
  RT t74 = RT(3) * t113;
  RT t73 = RT(3) * t87;
  RT t72 = RT(-8) * t113;
  RT t71 = RT(-3) * t114;
  RT t70 = RT(-3) * t106;
  RT t69 = t10 * t125;
  RT t68 = RT(2) * t29;
  RT t67 = t14 * t101;
  RT t66 = t21 * t114;
  RT t65 = t21 * t92;
  RT t64 = t26 * t112;
  RT t63 = t25 * t85;
  RT t62 = t26 * t86;
  RT t61 = t22 * t87;
  RT t60 = t24 * t107;
  RT t59 = t25 * t96;
  RT t58 = t29 * t84;
  RT t57 = t25 * t105;
  RT t56 = t27 * t94;
  RT t55 = t27 * t100;
  RT t54 = RT(-2) * t66;
  RT t53 = -t14 + t76;
  RT t52 = (RT(-2) * t91 + t4) * t15;
  RT t51 = t28 * t26 + t114;
  RT t50 = RT(4) * t116 + RT(2) * t95;
  RT t49 = RT(3) * t110 + RT(2) * t107;
  RT t48 = RT(-6) * t111 - RT(3) * t92;
  RT t47 = t61 - t108;
  RT t46 = t22 * t71 + t103;
  RT t3 = t20 * t20;
  RT t7 = t24 * t10;
  RT t45 = t7 * t128 + (t3 + t19 * t126) * t12;
  RT t44 = -t88 + RT(2) * t25 * t103 + t27 * t77;
  RT t43 = t121 + (RT(-2) * t111 - t92) * t23;
  RT t18 = t28 * t28;
  RT t11 = t14 * t26;
  RT t9 = t18 * t28;
  RT t8 = t25 * t13;
  RT t2 = t27 * A * B + t25 * t30;
  RT t1 = -t24 * t30 + (-t26 * B - t29 * A) * A;

  p[1][2] = t93;
  p[1][1] = t2;
  p[1][0] = -t1;

  p[2][2] = -t93 *c[5];
  p[2][1] = -t2 * c[5];
  p[2][0] = t1 * c[5];

  p[3][4] = (t18 * t29 - t28 * t17) * t23 * B;
  p[3][3] = t22 * B * t55 + ((2 * t55 - t16) * A
			     + (-2 * t28 * t85 + (-t17 + 2 * t100) * t25) * B) * t23;
  p[3][2] = ((-t28 * t15 + t17 * t29) * t22 + (-3 * t17 * t26 + t51 * t68) * t23) * A
    + (t25 * t58 + t47 * t28 + (-2 * t63 - t117 + t13 * t29 + t53 * t28) * t23) * B;
  p[3][1] = ((t101 - 2 * t23 * t26) * t27 * t24 + (t53 * t23 + t47) * t25) * B
    + ((-t15 * t22 + 2 * t23 * t87) * t25
       + (-t108 + 2 * t61 + (-3 * t14 + t76) * t23) * t27) * A;
  p[3][0] = (t102 + t29 * t109 + (-t130 + t47) * t24) * B
    + (t12 * t20 + t67 + (-t11 + t26 * t76) * t23 + (-t86 - t107) * t15) * A;

  p[4][8] = t18 * t18 * t5;
  p[4][7] = (t25 * t128 - t23 * t84) * t9;
  p[4][6] = (RT(6) * t131 + t46 * t23) * t18 + (t82 + t43) * t9;
  p[4][5] = t8 * t28 * t128 + (-t21 * t58 + (RT(3) * t121 + RT(12) * t123) * t25) * t18
    + (t44 * t28 + (t48 * t25 + (RT(-3) * t86 + t49) * t27) * t18) * t23;
  p[4][4] = (RT(12) * t90 + t13) * t131 + (t79 + t52) * t18 + ((t83 - t65) * t18
							       + (t6 * t127 + (t105 + t54) * t22) * t28) * t29
    + ((RT(2) * t116 + t95 - RT(3) * t62 + (RT(3) * t120 - RT(6) * t119) * t29) * t18
       - t8 * t84 - t25 * t88
       + (t48 * t13 + (RT(-6) * t86 + RT(6) * t110 + RT(4) * t107) * t114) * t28
       + (t94 + (RT(-4) * t29 * t19 - RT(3) * t120 + RT(2) * t119) * t28 + t115) * t17) * t23;
  p[4][3] = t8 * t82 + (t8 * t6 + (-t56 - t104 + t57) * t22) * t29
    + ((-t15 * t99 + (t78 + (t120 - t119) * t68) * t22) * t27
       + (RT(12) * t124 + (RT(6) * t122 - RT(2) * t65) * t29
	  + (t129 - RT(4) * t91) * t15) * t25) * t28
    + (t44 * t24 + (t96 * t127 + t8 * t126 + t25 * t75) * t29
       + (RT(4) * t104 - RT(3) * t57 - t8 * t22 + RT(2) * t56) * t26
       + ((RT(4) * t60 + t70 + t69 + (RT(6) * t112 + t72) * t29) * t27
	  + (RT(-6) * t62 + (RT(-12) * t119 + RT(6) * t120) * t29 + t50) * t25) * t28) * t23;
  p[4][2] = (t79 + t29 * t83) * t13 + (t13 * t4 + (t115 - t59) * t21) * t15
    + (t29 * t81 + (-t26 * t99 + t80) * t15 + t45) * t28
    + ((RT(3) * t19 * t114 - RT(2) * t13 * t20 + (RT(-4) * t112 + t74) * t28) * t15
       + (t24 * t54 + (t117 + t118) * t20 + (RT(2) * t59 - RT(3) * t115
					     + (-t13 - RT(2) * t90) * t21) * t26) * t29) * t22
    + (t14 * t94 + (t14 * t71 - t28 * t11 + t13 * t73) * t20
       + (RT(2) * t15 * t13 + RT(6) * t17 * t14 + (RT(-4) * t118 - RT(8) * t63) * t29) * t19
       + (t48 * t28 + t46) * t10 + (t50 * t28 + (RT(4) * t66 - RT(3) * t105 + t77) * t26
				    + (RT(-6) * t94 + t75 + RT(6) * t51 * t20) * t29) * t24) * t23;
  p[4][1] = (-t20 * t102 + (RT(2) * t113 - t112) * t108
	     + (t24 * t78 + (RT(-3) * t98 - t97 + RT(2) * t64) * t29) * t22
	     + (RT(4) * t11 * t19 - t7 * t22 + (t29 * t72 + t70) * t24 + t49 * t10) * t23) * t27
    + ((t81 + (RT(-2) * t60 + t106) * t22) * t29 + (t22 * t74 + t80
						    + (RT(-4) * t86 - t107) * t20) * t15 + (-t11 * t20 + t26 * t69 + t24 * t50
											    + (RT(6) * t64 - RT(6) * t97 - RT(4) * t98) * t29) * t23 + t45) * t25;
  p[4][0] = (-t11 * t101 + (RT(-4) * t89 + t14) * t130 + (-t120 - RT(2) * t119) * t12
	     + (t95 + RT(3) * t62 + RT(2) * t109 + t116) * t15) * t19 + t43 * t7
    + (t12 * t3 + (-t11 * t23 + t67 - t15 * t107) * t20) * t24
    + (-t21 * t61 + (t20 * t73 + t95) * t23 + t52 + t124) * t10;

  return p;
}




template < typename MPoly,
           typename RT,
	   typename POLY,
	   typename RO_t >
int StHa_x4_cx2y2(
		  const MPoly& c,
		  const POLY& fx, const int& sgn, const RT& A, const RT& B,
		  const RO_t& y)
{

  // The conic is of degree 2 in X
  // fx is of degree 4 in X
  //  RT c[6]  = {-10, 0, 0, 0, 1, 1};
  std::vector<POLY> p = compute_StHa_x4_cx2y2(c, fx, A, B);

  Seq<int> V;
  V.push_back( sgn );
  if (UPOLDAR::degree(p[1]) != -1)  V.push_back( sign_at(p[1], y) );
  if (UPOLDAR::degree(p[2]) != -1)  V.push_back( sign_at(p[2], y) );
  if (UPOLDAR::degree(p[3]) != -1)  V.push_back( sign_at(p[3], y) );
  if (UPOLDAR::degree(p[4]) != -1)  V.push_back( sign_at(p[4], y) );

  //  for (int i=0; i<= V.size(); ++i)
  //	cout << V[i] << " " << endl;
  //  cout << endl;
  //  int sgn1, sgn2, sgn3, sgn4, sgn5;
  //  sgn1 = sign_at(p[1], y);
  //  sgn2 = sign_at(p[2], y);
  //  sgn3 = sign_at(p[3], y);
  //  sgn4 = sign_at(p[4], y);
  //  sgn5 = sign_at(p[5], y);
  //  cout << sgn2 << " " << sgn3 << " " << sgn4 << " " << sgn5 << endl;

  return Wvar(V);
}



//-----------Help functions

template < typename RT >
UPolDse< UPolDse<RT> >
construct_X_minus_aY(const RT&  a = RT(1))
{
  typedef UPolDse<RT>               Poly_1;
  typedef UPolDse< UPolDse<RT> >    Poly_2;

  Poly_1 t0(2, AsSize()); t0[1] = RT(1);
  Poly_1 t1(1, AsSize()); t1[0] = a;

  Poly_2 h(2, AsSize());
  h[0] = t0;
  h[1] = t1;

  Variable< UPolDse< UPolDse<RT> > >() = "Y";
  return h;
}


template < typename RT >
UPolDse< UPolDse<RT> >
eval_upoly_at_upoly_2(const UPolDse<RT>& f, const UPolDse< UPolDse<RT> >& g)
{
  typedef UPolDse<RT>               Poly_1;
  typedef UPolDse< UPolDse<RT> >    Poly_2;

  Poly_1 tmp(1, AsSize());
  tmp[0] = lcoeff(f);

  Poly_2 H0(1, AsSize());
  H0[0] = tmp;

  Poly_2 H1;

  int p = UPOLDAR::degree(f);
  for (int i = 1; i <= p; ++i) {
    tmp[0] = f[p - i];
    H1 = g * H0 + tmp;

    std::swap(H1, H0);
  }

  return H0;
}


template < typename RT >
UPolDse< UPolDse<RT> >
compose(const UPolDse< UPolDse<RT> >& f, const UPolDse< UPolDse<RT> >& g)
{
  typedef UPolDse<RT>               Poly_1;
  typedef UPolDse< UPolDse<RT> >    Poly_2;

  Poly_1 one(1, AsSize()); one[0] = RT(1);
  Poly_2 result;

  int dgf = UPOLDAR::degree(f);
  for (int i = 0; i <= dgf; ++i) {
    if (UPOLDAR::degree(f[i]) >= 0) {
      Poly_2 t1(i+1, AsSize()); t1[i] = one;
      Poly_2 t2 = eval_upoly_at_upoly_2(f[i], g);
      result += t1 * t2;
    }
  }

  return result;
}


template < typename RT >
UPolDse<RT>
partial_eval( const UPolDse< UPolDse<RT> >& p, const RT& a)
{
  typedef UPolDse<RT>   UPoly;
  UPoly t(1, AsSize()); t[0] = a;
  UPoly h = UPOLDAR::eval_horner(p, t);
  return h;
}


template < typename RT >
RT homo_eval_horner(const UPolDse<RT>& f, const RT& b, const RT& c)
{
  RT d(1);

  int p = UPOLDAR::degree(f);
  RT H0 = UPOLDAR::lcoeff(f);
  RT H1;

  for (int i = 1; i <= p; ++i) {
    d = d * c;
    H1 = b * H0 + d * f[p-i];
    std::swap(H0, H1);
  }
  return H0;
}


template < typename RT >
UPolDse<RT>
special_eval_horner(const UPolDse< UPolDse<RT> >& f, const typename NumberTraits<RT>::FT& x)
{
  UPolDse<RT> b(1, AsSize()); b[0] = numerator(x);
  UPolDse<RT> c(1, AsSize()); c[0] = denominator(x);

  return homo_eval_horner(f, b, c);
}



template < typename RT >
UPolDse< UPolDse<RT> >
interchange_variables(const UPolDse< UPolDse<RT> >& f)
{
  typedef UPolDse<RT>              Poly;
  typedef UPolDse< UPolDse<RT> >   Poly_2;

  Poly_2 result;

  for (int i = 0; i <= f.get_degree(); ++i) {
    Poly ut(i+1, AsSize());
    ut[i] = RT(1);
    Poly_2 mt = let::convert_upoly_to_upoly_2(f[i]);
    result += mt * ut;
  }
  return result;
}

//-----------End Help functions



template < typename MPoly,
	   typename RO_t >
int
sign_at_line(const MPoly& f, const RO_t& x, const RO_t& y)
{
  assert(f.degree() <= 1);

  typedef typename RO_t::RT    RT;
  
  RO_t rx(x);
  RO_t ry(y);

  rx *= f[2];
  ry *= RT(-f[1]);
  ry -= f[0];

  return compare(rx, ry);
}



/**
 * @brief Computes the sign of @f@ evaluated over @x@ and @y@.
 *
 * @param f Bivariate polynomial of degree up to 2 (conic)
 * @param x algebraic number of degree up to 4
 * @param y algebraic number
 */
template < typename RT > inline
int
sign_at(const MPoly_2_2<RT>& f, const root_of<RT>& x, const root_of<RT>& y)
{
  typedef MPoly_2_2<RT>    MPoly;
  typedef root_of<RT>      RO_t;
  typedef typename root_of<RT>::POLY   POLY;
  

  typedef int (* signFunc)(const MPoly& f, const RO_t& x, const RO_t& y);

  const signFunc sgntab_[4] = {
    sign_at_x1_cx2y2<MPoly, RO_t>,
    sign_at_x2_cx2y2<MPoly, RO_t>,
    sign_at_x3_cx2y2<MPoly, RO_t>,
    sign_at_x4_cx2y2<MPoly, RO_t>
  };
  assert(x.degree() > 0);
  assert(x.degree() < 5);
  assert(y.degree() > 0);

  typedef typename RO_t::POLY POLY;
  typedef typename RO_t::FT    FT;

  if (f.degree() <= 1) return sign_at_line(f, x, y);

  if (x.degree() == 1) {
    POLY h = specialize_x<POLY>(f, x.left());
    return sign_at(h, y);
  }

  if (y.degree() == 1) {
    POLY h = specialize_y<POLY>(f, y.left());
    return sign_at(h, x);
  }
  

  MPoly g(f);  
  g.make_lcoeff_X_positive();
  int sgn;
  if (f == g) sgn = 1;
  else sgn = -1;

  return sgn * sgntab_[x.degree()-1](g, x, y);
}


/**
 * \brief Computes the sign of a bivariate polynomial evaluated over two algebraic numbers. Notice
 * that the polynomial is of type UPolDse< UPolDse <RT> >.
 *
 * \param f Bivariate polynomial.
 * \param x Real algebraic number.
 * \param y Real algebraic number.
 * \return the sign of the expression
 *
 * \sa root_of, UPolDse
 */
template < typename RT >
int
sign_at( const UPolDse< UPolDse<RT> >& f, const root_of<RT>& x, const root_of<RT>& y)
{
  typedef UPolDse<RT>               Poly;
  typedef UPolDse< UPolDse<RT> >    Poly_2;

  if (y.is_rational()) {
    Poly tf = special_eval_horner(f, y.left());
    return sign_at(tf, x);
  }

  if (x.is_rational()) {
    Poly_2 g = interchange_variables(f);
    Poly tf = special_eval_horner(g, x.left());
    return sign_at(tf, y);
  }

  Poly_2 g = let::convert_upoly_to_upoly_2(y.poly());

  SturmSeq<Poly> sth(g, f, HABICHT());
  Seq<Poly> sth_y_left;
  for (unsigned i = 0; i < sth.size(); ++i) {
    sth_y_left.push_back(special_eval_horner(sth[i], y.left()));
  }

  Seq<Poly> sth_y_right;
  for (unsigned i = 0; i < sth.size(); ++i) {
    sth_y_right.push_back(special_eval_horner(sth[i], y.right()));
  }

  std::vector<int> sgn_left;
  for (unsigned i = 0; i < sth_y_left.size(); ++i) {
    sgn_left.push_back( sign_at(sth_y_left[i], x) );
  }

  std::vector<int> sgn_right;
  for (unsigned i = 0; i < sth_y_right.size(); ++i) {
    sgn_right.push_back( sign_at(sth_y_right[i], x) );
  }

  return (y.sign_up() - y.sign_low()) * (Wvar(sgn_left) - Wvar(sgn_right));
}


/**
 * \brief Computes the sign of a bivariate polynomial, evaluated over two algebraic numbers.
 *
 * \param f Bivariate polynomial
 * \param x Real algebraic number
 * \param y Real algebraic number
 * \return The sign of the expression
 *
 * \sa root_of, MPol
 */
template < typename RT, typename O, typename R >
int
sign_at(const MPol<RT, O, R>& f, const root_of<RT>& x, const root_of<RT>& y)
{
  assert((NbVar(f) <= 2));

  typedef UPolDse<RT>               Poly;
  typedef UPolDse< UPolDse<RT> >    Poly_2;

  if ((MPOLDST::degree(f) <= 2) && (x.degree() <= 4)) {
    MPoly_2_2<RT> g = let::convert(f, type::As< MPoly_2_2<RT> >());
    return sign_at(g, x, y);
  }

  if (x.is_rational()) {
    Poly_2 g = let::convert_to_upoly_2(f, POLY_wrt_Y);
    Poly tf = special_eval_horner(g, x.left());
    return sign_at(tf, y);
  }

  if (y.is_rational()) {
    Poly_2 g = let::convert_to_upoly_2(f, POLY_wrt_X);
    Poly tf = special_eval_horner(g, y.left());
    return sign_at(tf, x);
  }

  Poly_2 g = let::convert_to_upoly_2(f, POLY_wrt_Y);
  return sign_at(g, x, y);
}



//Additional functions for solve
//----------------------------------------------------------------------------
template < typename MPoly,
           typename RT,
	   typename POLY >
std::vector<POLY>
compute_StHa_x_cx2y2(
		     const MPoly& c,
		     const POLY& fx,
		     const RT& A, const RT& B )
{
  typedef std::vector<POLY>
    (* StHaFunc)(const MPoly& c, const POLY& fx, const RT& A, const RT& B);

  const StHaFunc functab_[4] = {
    compute_StHa_x1_cx2y2<MPoly, RT, POLY>,
    compute_StHa_x2_cx2y2<MPoly, RT, POLY>,
    compute_StHa_x3_cx2y2<MPoly, RT, POLY>,
    compute_StHa_x4_cx2y2<MPoly, RT, POLY>
  };

  //  assert(fx.degree() > 1);
  return functab_[fx.degree()-1](c, fx, A, B);
}




template < typename POLY,
           typename RO_t >
int sign_on_StHa(
		 const std::vector<POLY>& StHa,
		 const int& sgn_a,
		 const RO_t& y)
{
  print(__FUNCTION__);

  std::vector<int> V;
  V.push_back( sgn_a );

  for (unsigned int i = 1; i < StHa.size(); ++i) {
    if (UPOLDAR::degree(StHa[i]) != -1)  V.push_back( sign_at(StHa[i], y) );
  }

  return Wvar(V);
}


__ALGEBRAIC_END_NAMESPACE







#endif // SYNAPS_USOLVE_ALGEBRAIC_SIGN_SIGN_FCT_2_H

