/********************************************************************
 *   This file is part of the source code of the SYNAPS library
 *   Author(s): B. Mourrain, GALAAD, INRIA
 *   History:
 *     The Rockwood method for computing the first positive real root added.
 *   $Id: Std.h,v 1.1 2005/09/28 06:30:41 mourrain Exp $
 *********************************************************************/
/** \file                                                           **/
//--------------------------------------------------------------------
#ifndef synaps_solve_bezier_Std_H
#define synaps_solve_bezier_Std_H
//--------------------------------------------------------------------
#include <synaps/init.h>
#include <synaps/arithm/binomial.h>
#include <synaps/base/Seq.h>
#include "synaps/arithm/sign.h"
#include <synaps/upol/bezier/rep1d.h>
#include <synaps/upol/bezier/BEZIER.m>
#include <synaps/usolve/bezier/Bezier.h>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
struct Opt {};
struct Rkw {};
struct Fst {};
//====================================================================
namespace bezier 
{
  using namespace BEZIER;
  //------------------------------------------------------------------
  /** Apply the subdivision algorithm for the polynomial on $]a,b[$, for the
      polynomial which coefficients in the Bernstein basis are @v1@.
      The algorithm subdivides the interval $]a,b[$ if its size is larger
      than @e@ and the sign variations $\ge 2$. The computed values are
      such that the real roots of the polynomial @v1@, within the
      distance $e$ from these values. They are added at the end of the
      sequence @sol@.  The routine ouputs the number of real roots,
      within the precision @e@.
  **/
  template<class S,class T, class C>
  unsigned isole_eps(S& sol,  T& v1, const C& a, const C& b, const C& e)
  {
    //    cout<<"-> "<<a<<" "<<b<<" "; VECTOR::print(cout,v1); cout<<endl;
    unsigned nbr=0;
    int d=(v1.size())-1;
    unsigned var=sign_variation(v1,2);
    if(b-a<e && b-a>-e & var>0)
      {
	sol.push_back((a+b)/2);
	return 1;
      }
    else if (var==1)
      {
	if(v1[d]==0)
	  sol.push_back(b);
	C s =dichotom(v1,a,b,e);
	// cout <<"S v1 "<<s<<endl;
	sol.push_back(s); return 1;
      }
    else if (var>1)
      {
	T v0(d+1);
	BEZIER::subdivide(v0,v1);
	nbr+= bezier::isole_eps(sol,v0,a,C((a+b)/2),e);
	if(v1[0]< e && v1[0]>-e) {sol.push_back(C((a+b)/2)); nbr++; }
	nbr+= bezier::isole_eps(sol,v1,C((a+b)/2),b,e);
      }
    return nbr;
  }

  //--------------------------------------------------------------------
  template<class T>
  T bezierbsearch(T* src, unsigned size, const T& precision )
  {
    typedef T real;
    real d,g;
    g = 0.0; d = 1.0;
    if ( src[0] > src[size-1])
      {
	BEZIER::bezier2monomials<T>(src,size);
	do
	  {
	    real m = (g+d)/2.0;
	    unsigned i = size-1;
	    real v = 0;
	    do { v += src[i]; v *= m; i--; } while(i);
	    v += src[0];
	    if ( v >= 0 ) g = m; else d = m;
	  }
	while ( d-g > precision );
      }
    else
      {
	BEZIER::bezier2monomials<T>(src,size);
	do
	  {
	    real m = (g+d)/2.0;
	    real v = 0;
	    unsigned i = size-1;
	    do { v += src[i]; v *= m; i--; } while(i);
	    v += src[0];
	    if ( v <= 0 ) g = m; else  d = m;
	  }
	while ( d-g > precision );
      };
    return (d+g)/2.0;
  };

  //--------------------------------------------------------------------
  template<class T>
  unsigned bezier_roots_alloc(T ** vals, T ** parms, unsigned size,
			       const T& precision )
  {
    unsigned maxdeep=0,maxsize,allocated;
    T ex=1;
    while(ex>precision) {ex/=2; maxdeep++;}
    maxdeep *= 2; maxdeep++;
    //(unsigned)(2*log(T(1.0/precision))/log(T(2)))+1;
    maxsize = (2+size)*maxdeep;
    while( !(*vals=new T[maxsize]) ) maxsize /= 2;
    allocated = maxsize/(2+size);
    *parms = *vals + allocated*size;
    return allocated;
  };

  template<typename T,typename C> inline
  void push_back(T &  b, const C& c, unsigned s) {b[s]=c;}

  template<typename C> inline
  void push_back(std::vector<C> &  b, const C& c, unsigned s){b.push_back(c);}

  template<typename C> inline
  void push_back(std::vector<std::pair<C,C> > &  b, const C& c, unsigned s)
           {b.push_back(std::pair<C,C>(c,c));}

  template<typename C> inline
  void push_back(std::vector<std::pair<C,C> > &  b, const C& c, const C& d,
		 unsigned s)
           {b.push_back(std::pair<C,C>(c,d));}

  template<class T>
  void bezier_roots_free( T * vals, T * parms ) { delete[] vals; };

  //----------------------------------------------------------------------
  template<class Dst,class T,class B>
  unsigned bezier_roots_loop( Dst& roots,
			      T * vals_mem,
			      T * parms_mem,
			      unsigned allocated,
			      const B& src,
			      const T& precision )
  {
    typedef T real;
    unsigned size = src.size();
    unsigned nroots = 0;
    real * curr;
    real * parms;

    parms = parms_mem;
    curr  = vals_mem;
    std::copy(src.begin(),src.end(),curr);
    vals_mem -= size;
    parms[0] = 0.0;
    parms[1] = 1.0;
    do
    {
      unsigned c = sign_variation( curr, curr+size );
      if ( c != 0 )
	{
	  if ( parms[1] - parms[0]  < precision )
	    {
	      push_back(roots,T((parms[1]+parms[0])/2),nroots);
	      //push_back(roots,parms[0],parms[1],nroots);
	      nroots++;
	      curr  -= size;
	      parms -= 2;
	    }
	  else
	    {
	      if ( c > 1 )
		{
		  if ( parms == parms_mem - allocated ) return -1;

		  real * left  = curr+size;
		  real * right = curr;

		  // split at 1/2
		  subdivide_size(left,right,size);

		  // push new parameters on the stack
		  parms[3] =  (parms[1]+parms[0])/2.0;
		  parms[2] =  parms[0];
		  parms[0] =  parms[3];
		  parms += 2;
		  curr  += size;
		}
	      else
		{
		  // binary search
		  real delta,m;
		  delta = precision/(parms[1]-parms[0]);
		  m = bezierbsearch(curr,size,delta);
		  push_back(roots,T((1-m)*parms[0]+m*parms[1]),nroots);
		  //  push_back(roots,parms[0],parms[1],nroots);
		  nroots++;
		  curr -= size;
		  parms -=2;
		};
	    };
	}
      else
	{
	  // return to previous call
	  curr-=size;
	  parms-=2;
	};
    }
  while( curr != vals_mem );
  return nroots;
  };

  //--------------------------------------------------------------------
  // solveur
  template<class Dst,class BezierRep>
  inline
  int isole_eps_fast(Dst& roots, const BezierRep& Br,
	    const typename BezierRep::value_type& prec)
  {
    typedef typename BezierRep::value_type real;
    real *vals_mem,*parms_mem;
    unsigned allocated;
    int result;
    allocated = bezier_roots_alloc( &vals_mem, &parms_mem, Br.size(), prec );
    result = bezier_roots_loop( roots,
				vals_mem, parms_mem,
				allocated,
				Br, prec );
    bezier_roots_free(vals_mem,parms_mem);
    return result;
  };

  //----------------------------------------------------------------------
  template<class Dst,class T,class B>
  unsigned bezier_roots_isole_loop( Dst& roots,
			      T * vals_mem,
			      T * parms_mem,
			      unsigned allocated,
			      const B& src,
			      const T& precision )
  {
    typedef T real;
    unsigned size = src.size();
    unsigned nroots = 0;
    real * curr;
    real * parms;

    parms = parms_mem;
    curr  = vals_mem;
    std::copy(src.begin(),src.end(),curr);
    vals_mem -= size;
    parms[0] = 0.0;
    parms[1] = 1.0;
    do
    {
      unsigned c = Variation( curr, size );
      if ( c != 0 )
	{
	  if ( parms[1] - parms[0]  < precision )
	    {
	      //push_back(roots,(parms[1]+parms[0])/2.0,nroots);
	      push_back(roots,parms[0],parms[1],nroots);
	      nroots++;
	      curr  -= size;
	      parms -= 2;
	    }
	  else
	    {
	      if ( c > 1 )
		{
		  if ( parms == parms_mem - allocated ) return -1;

		  real * left  = curr+size;
		  real * right = curr;

		  // split at 1/2
		  subdivide_size(left,right,size);

		  // push new parameters on the stack
		  parms[3] =  (parms[1]+parms[0])/2.0;
		  parms[2] =  parms[0];
		  parms[0] =  parms[3];
		  parms += 2;
		  curr  += size;
		}
	      else
		{
		  // binary search
		  //  real delta,m;
		  //  delta = precision/(parms[1]-parms[0]);
		  //m = bezierbsearch(curr,size,delta);
		  //push_back(roots,(1.0-m)*parms[0]+m*parms[1],nroots);
		  push_back(roots,parms[0],parms[1],nroots);
		  nroots++;
		  curr -= size;
		  parms -=2;
		};
	    };
	}
      else
	{
	  // return to previous call
	  curr-=size;
	  parms-=2;
	};
    }
  while( curr != vals_mem );
  return nroots;
  };

  // solveur
  template<class Dst,class BezierRep>
  inline
  int isole_fast(Dst& roots, const BezierRep& Br,
	    const typename BezierRep::value_type& prec)
  {
    typedef typename BezierRep::value_type real;
    real *vals_mem,*parms_mem;
    unsigned allocated;
    int result;
    allocated = bezier_roots_alloc( &vals_mem, &parms_mem, Br.size(), prec );
    result = bezier_roots_isole_loop( roots,
				vals_mem, parms_mem,
				allocated,
				Br, prec );
    bezier_roots_free(vals_mem,parms_mem);
    return result;
  };

  //--------------------------------------------------------------------
  /** Compute the first real positive root of a polynomial with coefficients
      @bz@ in the Bernstein basis for the interval $]u,v[$. It returns
      @v@ when no root is found.
  **/
  template<class C,class B>
  C solve_rockwood(const B & bz,const C & u, const C& v, const C & eps)
  {
    unsigned sz =bz.size();
    C* tab= new C[2*sz];
    C* t0=tab, *t1= tab+sz, *t2;
    for(unsigned i=0;i<sz;++i) t1[i]=bz[i];


    C x=u, a=u, b=v, c=u/*, h*/;
    unsigned i=0;
    bool s;

    while ( (t1[0]>eps || t1[0]<-eps) && (b-a >eps) )
    {
      //  for(unsigned k=0;k<sz;++k) std::cout<<t1[k]<<" "; std::cout<<std::endl;
      i=1;
      s = (t1[0]>0);
      while( s == (t1[i]>0) && i <sz) i++;
      if(i==sz) return v;
      x = ((i-1) - t1[i-1]/(t1[i]-t1[i-1]))/(sz-1);
      if(i==(sz-1) && t1[sz-1]<eps && t1[sz-1]>-eps) x=(C(sz)-C(2))/(sz-1);
         //return b;
      subdivide_size(t0,t1,sz,x);
      s = (t0[0]>0); i=1;
      while( s == (t0[i]>0)  && i <sz) ++i;
      if(i<sz)
	{ t2=t1;t1=t0;t0=t2; b-=(1-x)*(b-a); c=b;}
      else
	{ a+=x*(b-a); c=a;}
    }
    delete [] tab;
    return c;
  }
  //--------------------------------------------------------------------
  /** Assuming there is one root in $[a,b[$, it computes it within the
      precision e. **/
  template<class C,class T>
  C dichotom(const T & p, const C&  a, const C&  b, const C&  e)
  {
    C u0=0, u1=1, re=e/(b-a);
    int s0=sign(p[0]), s1=sign(p[p.degree()]);
    for(unsigned i=1; i < p.size() && s0==0; i++)  s0=sign(p[i]);
    while((u1-u0)>re)
      {
	int sm= sign(eval(p,C((u0+u1)/2)));
	if(sm != s0)
	  { s1=sm; u1=(u0+u1)/2; }
	else if(sm != s1)
	  { s0=sm; u0=(u0+u1)/2; }
	else if(sm!=0)
	  {
	    std::cerr <<"Error in sign evaluation"<< std::endl;
	    return 100000;//	  exit(1);
	  }
      }
    return (a+(u0+u1)*(b-a)/2);
  }
  //--------------------------------------------------------------------
  /** Compute the first root of a polynomial @v1@ in the interval $[a,b]$.
      Put the result in @x@ and return @true@, if a root is found.  Else if
      returns @false@. Step the subdivision when the size of the interval is
      small than @e@.
  **/
  template<class C, class T>
  bool firstroot(C & x, T& v1, const C& a, const C& b, const C& e)
  {
//    unsigned nbr=0;
    int d=(v1.size())-1;
    unsigned var= sign_variation(v1,2);
    if(b-a<e && b-a>-e && var>0)
      {
	x= (a+b)/C(2); return true;
      }
    else if(v1[0]< e && v1[0]>-e)
      {x= a; return true;}
    else if (var==1)
      {
	x = bezier::dichotom(v1,a,b,e); return true;
      }
    else if (var>1)
      {
	T v0(d+1);
	BEZIER::subdivide(v0,v1);
	C m=(a+b)/2;
	unsigned nv0 = sign_variation(v0,2);
	if(nv0 >0 && bezier::firstroot(x,v0,a,m,e))
	  return true;
	else
	  return bezier::firstroot(x,v1,m,b,e);
      }
    else
      return false;
  }

  //---------------------------------------------------------------------
  /**  Compute the $r^{th}$ first root on the interval $[a,b]$.
       If it founds it, it returns @true@ and put the isolating  interval
       in @x@. If it does not found it, it decrease @r@ by the number of
       roots found and return false.
       Roots are counted with multiplicity.
  **/
  template<class C, class T>
  bool isole_nth_first_root(std::pair<C,C> & x, T& v1,
			    const C& a, const C& b,
			    int & r, const C& e)
  {
//    unsigned nbr=0;
    int d=(v1.size())-1;
    unsigned var=Variation2(v1);
    if(v1[0]< e && v1[0]>-e)
      {
	unsigned mu=0;
	while((int)mu<d && v1[mu]< e && v1[mu]>-e) mu++;
	if(r<=(int)mu)
	  { x = std::pair<C,C>(a,b); r=0; return true; }
	else
	  { r-=mu;  return false; }
      }
    else if(var>0 && b-a<e && b-a>-e)
      {
	if ((int)var>=r)
	  {  x = std::pair<C,C>(a,b); r-=var; return true;}
	else
	  {r-=var; return false;}
      }
    else if (var==1)
      {
	if(r==1)
	  {
	    C u= shrinkinf(v1,a,b);
	    x = std::pair<C,C>(a,u);
	    r=0;
	    return true;
	  }
	else
	  { r--; return false; }
      }
    else if (var>1)
      {
	T v0(d+1);
	BEZIER::subdivide(v0,v1);
	C m=(a+b)/2;
	bool nth_root_found = bezier::isole_nth_first_root(x,v0,a,m,r,e);
	if(nth_root_found)
	  return true;
	else
	  return bezier::isole_nth_first_root(x,v1,m,b,r,e);
      }
    else
      return false;
  }

//---------------------------------------------------------------------
/**  Compute the $r^{th}$ first root on the interval $[a,b]$.
       If it founds it, it returns @true@ and put the isolating  interval
       in @x@. If it does not found it, it decrease @r@ by the number of
       roots found and return false.
       Roots are counted with multiplicity.
**/
  template<class C, class T>
  bool isole_nth_last_root(std::pair<C,C> & x, T& v1,
			    const C& a, const C& b,
			    int & r, const C& e)
  {
//    unsigned nbr=0;
    int d=(v1.size())-1;
    unsigned var=Variation2(v1);
    if(v1[d]< e && v1[d]>-e)
      {
	unsigned mu=d;
	while(mu>=0 && v1[mu]< e && v1[mu]>-e) mu++;
	if(r<=(int)mu)
	  { x = std::pair<C,C>(a,b); r=0; return true; }
	else
	  { r-=mu;  return false; }
      }
    else if(var>0 && b-a<e && b-a>-e)
      {
	if ((int)var>=r)
	  { x = std::pair<C,C>(a,b); r-=var; return true;}
	else
	  { r-=var; return false;}
      }
    else if (var==1)
      {
	if(r==1)
	  {
	    C u= shrinkinf(v1,a,b);
	    x = std::pair<C,C>(a,u);
	    r=0;
	    return true;
	  }
	else
	  { r--; return false; }
      }
    else if (var>1)
      {
	T v0(d+1);
	BEZIER::subdivide(v0,v1);
	C m=(a+b)/2;
	bool nth_root_found = bezier::isole_nth_last_root(x,v1,m,b,r,e);
	if(nth_root_found)
	  return true;
	else
	  return bezier::isole_nth_last_root(x,v0,a,m,r,e);
      }
    else
      return false;
  }
} // namespace  bezier
//----------------------------------------------------------------------
/** Compute the real roots of p, by subdivision of the control polygon,
    within a precision @eps@ after the change of coordinnates
    $x=\frac{t}{1-t}$.
**/
template<class C, class UPOL>
Seq<C> solve(const UPOL & p, const Bezier<C,Std> & mth)
{
  Seq<C> sol;

  unsigned vl=0; for(;vl<p.size() && p[vl]==0;vl++);
  unsigned d1=p.size()-vl;
  bezier::rep1d<C> bp(d1);
  let::assign(bp,p,vl);
  bezier::rep1d<C> bm(bp);

  using bezier::isole_eps;

  for (unsigned k=1;k<d1;k+=2) bm[k]*=(-1);
  unsigned n0 = isole_eps(sol.rep(),bm,C(0),C(1),mth.eps);
  for(unsigned j=0;j<n0; ++j) sol[j]/=(sol[j]-1);

  if(vl !=0) {sol.push_back(0); n0++;}

  unsigned n1 = isole_eps(sol.rep(),bp,C(0),C(1),mth.eps);
  for(unsigned j=0;j<n1; ++j) sol[j+n0]/=(1-sol[j+n0]);

  return sol;
}
//----------------------------------------------------------------------
/**
    Compute the real roots of p, by subdivision of the control polygon,
    within a precision \c eps after the change of coordinnates
    $x=\frac{t}{1-t}$.
**/
template<class C, class UPOL>
Seq<C> solve(const UPOL & p, const Bezier<C,Opt> & mth)
{
  Seq<C> sol;


  unsigned vl=0; for(;vl<p.size() && p[vl]==0;vl++);
  unsigned d1=p.size()-vl;
  bezier::rep1d<C> bp(d1);
  let::assign(bp,p,vl);
  bezier::rep1d<C> bm(bp);

  using namespace bezier;

  for (unsigned k=1;k<d1;k+=2) bm[k]*=(-1);
  unsigned n0 = isole_eps_fast(sol.rep(),bm,mth.eps);
  for(unsigned j=0;j<n0; ++j) sol[j]/=(sol[j]-1);

  if(vl !=0) {sol.push_back(0); n0++;}

  unsigned n1 = isole_eps_fast(sol.rep(),bp,mth.eps);
  for(unsigned j=0;j<n1; ++j) sol[j+n0]/=(1-sol[j+n0]);

  return sol;
}
//----------------------------------------------------------------------
/** Compute the first real positive roots of p, by subdivision of the control
    polygon, within a precision @eps@ after the change of coordinnates
    $x=\frac{t}{1-t}$.
**/
template<class C, class UPOL>
C RootOf(const UPOL & p, const Bezier<C,Rkw> & mth)
{
  unsigned d1=p.size();
  bezier::rep1d<C> bz(d1); let::assign(bz,p);
  using namespace bezier; C x = solve_rockwood(bz,C(0),C(1),mth.eps);
  if (x==1)
    {
      std::cerr<< "No real positive root"<<std::endl;
    }
  return x/(1-x);
}
//----------------------------------------------------------------------
/** Compute the first real positive roots of p, by subdivision of the control
    polygon, within a precision @eps@ after the change of coordinnates
    $x=\frac{t}{1-t}$.
**/
template<class C, class UPOL>
C RootOf(const UPOL & p, const Bezier<C,Fst> & mth)
{
  unsigned d1=p.size();
  bezier::rep1d<C> bz(d1); let::assign(bz,p);
  using namespace bezier; C x;
  bool b=firstroot(x,bz,C(0),C(1),mth.eps);
  if (!b)
    {
		std::cerr<< "No real positive root found"<<std::endl;
    }
  return x/(1-x);
}
//====================================================================
__END_NAMESPACE_SYNAPS
/********************************************************************/
#endif // synaps_solve_bezier_Std_H

