
#ifndef _synaps_usolve_contfrac_lower_bound_h_
#define _synaps_usolve_contfrac_lower_bound_h_

#include <queue>
#define ilog2(a) mpz_sizeinbase(a.get_mpz_t(),2)

/* As a power of 2 */

struct CauchyBound
{

  template < typename Poly >
  typename Poly::value_type lower( const Poly& f)
  {
    typedef typename Poly::value_type    RT;
    
    long deg = degree( f);
    long l = 0;
    
    for (int i = 1; i <= deg; ++i) {
      if ( sign( f[i]) * sign( f[0]) < 0 ) {
	++l;
      }
    }
    
    long ba0  = ilog2(f[0]) - 1; /* puiss de 2 < an */
    bool bound_set = false;
    long K = 0;
    
    for (int i = 1; i <= deg; ++i) {
      if ( sign( f[i]) * sign( f[0]) < 0 ) {
	long bai  = ilog2( RT( l * f[i])) - 1;
	long p = bai - ba0 - 1 ;
	long k = i;
	long q = p / k;
	long r = p % k;
	
	long rq = 0;
	if (r <= k-2) {
	  rq = q + 1;
	} else {
	  rq = q  + 2;
	}
	
      if ( (bound_set == false) || (K < rq) ) {
	K = rq;
	bound_set = true;
      }
      }
    }
    if ( K < 0 ) {return pow( RT(2), abs( K) );}

    return 0;
  }

};

//--------------------------------------------------------------------
/** A lower bound using the Bernstein representation of the polynomial.
    The idea is to compute the Bernstein form of the polynomial over
    the positive axis and find the least positive intersection point of the
    convex hull with the axis. From the convex hull property it follows
    that this quantity is a lower bound on the positive roots of the
    polynomial. The value returned is the floor of the intersection point.
**/
struct BezierBound
{
  template < typename Poly >
  typename Poly::value_type lower( const Poly& f)
  {
    //std::cout <<"Bezier"<<std::endl;
    typedef typename Poly::value_type    RT;
    unsigned deg = degree(f);
    bezier::rep1d<RT> bz(deg+1); 
    let::assign(bz, f);  // The Bernstein form of f w.r.t. the interval [0,\infinity]
    
    int sn = sign(bz[0]);
    if(sn == 0) return 0;// That is, zero is a root, this should not be the case
    
    //std::cout << "bz[0] = "<< bz[0] << std::endl;
    
     RT lowerBoundN =-1, lowerBoundD=1;// Numerator and denominator of lower bound
    //Now we find the least positive intersection of the convex hull.
    for(int i=1; i <= deg; i++){
      if(sn * sign(bz[i]) < 0){
	/*
	  std::cout << "bz["<<i<<"] = "<< bz[i] << std::endl;
	  std::cout << "bz[0]*i "<< bz[0] * i << std::endl;
	  std::cout << "(deg*(bz[0]-bz[i]))" << (deg*(bz[0]-bz[i])) << std::endl;
	*/
	if(lowerBoundN * deg *(bz[0]-bz[i]) > bz[0]*i*lowerBoundD 
	   || lowerBoundN == -1){
	  lowerBoundN = bz[0]*i;
	  lowerBoundD = deg*(bz[0]-bz[i]);
	}
      }
    }
    lowerBoundD -= lowerBoundN;// Have to scale back the result to the orignial setting of [0, B]
    long K = ilog2(lowerBoundN)- ilog2(lowerBoundD) -1;
    if(K < 0)
      return RT(0);
    else if(lowerBoundN == lowerBoundD)
      return 1;
    else{// K > 0
      return pow( RT(2), K);    
    }
  }
};
//--------------------------------------------------------------------
/**
   A lower bound based upon the upper bound given by Hong. This should be the
   best possible, since Hong's bound is better than Kioustelidis' bound.
   The problem is that it is a \Theta(n^2) procedure.
**/
struct HongBound
{
  template < typename Poly >
  typename Poly::value_type lower( const Poly& f)
  {
    typedef typename Poly::value_type    RT;
    unsigned deg = degree(f);
    
    long lB=0, gB=0;
    bool localBoundSet = false, globalBoundSet = false;
    long temp, q;
    int s = sign(f[0]);
    //std::cout << "f[0]= " << f[0] << std::endl;
    for(int i= deg; i > 0; i--){
      //std::cout << "f["<< i << "]" << " = " << f[i] << std::endl;
      if(sign(f[i]) * s < 0){
	for(int k=i-1; k >= 0; k--){
	  if(sign(f[k]) * s > 0){
	    temp = ilog2( RT(f[i]) ) - ilog2( RT(f[k]) ) - 1;
	    q = temp /(i-k);
	    if(!localBoundSet || lB > q + 2){// Choose the minimum among localBounds
	    localBoundSet = true;
	    lB = q+2;
	    }
	  }
	}
	localBoundSet = false;
	if(!globalBoundSet || gB < lB){// Choose the maximum among globalBounds
	  globalBoundSet = true;
	  gB = lB;
      }
      }
      //std::cout << "Cout gb after " << i << " loop = "<< gB << std::endl;
    }
    if ( gB+1 < 0 ) return pow( RT(2), abs( gB+1) );
    return 0;
  }
};

//--------------------------------------------------------------------
/**
   A \Theta(n) time algorithm for computing a lower bound. It is based
   upon an upper bound by Akritas et al. that returns a value some what
   better than Kioustelidis' bound, but is not better than the bound
   by Hong. The idea is to pair the positive and negative coefficients
   in a certain manner. Assume the constant coefficient is not zero and there
   is at least one sign variation in the coefficients.
**/
struct AkritasBound
{
  
  template < typename Poly >
  typename Poly::value_type lower( const Poly& f)
  {
    //    std::cout <<"Poly is " << f << std::endl;
    typedef typename Poly::value_type    RT;
    typedef std::pair<RT, int>	Coeff_deg;
    Coeff_deg T1, T2;

    unsigned deg=degree(f);
    int s=sign(f[0]), len;
    long B=0;// The logarithm of the bound to be returned
    bool boundSet=false;// The bound is not set initially
    long temp,q;
    int posCounter=0, negCounter=0, nPosCounter=0;
    // The coefficients from posCounter to negCounter-1 have the same sign
    // as f[0]; the coefficients from negCounter to nPosCounter-1 have the
    // opposite sign of f[0].
    std::queue< Coeff_deg > Neg;// Queue that stores the negative (between
                              // negCounter and nPosCounter) coefficients.
    std::stack< Coeff_deg > Pos; // Stack that stores the positive (between
                              // posCounter and negCounter) coefficients.

    int l1, l2;// size of the above two queues

    while(posCounter != deg+1){
      // Find the first change in sign larger than posCounter and assign negCounter
      // to this value. If no such sign variation occurs then
      // set negCounter to -1.
      negCounter = -1;// By default there is no such coefficient
      //Pos.push(Coeff_deg(RT(f[posCounter]), posCounter));// Push the first "positive" coefficient
      Pos.push(Coeff_deg(RT(f[posCounter]), posCounter));// Push the first "positive" coefficient
      for(int i=posCounter + 1;i <= deg; i++){
	if(sign(f[i]) * s < 0){
	  negCounter=i; break;
	}
	if(sign(f[i]) *s > 0){// Store the list of "positive" coefficients
	  //Pos.push(Coeff_deg(RT(f[i]), i));// Note the actual degree is deg-i, but
	                               // later we subtract degrees so we store only i
	  Pos.push(Coeff_deg(RT(f[i]), i));
	}
      }
      if(negCounter == -1){
	//	std::cout <<"Returning the lower bound " << B << std::endl;
	if ( B < 0 ) 
	  return pow( RT(2), abs(B) );
	else
	  return 0;
      }
      // Now find the next coefficient that has the same sign as f[0].
      nPosCounter = deg+1;// By default this is deg+1
      Neg.push(Coeff_deg(RT(f[negCounter]), negCounter));
      for(int i=negCounter + 1;i <= deg; i++){
	if(sign(f[i]) * s > 0){
	  nPosCounter=i; break;
	}
	if(sign(f[i]) *s < 0)// Store the list of "negative" coefficients
	  Neg.push(Coeff_deg(RT(f[i]), i));
      }

      //      std::cout <<" Assigning to the queues done" << std::endl;
      // Start computing the bound
      // If the number of "positive" terms from posCounter to negCounter -1 are
      // greater or equal to  the number of "negative" terms from negCounter to
      // nPosCounter -1 then we have the straightforward matching of "negative"
      // coefficients with the "positive" ones.
      l1 = Neg.size(); l2 = Pos.size();
      if(l2 >= l1){
	//	std::cout <<"Pos length " << l2 << " greater than Neg length " 
	//		  << l1 << std::endl;
	// or equally negCounter - posCounter >= nPosCounter - negCounter
	for(int i=0; i < l1; i++){
	  T1= Neg.front(); Neg.pop();
	  //T2= Pos.front(); Pos.pop();
	  T2=Pos.top();Pos.pop();
	  //	  std::cout <<"Neg coeff " << T1.first << " Pos coeff " << T2.first
	  //		    << " Deg difference " << T1.second - T2.second 
	  //		    << std::endl;
	  temp = ilog2( T1.first ) - ilog2( T2.first ) - 1;
	  q = temp/(T1.second - T2.second);
	  //	  std::cout << "Difference between logs " << temp 
	  //		    << " temporary bound value " << B << std::endl;
	  if(!boundSet || B < q + 2){// Choose the maximum
	    boundSet = true;
	    B = q+2;
	  }
	}
      }else{//Else split the coefficient f[posCounter] to compensate for
	// the paucity of "positive" terms and then do the matching.
	//	std::cout <<" Pos length "<< l2 << " smaller than Neg length " 
	//		  << l1 << std::endl;
	
	//T2=Pos.front(); Pos.pop();
	T2= Pos.top();Pos.pop();
	for(int i=0; i <= l1-l2; i++){
	  T1= Neg.front(); Neg.pop();
	  temp = ilog2( T1.first ) - ilog2( T2.first ) + ilog2(RT(l1-l2+1)) - 1;
	  q = temp/(T1.second-T2.second);
	  if(!boundSet || B < q + 2){// Choose the maximum
	    boundSet = true;
	    B = q+2;
	  }
	}
	//	std::cout <<" Splitting the leading coefficient done" << std::endl;
	l1 = Neg.size();// Now the size of Neg and Pos should be the same
	for(int i=0; i < l1; i++){
	  T1= Neg.front(); Neg.pop();
	  //T2= Pos.front(); Pos.pop();
	  Pos.pop();
	  temp = ilog2( T1.first ) - ilog2( T2.first ) - 1;
	  q = temp/(T1.second - T2.second);
	  if(!boundSet || B < q + 2){// Choose the maximum
	    boundSet = true;
	    B = q+2;
	  }
	}
      }// end else
      //      std::cout <<" Matching the coefficients done" << std::endl;
      posCounter = nPosCounter; Neg.empty(); Pos.empty();
    }//end while
    //    std::cout <<"Returning the lower bound " << B << std::endl;
    if ( B < 0 ) return pow( RT(2), abs(B) );
    return 0;
  }
};

  

#endif //_synaps_usolve_contfrac_lower_bound_h_
