// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)


// This code is a modification of the code from
// Polynomial root finder (polynoms up to degree 4)
// AUTHOR: Jochen SCHARZE
//           (See 'Cubic & Quartic Roots' in 'Graphics Gems 1', AP)
// CUBIC: Modification from the GNU gsl library
// QUARTIC: We use the gnu gsl-solver

// 27/11/2003: Modification mourrain@sophia.inria.fr for generic coefficients.
// 05/12/2003: Modification et@di.uoa.gr multiple roots and vectors as input


#ifndef SYNAPS_ARITHM_ALGEBRAIC_SOLVE_BY_RADICAL_H
#define SYNAPS_ARITHM_ALGEBRAIC_SOLVE_BY_RADICAL_H

#include <synaps/init.h>
#include <synaps/usolve/Traits.h>

__BEGIN_NAMESPACE_SYNAPS

#define EQN_EPS 1e-5

static int isZero(double x)
{
	return x > -EQN_EPS && x < EQN_EPS;
}


template<class V>
int solve_quadratic(std::vector<double>& s, const V& c)
{
	typedef typename V::value_type                RT;
	typedef typename Integer_type_traits<RT>::FT  FT;
	typedef Rational_traits<FT>                   Rat_traits;

	RT D = c[1] * c[1] - RT(4) *  c[2] * c[0];
	if (D == RT(0)) {
		// one double root
		FT r = Rat_traits::make_rational(RT(-c[1]),  RT(c[2] * 2));
		s.push_back(to_double(r));
		s.push_back(to_double(r));
		return 2;
	}

	if (D < RT(0)) return 0;

	// two real roots
	double sqrt_D = std::sqrt(to_double(D));


//	FT r1 = Rat_traits::make_rational( RT(-c[1] - sqrt_D),   RT(c[2] * 2) );
//	FT r2 = Rat_traits::make_rational( RT(-c[1] + sqrt_D),   RT(c[2] * 2) );

	FT r1 = Rat_traits::make_rational( RT(-c[1] ),   RT(c[2] * 2) );
	FT r2 = Rat_traits::make_rational( RT(-c[1] ),   RT(c[2] * 2) );

	r1 += FT(sqrt_D) /  RT(c[2] * 2);
	r1 -= FT(sqrt_D) /  RT(c[2] * 2);

	s.push_back(to_double(r1));
	s.push_back(to_double(r2));
	return 2;
}

	template<class NT, class VV>
int solve_cubic(VV& s, const VV& coeff)
{
	double a = coeff[2] / coeff[3];
	double b = coeff[1] / coeff[3];
	double c = coeff[0] / coeff[3];

	double q = (a * a - 3 * b);
	double r = (2 * a * a * a - 9 * a * b + 27 * c);

	double Q = q / 9;
	double R = r / 54;

	double Q3 = Q * Q * Q;
	double R2 = R * R;

	double CR2 = 729 * r * r;
	double CQ3 = 2916 * q * q * q;

	if (R == 0 && Q == 0) {
		s.push_back( - a / 3 );
		s.push_back( - a / 3 );
		s.push_back( - a / 3 );
		return 3 ;
	} else if (CR2 == CQ3) {
		/* this test is actually R2 == Q3, written in a form suitable
		   for exact computation with integers */

		/* Due to finite precision some double roots may be missed, and
		   considered to be a pair of complex roots z = x +/- epsilon i
		   close to the real axis. */
		double sqrtQ = sqrt (Q);
		s.push_back( -2 * sqrtQ  - a / 3 );
		s.push_back( sqrtQ - a / 3 );
		s.push_back( sqrtQ - a / 3 );
		sort(s.begin(), s.end());
		return 3 ;
	} else if (CR2 < CQ3)  { /* equivalent to R2 < Q3 */
		double sqrtQ = sqrt (Q);
		double sqrtQ3 = sqrtQ * sqrtQ * sqrtQ;
		double theta = acos (R / sqrtQ3);
		double norm = -2 * sqrtQ;
		s.push_back( norm * cos (theta / 3) - a / 3 );
		s.push_back( norm * cos ((theta + 2.0 * M_PI) / 3) - a / 3 );
		s.push_back( norm * cos ((theta - 2.0 * M_PI) / 3) - a / 3 );

		sort(s.begin(), s.end());
		return 3;
	} else {
		double sgnR = (R >= 0 ? 1 : -1);
		double A = -sgnR * pow (fabs (R) + sqrt (R2 - Q3), 1.0/3.0);
		double B = Q / A ;
		s.push_back( A + B - a / 3 );
		return 1;
	}
}

__END_NAMESPACE_SYNAPS

#endif // SYNAPS_ARITHM_ALGEBRAIC_SOLVE_BY_RADICAL_H

